#include "UserDefinedMapping1.h"
#include "MappingInformation.h"
#include "StretchMapping.h"

static StretchMapping *stretch=NULL;


//\begin{>UserDefinedMapping1Include.tex}{\subsection{Constructor}}
UserDefinedMapping1::
UserDefinedMapping1( )
  : Mapping(2,2,parameterSpace,cartesianSpace)  
//===========================================================================
// /Purpose: Build a user defined Mapping
//\end{UserDefinedMapping1Include.tex}
//===========================================================================
{ 
  UserDefinedMapping1::className="UserDefinedMapping1";
  setName( Mapping::mappingName,"userDefinedMapping1");

  mappingType=unitSquare;

  // mapping is from R^2 -> R^2 by default:
  setDomainDimension(2);
  setRangeDimension(2);

  setGridDimensions( axis1,11 );
  setGridDimensions( axis2,11 );


  setBasicInverseOption(canInvert);  // uncomment if inverse is defined in this file

  mappingHasChanged();

}

// Copy constructor is deep by default
UserDefinedMapping1::
UserDefinedMapping1( const UserDefinedMapping1 & map, const CopyType copyType )
{
  UserDefinedMapping1::className="UserDefinedMapping1";
  if( copyType==DEEP )
  {
    *this=map;
  }
  else
  {
    cout << "UserDefinedMapping1:: sorry no shallow copy constructor, doing a deep! \n";
    *this=map;
  }
}

UserDefinedMapping1::
~UserDefinedMapping1()
{ 
  if( debug & 4 )
    cout << " UserDefinedMapping1::Destructor called" << endl;
  delete stretch;
}

UserDefinedMapping1 & UserDefinedMapping1::
operator=( const UserDefinedMapping1 & x )
{
  if( UserDefinedMapping1::className != x.getClassName() )
  {
    cout << "UserDefinedMapping1::operator= ERROR trying to set a UserDefinedMapping1 = to a" 
      << " mapping of type " << x.getClassName() << endl;
    return *this;
  }
  this->Mapping::operator=(x);            // call = for derivee class
  mappingType=x.mappingType;
  
  rp.redim(0); rp=x.rp;
  ip.redim(0); ip=x.ip;
  
  return *this;
}

void UserDefinedMapping1::
map( const realArray & r, realArray & x, realArray & xr, MappingParameters & params )
{
#ifdef USE_PPP
  Overture::abort("UserDefinedMapping1::map: ERROR: fix me Bill!");
#else
  mapS(r,x,xr,params);
#endif
}

void UserDefinedMapping1::
mapS( const RealArray & r, RealArray & x, RealArray & xr, MappingParameters & params )
// ========================================================================================
// /Description:
//     Define the mapping here.
// /r (input) : parameter space coordinates on the unit line or unit square or unit cube.
// ========================================================================================
{
  if( params.coordinateType != cartesian )
    cerr << "UserDefinedMapping1::map - coordinateType != cartesian " << endl;

  Index I = getIndex( r,x,xr,base,bound,computeMap,computeMapDerivative );

  if( mappingType==unitSquare )
  {
    if( computeMap )
    {
      x(I,axis1)=r(I,axis1); 
      x(I,axis2)=r(I,axis2); 
    }
    if( computeMapDerivative )
    {
      xr(I,axis1,axis1)=1.;
      xr(I,axis1,axis2)=0.;
      xr(I,axis2,axis1)=0.;
      xr(I,axis2,axis2)=1.;
    
    }
  }
  else if( mappingType==helicalWire )
  {
    assert( rp.getLength(0)==10 );
    const real ra=   rp(0);
    const real rb=   rp(1);
    const real za=   rp(2);
    const real zb=   rp(3);
    const real x0=   rp(4);
    const real y0=   rp(5);
    const real rh=   rp(6);
    const real hfreq=rp(7);
    const real arg0= rp(8);

//       printF("The helical write is defined by \n"
//              "   x = xh(t) + rr*cos(theta)\n"
//              "   y = yh(t) + rr*sin(theta)\n"
//              "   y = yh(t) + rr*sin(theta)\n"
//              "   z = za + (zb-za)*t\n"
//              " where\n"
//              "   theta = 2*pi*(r(1)) \n"
//              "   rr = ra+(rb-ra)*(r(2))   (ra,rb are the inner and outer radii of the wire grid)\n"
//              "  t= r(3) \n"
//              "   (xh(t),yh(t)) = (x0,y0) + rh*(cos(arg),sin(arg))  (helix: center of the wire))\n"
//              "  arg = 2*pi*hfreq*(t-arg0) \n"
//              "  hfreq = helix frequency, arg0=helix angle offset\n"
//              "  rh= radius of the helix, phi0=offset for helix angle\n");

    real startAngle=0., endAngle=1.;
     
    const real scale=twoPi*(endAngle-startAngle);
    const real rad=rb-ra;
    const real length = zb-za;

    RealArray angle(I);
    angle=scale*r(I,axis1)+startAngle*twoPi;

#define RADIUS(x) (rad*(x)+ra)

    RealArray phi(I),radius(I);
    radius=RADIUS(r(I,axis2));
    phi=twoPi*hfreq*(r(I,axis3)-arg0);

    if( computeMap )
    {
      RealArray xh(I),yh(I); // helix

      xh=x0+rh*cos(phi);
      yh=y0+rh*sin(phi);

      x(I,0)=radius*cos(angle)+xh;
      x(I,1)=radius*sin(angle)+yh;
      x(I,2)=length*r(I,axis3)+za;
    }
    if( computeMapDerivative )
    {

      xr(I,axis1,axis1)=-radius*scale*sin(angle);
      xr(I,axis2,axis1)= radius*scale*cos(angle);
      xr(I,axis3,axis1)=0;

      xr(I,axis1,axis2)= rad*cos(angle);
      xr(I,axis2,axis2)= rad*sin(angle);
      xr(I,axis3,axis2)=0.;

      xr(I,axis1,axis3)=(-rh*twoPi*hfreq)*sin(phi);
      xr(I,axis2,axis3)=( rh*twoPi*hfreq)*cos(phi);
      xr(I,axis3,axis3)=length;


    }
  }
  else if( mappingType==filletForTwoCylinders )
  {
    // Here is a surface-grid  fillet grid for joining two intersecting cylinders.  
    // 
    // cyl1 is parallel to the x-axis :  y^2 + z^2 = a^2
    // cyl2 is parallel to the y-axis :  x^2 + y^2 = b^2 
    // 
    // The curve of intersection is
    //      x=b*cos(t), z=b*sin(t), y = sqrt( a^2 - (b*sin(t))^2 )

    real a = rp(0);
    real b = rp(1);
    real width=rp(2);

    assert( a>=b-width );

    real aSq=a*a, bSq=b*b;
    
    // RealArray xh(I),yh(I); // helix

    // xh=x0+rh*cos(phi);
    // yh=y0+rh*sin(phi);
    RealArray theta(I), st(I), ct(I), x1(I), y1(I), z1(I), x2(I), y2(I), z2(I), bb(I), temp;
    RealArray s(I), alpha(I), alphas(I), beta(I), betas(I);

    theta(I)=2.*Pi*r(I,0);
    st= sin(theta);
    ct= cos(theta);

    // ****************************************************************************
    //   The fillet is defined as S(r,s) = (1-alpha(s))*S1(r,1-s) + alpha(s)*S2(r,s)
    // ****************************************************************************

    // To make sharper use : 
    //      S(r,s) = (1-alpha(s))*S1(r,beta(1-s)) + alpha(s)*S2(r,beta(s))


    // Here is a function that goes from 0 to 1 and has zero derivatives at 0 and 1: 
    // alpha(I) = 6.*SQR(r(I,1))*( .5 - r(I,1)/3. ); // blending function goes from 0 to 1. 

    assert( stretch!=NULL );
    s(I)=r(I,1);
    stretch->mapS( s,alpha,alphas );  // blending function

    beta = alpha*s;
    betas = alpha + alphas*s;

    RealArray sm(I),alpham(I),alphams(I),betam(I),betams(I);
    sm=1.-s(I);
    stretch->mapS( sm,alpham,alphams ); 
    betam= alpham*sm;
    betams = -alpham - alphams*sm;
    

    // Surface 1:  S1(r,1-s) 
    temp = sqrt( aSq - bSq*SQR(st) );
    x1=b*ct;
    y1=temp + betam*width; 
    z1=b*st;

    // Surface 2: S2(r,s)
    real db=width;
    bb(I) = b + db*beta;
    x2=bb*ct;
    y2=sqrt( aSq - SQR(bb*st) );
    z2=bb*st;

      
    if( computeMap )
    {

      x(I,0)=(1.-alpha)*x1 + alpha*x2;
      x(I,1)=(1.-alpha)*y1 + alpha*y2;
      x(I,2)=(1.-alpha)*z1 + alpha*z2;
    }
    if( computeMapDerivative )
    {
      // check these : 
      xr(I,axis1,0)=(1.-alpha)*(-b*2.*Pi*st)            + alpha*(-bb*2.*Pi*st);
      xr(I,axis2,0)=(1.-alpha)*(-bSq*2.*Pi*st*ct/temp)  + alpha*(-bb*bb*2.*Pi*st*ct/y2);
      xr(I,axis3,0)=(1.-alpha)*( b*2.*Pi*ct)            + alpha*( bb*2.*Pi*ct);

      xr(I,axis1,1)=-alphas*x1 + alphas*x2                           + alpha*db*betas*ct;
      xr(I,axis2,1)=-alphas*y1 + alphas*y2  +(1.-alpha)*betams*width - alpha*bb*betas*st*st*db/y2;
      xr(I,axis3,1)=-alphas*z1 + alphas*z2                           + alpha*db*betas*st;

    }


  }
  else
  {
    printF("UserDefinedMapping1:ERROR: unknown mappingType=%i\n",mappingType);
    Overture::abort();
  }
  
}

void UserDefinedMapping1::
basicInverse( const realArray & x, realArray & r, realArray & rx, MappingParameters & params )
{
#ifdef USE_PPP
  Overture::abort("UserDefinedMapping1::basicInverse: ERROR: fix me Bill!");
#else
  basicInverseS(x,r,rx,params);
#endif
}



void UserDefinedMapping1::
basicInverseS( const RealArray & x, RealArray & r, RealArray & rx, MappingParameters & params )
//==================================================================================
// /Description:
//   Define the inverse here if you know it. You must also set the bascInverseOption in the
//  constructor or else this function will not be called.
//=================================================================================
{
  Index I = getIndex( x,r,rx,base,bound,computeMap,computeMapDerivative );
  if( mappingType==unitSquare )
  {
    if( computeMap )
    {
      r(I,axis1)=x(I,axis1);
      r(I,axis2)=x(I,axis2);
      periodicShift(r,I);   // shift r in any periodic directions
    }
    if( computeMapDerivative )
    {
      rx(I,axis1,axis1)=1.;
      rx(I,axis1,axis2)=0.;
      rx(I,axis2,axis1)=0.;
      rx(I,axis2,axis2)=1.;
    }
  }
  else if( mappingType==helicalWire )
  {
    assert( rp.getLength(0)==10 );
    const real ra=   rp(0);
    const real rb=   rp(1);
    const real za=   rp(2);
    const real zb=   rp(3);
    const real x0=   rp(4);
    const real y0=   rp(5);
    const real rh=   rp(6);
    const real hfreq=rp(7);
    const real arg0= rp(8);

    real startAngle=0., endAngle=1.;
     
    const real scale=twoPi*(endAngle-startAngle), inverseScale=1./scale;
    const real rad=rb-ra, inverseRad=1./rad;
    const real length = zb-za, inverseLength=1./length;


//       printF("The helical write is defined by \n"
//              "   x = xh(t) + rr*cos(theta)\n"
//              "   y = yh(t) + rr*sin(theta)\n"
//              "   y = yh(t) + rr*sin(theta)\n"
//              "   z = za + (zb-za)*t\n"
//              " where\n"
//              "   theta = 2*pi*(r(1)) \n"
//              "   rr = ra+(rb-ra)*(r(2))   (ra,rb are the inner and outer radii of the wire grid)\n"
//              "  t= r(3) \n"
//              "   (xh(t),yh(t)) = (x0,y0) + rh*(cos(arg),sin(arg))  (helix: center of the wire))\n"
//              "  arg = 2*pi*hfreq*(t-arg0) \n"
//              "  hfreq = helix frequency, arg0=helix angle offset\n"
//              "  rh= radius of the helix, phi0=offset for helix angle\n");


//       phi=twoPi*hfreq*(r(I,axis3)-arg0);
//       xh=x0+rh*cos(phi);
//       yh=y0+rh*sin(phi);

//       x(I,0)=radius*cos(angle)+xh;
//       x(I,1)=radius*sin(angle)+yh;
//       x(I,2)=length*r(I,axis3)+za;

    RealArray r3(I), phi(I),xh(I),yh(I); // helix

    r3=(x(I,axis3)-za)*inverseLength;

    phi=twoPi*hfreq*(r3-arg0);
    xh=x0+rh*cos(phi);
    yh=y0+rh*sin(phi);
    
    if( computeMap )
    {
      real theta0=twoPi*startAngle; //  theta1=twoPi*endAngle;

      // **NOTE** atan2(-y,-x) : result in [-pi,pi]
      r(I,axis3)=r3;

      r(I,axis1)=atan2(evaluate(yh-x(I,axis2)),evaluate(xh-x(I,axis1))); 
      r(I,axis1)=( r(I,axis1)+(Pi-theta0) )*inverseScale;
      r(I,axis1)=fmod(r(I,axis1)+1.,1.);  // map back to [0,1]

      r(I,axis2)=(sqrt(SQR(x(I,axis1)-xh)+SQR(x(I,axis2)-yh))-ra)*inverseRad;

      periodicShift(r,I);   // shift r in any periodic directions

    }
    if( computeMapDerivative )
    {
      RealArray radius(I);
      radius=sqrt(SQR(x(I,axis1)-xh)+SQR(x(I,axis2)-yh));

      rx(I,axis2,axis1)= (x(I,axis1)-xh)/(rad*radius);
      rx(I,axis2,axis2)= (x(I,axis2)-yh)/(rad*radius);
      rx(I,axis2,axis3)= ( (x(I,axis1)-xh)*( sin(phi)) + 
			   (x(I,axis2)-yh)*(-cos(phi)) )*(rh*twoPi*hfreq/(length*rad))/radius;

      radius=inverseScale/(SQR(radius));    // ** change defn of radius! **
      rx(I,axis1,axis1)=-(x(I,axis2)-yh)*radius;
      rx(I,axis1,axis2)= (x(I,axis1)-xh)*radius;
      rx(I,axis1,axis3)=((x(I,axis2)-yh)*(sin(phi))+
			 (x(I,axis1)-xh)*(cos(phi)) )*(-rh*twoPi*hfreq/length)*radius;

      rx(I,axis3,axis1)=0.;
      rx(I,axis3,axis2)=0.;
      rx(I,axis3,axis3)=inverseLength;

//       radius=sqrt(SQR(x(I,cylAxis1)-xc[cylAxis1])+SQR(x(I,cylAxis2)-xc[cylAxis2]));
//       // radius=pow(pow(x(I,cylAxis1)-xc[cylAxis1],2)+pow(x(I,cylAxis2)-xc[cylAxis2],2),.5);
//       rx(I,axis3,cylAxis1)= (x(I,cylAxis1)-xc[cylAxis1])/(rad*radius);
//       rx(I,axis3,cylAxis2)= (x(I,cylAxis2)-xc[cylAxis2])/(rad*radius);
//       rx(I,axis3,cylAxis3)=0.;
//       radius=inverseScale/(SQR(radius));    // ** change defn of radius! **
//       rx(I,axis1,cylAxis1)=-(x(I,cylAxis2)-xc[cylAxis2])*radius;
//       rx(I,axis1,cylAxis2)= (x(I,cylAxis1)-xc[cylAxis1])*radius;
//       rx(I,axis1,cylAxis3)=0.;

    }
  }
  else
  {
    printF("UserDefinedMapping1:ERROR: unknown mappingType=%i\n",mappingType);
    Overture::abort();
  }
  
}
  

//=================================================================================
// get a mapping from the database
//=================================================================================
int UserDefinedMapping1::
get( const GenericDataBase & dir, const aString & name)
{
  GenericDataBase & subDir = *dir.virtualConstructor();
  dir.find(subDir,name,"Mapping");

  subDir.get( UserDefinedMapping1::className,"className" ); 
  if( UserDefinedMapping1::className != "UserDefinedMapping1" )
  {
    cout << "UserDefinedMapping1::get ERROR in className!" << endl;
  }
  int temp;
  subDir.get( temp,"mappingType" ); mappingType=(UserDefinedMappingEnum)temp;
  subDir.get( rp,"rp" );
  subDir.get( ip,"ip" );

  Mapping::get( subDir, "Mapping" );
  delete &subDir;
  mappingHasChanged();
  return 0; 
}

int UserDefinedMapping1::
put( GenericDataBase & dir, const aString & name) const
{  
  GenericDataBase & subDir = *dir.virtualConstructor();      // create a derived data-base object
  dir.create(subDir,name,"Mapping");                      // create a sub-directory 

  subDir.put( UserDefinedMapping1::className,"className" );
  subDir.put( (int)mappingType,"mappingType" ); 
  subDir.put( rp,"rp" );
  subDir.put( ip,"ip" );            
  Mapping::put( subDir, "Mapping" );
  delete &subDir;
  return 0;
}

Mapping *UserDefinedMapping1::
make( const aString & mappingClassName )
{ // Make a new mapping if the mappingClassName is the name of this Class
  Mapping *retval=0;
  if( mappingClassName==UserDefinedMapping1::className )
    retval = new UserDefinedMapping1();
  return retval;
}

//\begin{>>UserDefinedMapping1Include.tex}{\subsection{getParameters}}
void UserDefinedMapping1::
getParameters(IntegerArray & ipar, RealArray & rpar ) const
//===========================================================================
// /Description: 
//   Return the current values for the parameters.
//\end{UserDefinedMapping1Include.tex}
//===========================================================================
{
  ipar.redim(0); ipar=ip;
  rpar.redim(0); rpar=rp;
}



//\begin{>>UserDefinedMapping1Include.tex}{\subsection{setParameters}}
void UserDefinedMapping1::
setParameters(const IntegerArray & ipar, 
              const RealArray & rpar )
//===========================================================================
// /Description:
//    Set any parameters. These are up to you to define and use as appropriate.
//\end{UserDefinedMapping1Include.tex}
//===========================================================================
{
  ip.redim(0); ip=ipar;
  rp.redim(0); rp=rpar;
}

    

//=============================================================================
//   Prompt for changes to parameters
//   
//=============================================================================
int UserDefinedMapping1::
update( MappingInformation & mapInfo ) 
{

  assert(mapInfo.graphXInterface!=NULL);
  GenericGraphicsInterface & gi = *mapInfo.graphXInterface;
  
  char buff[180];  // buffer for sprintf
//   aString menu[] = 
//     {
//       "!UserDefinedMapping1",
//       "unit square",
//       "helical wire",
//       " ",
//       "lines",
//       "boundary conditions",
//       "share",
//       "mappingName",
//       "periodicity",
//       "show parameters",
//       "plot",
//       "help",
//       "exit", 
//       "" 
//      };

  // defaults for helical wire: 
  real ra=.2, rb=.3;
  real za=0., zb=5.;
  real x0=0., y0=0., rh=.5;
  real hfreq=1., arg0=0.;
  if( mappingType==helicalWire && rp.getLength(0)==10 )
  {
    ra=   rp(0);
    rb=   rp(1);
    za=   rp(2);
    zb=   rp(3);
    x0=   rp(4);
    y0=   rp(5);
    rh=   rp(6);
    hfreq=rp(7);
    arg0= rp(8);
  }
  

  GUIState dialog;
  bool buildDialog=true;
  if( buildDialog )
  {
    dialog.setWindowTitle("UserDefinedMapping");
    dialog.setExitCommand("exit", "exit");

    // option menus
    dialog.setOptionMenuColumns(1);

    aString opCommand1[] = {"unit square",
			    "helical wire",
                            "fillet for two cylinders",
			    ""};
    
    dialog.addOptionMenu( "type:", opCommand1, opCommand1, mappingType); 


    aString cmds[] = {"lines",
		      "boundary conditions",
		      "share",
		      "mappingName",
		      "periodicity",
		      "show parameters",
		      "plot",
                      "check",
                      "check inverse",
		      "help",
		      ""};
    int numberOfPushButtons=10;  // number of entries in cmds
    int numRows=(numberOfPushButtons+1)/2;
    dialog.setPushButtons( cmds, cmds, numRows ); 

//   bool plotVolumeGrid=false;
//   bool plotSurfaceGrid=true;
//   bool plotStitching=true;
//   bool interactiveStitcher=true;
//   aString tbCommands[] = {"plot volume grid",
//                           "plot surface grid",
// 			  "plot stitching",
//                           "interactive stitcher",
// 			  ""};
//   int tbState[10];
//   tbState[0] = plotVolumeGrid; 
//   tbState[1] = plotSurfaceGrid; 
//   tbState[2] = plotStitching;
//   tbState[3] = interactiveStitcher;
//   int numColumns=1;
//   dialog.setToggleButtons(tbCommands, tbCommands, tbState, numColumns);

    const int numberOfTextStrings=7;  // max number allowed
    aString textLabels[numberOfTextStrings];
    aString textStrings[numberOfTextStrings];


    int nt=0;
    textLabels[nt] = "ra,rb:";  sPrintF(textStrings[nt],"%g,%g",ra,rb);  nt++; 
    textLabels[nt] = "za,zb:";  sPrintF(textStrings[nt],"%g,%g",za,zb);  nt++; 
    textLabels[nt] = "x0,y0,rh:";  sPrintF(textStrings[nt],"%g,%g,%g",x0,y0,rh);  nt++; 
    textLabels[nt] = "hfreq,arg0:";  sPrintF(textStrings[nt],"%g,%g",hfreq,arg0);  nt++; 

    // null strings terminal list
    textLabels[nt]="";   textStrings[nt]="";  assert( nt<numberOfTextStrings );
    dialog.setTextBoxes(textLabels, textLabels, textStrings);

    // dialog.buildPopup(menu);
    gi.pushGUI(dialog);
  }
  


  aString help[] = 
    {
      "lines              : specify number of grid lines",
      "boundary conditions: specify boundary conditions",
      "share              : specify share values for sides",
      "mappingName        : specify the name of this mapping",
      "periodicity        : specify periodicity in each direction",
      "show parameters    : print current values for parameters",
      "plot               : enter plot menu (for changing ploting options)",
      "help               : Print this list",
      "exit               : Finished with changes",
      "" 
    };

  aString answer,line; 

  bool plotObject=TRUE;
  GraphicsParameters parameters;
  parameters.set(GI_PLOT_THE_OBJECT_AND_EXIT,TRUE);

  gi.appendToTheDefaultPrompt("Square>"); // set the default prompt
  int len=0;
  for( int it=0;; it++ )
  {
    if( it==0 && plotObject )
      answer="plotObject";
    else
      gi.getAnswer(answer,"");
 

    if( answer=="unit square" )
    {
      mappingType=unitSquare;

      setDomainDimension(2);
      setRangeDimension(2);

      setGridDimensions( axis1,11 );
      setGridDimensions( axis2,11 );

      mappingHasChanged();

    }
    else if( answer=="helical wire" )
    {
      mappingType=helicalWire;

      setDomainDimension(3);
      setRangeDimension(3);

      setGridDimensions( axis1,41 ); // theta
      setGridDimensions( axis2,7  ); // r 
      setGridDimensions( axis3,101); // z 

      signForJacobian=-1.;            // mapping is left handed

      setIsPeriodic(axis1,functionPeriodic );       
      setBoundaryCondition( Start,axis1,-1 );
      setBoundaryCondition(   End,axis1,-1 );

      setBoundaryCondition( Start,axis2, 1 );
      setBoundaryCondition(   End,axis2, 2 );
      setBoundaryCondition( Start,axis3, 3 );
      setBoundaryCondition(   End,axis3, 4 );

      if( !gi.readingFromCommandFile() )
      {
	printF("The helical wire is defined by \n"
	       "   x = xh(t) + rr*cos(theta)\n"
	       "   y = yh(t) + rr*sin(theta)\n"
	       "   y = yh(t) + rr*sin(theta)\n"
	       "   z = za + (zb-za)*t\n"
	       " where\n"
	       "   theta = 2*pi*(r(1)) \n"
	       "   rr = ra+(rb-ra)*(r(2))   (ra,rb are the inner and outer radii of the wire grid)\n"
	       "  t= r(3) \n"
	       "   (xh(t),yh(t)) = (x0,y0) + rh*(cos(arg),sin(arg))  (helix: center of the wire))\n"
	       "  arg = 2*pi*hfreq*(t-arg0) \n"
	       "  hfreq = helix frequency, arg0=helix angle offset\n"
	       "  rh= radius of the helix, phi0=offset for helix angle\n");
      }
      
      if( rp.getLength(0)!=10 )
      {
        rp.redim(10);
      }
      
      rp(0)=ra;
      rp(1)=rb;
      rp(2)=za;
      rp(3)=zb;
      rp(4)=x0;
      rp(5)=y0;
      rp(6)=rh;
      rp(7)=hfreq;
      rp(8)=arg0;

      mappingHasChanged();
      
    }
    else if( len=answer.matches("ra,rb:") )
    {
      sScanF(answer(len,answer.length()-1),"%e %e",&ra,&rb);
      if( !gi.isGraphicsWindowOpen() )
        dialog.setTextLabel("ra,rb:",sPrintF(answer,"%g,%g",ra,rb));
      rp(0)=ra;
      rp(1)=rb;
      mappingHasChanged();
    }
    else if( len=answer.matches("za,zb:") )
    {
      sScanF(answer(len,answer.length()-1),"%e %e",&za,&zb);
      if( !gi.isGraphicsWindowOpen() )
        dialog.setTextLabel("za,zb:",sPrintF(answer,"%g,%g",za,zb));
      rp(2)=za;
      rp(3)=zb;
      mappingHasChanged();
    }
    else if( len=answer.matches("x0,y0,rh:") )
    {
      sScanF(answer(len,answer.length()-1),"%e %e %e",&x0,&y0,&rh);
      if( !gi.isGraphicsWindowOpen() )
        dialog.setTextLabel("x0,y0,rh:",sPrintF(answer,"%g,%g,%g",x0,y0,rh));
      rp(4)=x0;
      rp(5)=y0;
      rp(6)=rh;
      mappingHasChanged();
    }
    else if( len=answer.matches("hfreq,arg0:") )
    {
      sScanF(answer(len,answer.length()-1),"%e %e",&hfreq,&arg0);
      if( !gi.isGraphicsWindowOpen() )
        dialog.setTextLabel("hfreq,arg0:",sPrintF(answer,"%g,%g",hfreq,arg0));
      rp(7)=hfreq;
      rp(8)=arg0;
      mappingHasChanged();
    }
    else if( answer=="fillet for two cylinders" )
    {
      mappingType=filletForTwoCylinders;
      setBasicInverseOption(canDoNothing); 

      if( rp.getLength(0)!=10 )
      {
        rp.redim(10); rp=0.;
      }
      rp(0)=1.;  // radius a
      rp(1)=.5;  // radius b

      gi.inputString(answer,"Enter a,b,d  (a=cyl1-radius, b=cyl2-radius, d=width width a>= b-width)");
      sScanF(answer(len,answer.length()-1),"%e %e %e",&rp(0),&rp(1),&rp(2));
      printF("Setting a=%e, b=%e, d=%e\n",rp(0),rp(1),rp(2));

      // if( stretch==NULL ) stretch = new StretchMapping(StretchMapping::exponentialBlend);

      if( stretch==NULL ) stretch = new StretchMapping(StretchMapping::hyperbolicTangent);
      stretch->setIsNormalized(false);
      real beta=5.; // 10.;
      stretch->setHyperbolicTangentParameters(.5,0.,.5,beta,.5);  


      setDomainDimension(2);
      setRangeDimension(3);

      setGridDimensions( axis1,41 ); // theta
      setGridDimensions( axis2,7  ); // r 

      setIsPeriodic(axis1,functionPeriodic );       
      setBoundaryCondition( Start,axis1,-1 );
      setBoundaryCondition(   End,axis1,-1 );

      setBoundaryCondition( Start,axis2, 1 );
      setBoundaryCondition(   End,axis2, 2 );
//       setBoundaryCondition( Start,axis3, 3 );
//       setBoundaryCondition(   End,axis3, 4 );

      mappingHasChanged();
    }
    else if( answer=="show parameters" )
    {
      rp.display("real parameters");
      ip.display("int parameters");
      display();
    }
    else if( answer=="plot" )
    {
      parameters.set(GI_PLOT_THE_OBJECT_AND_EXIT,FALSE);
      parameters.set(GI_TOP_LABEL,getName(mappingName));
      gi.erase();
      PlotIt::plot(gi,*this,parameters); 
      parameters.set(GI_PLOT_THE_OBJECT_AND_EXIT,TRUE);
    }
    else if( answer=="help" )
    {
      for( int i=0; help[i]!=""; i++ )
        gi.outputString(help[i]);
    }
    else if( answer=="lines"  ||
             answer=="boundary conditions"  ||
             answer=="share"  ||
             answer=="mappingName"  ||
             answer=="periodicity" ||
             answer=="check" ||
             answer=="check inverse" )
    { // call the base class to change these parameters:
      mapInfo.commandOption=MappingInformation::readOneCommand;
      mapInfo.command=&answer;
      Mapping::update(mapInfo); 
      mapInfo.commandOption=MappingInformation::interactive;
    }
    else if( answer=="exit" )
      break;
    else if( answer=="plotObject" )
      plotObject=TRUE;
    else 
    {
      gi.outputString( sPrintF(buff,"Unknown response=%s",(const char*)answer) );
      gi.stopReadingCommandFile();
    }

    if( plotObject )
    {
      parameters.set(GI_TOP_LABEL,getName(mappingName));
      gi.erase();
      PlotIt::plot(gi,*this,parameters);   // *** recompute every time ?? ***

    }
  }
  gi.erase();
  gi.unAppendTheDefaultPrompt();  // reset

  if( buildDialog )
  {
    gi.popGUI(); // restore the previous GUI
  }
  return 0;
}
