#include "DataPointMapping.h"
#include "ParallelUtility.h"
#include "display.h"
#include "Inverse.h"          
#include "DistributedInverse.h"




void DataPointMapping::
mapS( const RealArray & r, RealArray & x, RealArray & xr, MappingParameters & params )
// ====================================================================================================
// /Description:
//    Evaluate the Mapping.
// ===================================================================================================
{
  if( params.coordinateType != cartesian )
    cerr << "DataPointMapping::map - coordinateType != cartesian " << endl;


#ifndef USE_PPP
  Index I = getIndex( r,x,xr,base,bound,computeMap,computeMapDerivative );
  mapScalar( r,x,xr,params,base,bound,computeMap,computeMapDerivative );
  return;
#else

  // ****************************************
  // ******* Parallel map function **********
  // ****************************************

  // ****** NOTE: this code is very similar to that in inverseMap.C

  int debug=3;
  
  if( debug >0 )
    openDebugFiles();


  Index I = getIndex( r,x,xr,base,bound,computeMap,computeMapDerivative );

  const MPI_Comm & OV_COMM = Overture::OV_COMM;
  
  const int np= max(1,Communication_Manager::numberOfProcessors());
  const int myid=max(0,Communication_Manager::My_Process_Number);

  const real * rp = r.Array_Descriptor.Array_View_Pointer1;
  const int rDim0=r.getRawDataSize(0);
#undef R
#define R(i0,i1) rp[i0+rDim0*(i1)]
#undef X
  real * xp = x.Array_Descriptor.Array_View_Pointer1;
  const int xDim0=x.getRawDataSize(0);
#define X(i0,i1) xp[i0+xDim0*(i1)]
  real * xrp = xr.Array_Descriptor.Array_View_Pointer2;
  const int xrDim0=xr.getRawDataSize(0);
  const int xrDim1=xr.getRawDataSize(1);
#undef XR
#define XR(i0,i1,i2) xrp[i0+xrDim0*(i1+xrDim1*(i2))]


  if( debug & 1 )
  {
    fprintf(pDebugFile,"DPM:mapS: myid=%i base=%i bound=%i computeMap=%i computeDer=%i name=%s \n",
	    myid,base,bound,computeMap,computeMapDerivative,
	    (const char*)getName(mappingName));
  }
      
  realArray & vertex = xy;  //   this array determines the parallel distribution 

  int numPoints=bound-base+1;

  int *nqs = new int [np];  // nqs[p] : number of queries to send to processor p
  int *nqr = new int [np];  // nqs[p] : number of queries received from processor p
  for( int p=0; p<np; p++ ) nqs[p]=0;
    
  // ip(j,p) = i, j=0,..,nqs[p] : means x(i,.) is the j'th point sent to processor p 
  int *pip=NULL;
  if( numPoints>0 )
    pip = new int [numPoints*np];    // ********* this is too large in general, use a SparseArray ? 
#define ip(j,p) (pip[(p)+np*(j)])

  real dr[3];
  for( int axis=0; axis<3; axis++ )
    dr[axis]=1./max(gridIndexRange(1,axis)-gridIndexRange(0,axis),1);

  int iv[3]={0,0,0}; //
  for( int i=base; i<=bound; i++ )
  {
    for( int axis=0; axis<domainDimension; axis++ )
    { // index = index of closest grid point on the donor grid -- is this correct ? 
      real rval =R(i,axis);
      if( isPeriodic[axis] ) rval=fmod(rval+1.,1.); // periodic wrap to [0,1]

      iv[axis] = int( rval/dr[axis]+ gridIndexRange(0,axis) +.5 );
    }

    int p = vertex.Array_Descriptor.findProcNum( iv );    // index iv[] is on processor p 

    ip(nqs[p],p)=i;  // this point should be sent to processor p for an answer 
    nqs[p]++;
  }
  if( debug & 1 )
  {
    for( int p=0; p<np; p++ ) 
      fprintf(pDebugFile,"DPM:mapS: myid=%i p=%i nqs[p]=%i\n",myid,p,nqs[p]);
  }
  

  // --- send/receive number of queries ---
  // we need to send computeMap and computeMapDerivative since these may not be the
  // same on all processors!

  int myInfo[3]={0,computeMap,computeMapDerivative}; //
  // info(0:2,p) : holds (nqr[p],computeMap[p],computeMapDerivative[p])
  int *pinfo = new int [3*np];
#define info(i,p) pinfo[(i)+3*(p)]

  MPI_Status status;
  const int tag0=309278;
  int numberOfQueries=0;      // total number of queries that we will receieve 
  for( int p=0; p<np; p++ )
  {
    // *** send computeMap computeMapDerivative here too !

    int tags=tag0+p, tagr=tag0+myid;
    myInfo[0]=nqs[p];
    MPI_Sendrecv(myInfo,     3, MPI_INT, p, tags, 
		 &info(0,p), 3, MPI_INT, p, tagr, OV_COMM, &status ); 

    nqr[p]=info(0,p);
    if( debug & 1 ) 
    {
      fprintf(pDebugFile,"DPM:mapS: myid=%i send %i queries to p=%i (map=%i,der=%i), "
	      "receive %i queries from p=%i, p: (map=%i der=%i)\n",
	      myid,nqs[p],p,computeMap,computeMapDerivative, nqr[p],p,info(1,p),info(2,p));
      fflush(pDebugFile);
    }
	
    numberOfQueries += nqr[p];
  }

  // For now -- we compute r and rx for the queries if there is any processor wants these results,
  // but we do not return rx if the requesting processor does noy want them.

  int computeMapAnswer=false, computeMapDerivativeAnswer=false;
  for( int p=0; p<np; p++ )
  {
    computeMapAnswer = computeMapAnswer || info(1,p);
    computeMapDerivativeAnswer = computeMapDerivativeAnswer || info(2,p);
  }

  // ******* BUFFERS ************
  //  sqb[p] : send queries buffer -- send queries to processor p
  //  rqb[p] : receive queries buffer
  //
  //  sqb[p] : send answers buffer -- send answers to processor p
  //  rqb[p] : receieve answers buffer 

  const int numDataToSendPerPoint=rangeDimension+domainDimension;  // we send x and r 

  // --- Allocate receive buffers ---
  real **rqb = new real* [np];
  for( int p=0; p<np; p++ )
  {
    rqb[p] = new real [max(1,nqr[p]*numDataToSendPerPoint)];
  }
  
  // --- post receives ---
  MPI_Request *receiveRequest     = new MPI_Request[np];  
  MPI_Request *sendRequest        = new MPI_Request[np];  
  MPI_Request *sendRequestAnswers = new MPI_Request[np];  
  const int tag1=2529743;
  for( int p=0; p<np; p++ )
  {  
    int tag=tag1+myid;
    MPI_Irecv( rqb[p], nqr[p]*numDataToSendPerPoint, MPI_Real, p, tag, OV_COMM, &receiveRequest[p] );
  }

  // -- allocate send buffers ---

  // pxs[p](i,m) : holds x and r values m=0,1,...,rangeDimension+domainDimension-1
  real **sqb = new real* [np];
  for( int p=0; p<np; p++ )
  {
    sqb[p] = new real [max(1,nqs[p]*numDataToSendPerPoint)];
  }

  // --- fill in send buffers ---
#define xs(i,n,p) sqb[p][(i)+nqs[p]*(n)]
#define rs(i,m,p) sqb[p][(i)+nqs[p]*(m+rangeDimension)]
  for( int p=0; p<np; p++ )
  {
    for( int j=0; j<nqs[p]; j++ )
    {   
      int i = ip(j,p);
      if( computeMap )
        for( int n=0; n<rangeDimension; n++ ) xs(j,n,p) = X(i,n); 

      for( int m=0; m<domainDimension; m++ )
      {
	real rval =R(i,m);
	if( isPeriodic[m] ) rval=fmod(rval+1.,1.); // periodic wrap to [0,1]
	rs(j,m,p) = rval;
      }
      
    }
  }
    
  // --- send buffers ---
  for( int p=0; p<np; p++ )
  {
    if( debug & 1 ) 
    {
      fprintf(pDebugFile,"DPM:mapS: Send %i queries from myid=%i to p=%i\n",nqs[p],myid,p);
      fflush(pDebugFile);
    }
	
    int tag=tag1+p;
    MPI_Isend(sqb[p], nqs[p]*numDataToSendPerPoint, MPI_Real, p, tag, OV_COMM, &sendRequest[p] );  
  }
  
  // --- wait for all the receives to finish ---
  MPI_Status *receiveStatus= new MPI_Status[np];  
  MPI_Waitall(np,receiveRequest,receiveStatus);


  if( true || debug & 4 )
  { // sanity check:

    fflush(0);
    Communication_Manager::Sync();

    for( int p=0; p<np; p++ )
    {
      int num=0;
      MPI_Get_count( &receiveStatus[p], MPI_Real, &num );
      assert( num==nqr[p]*numDataToSendPerPoint );
    }
  }


  // --- post receives for answers (which have not even been computed yet) ---
  //     return r and maybe rx values 
  int *numDataToReceivePerPoint = new int[np];
  real **rab = new real* [np];  // rab = "answers" 
  for( int p=0; p<np; p++ )
  {
    // return r, and optionally rx values

    // something is wrong here (c.f. stir.cmd)
    // ***        numDataToReceivePerPoint[p] =rangeDimension + info(2,p)*domainDimension*rangeDimension;  

    numDataToReceivePerPoint[p] =rangeDimension + domainDimension*rangeDimension;  
    // we receive back the number we sent: nqs[p] 
    rab[p] = new real [max(1,nqs[p]*numDataToReceivePerPoint[p])];
  }

  const int tag2=1329743;
  for( int p=0; p<np; p++ )
  {  
    int tag=tag2+myid;
    if( debug & 1 )
    {
      fprintf(pDebugFile,"DPM:mapS: myid=%i, expect %i answers from p=%i\n",myid,nqs[p],p);
      fflush(pDebugFile);
    }
	
    MPI_Irecv( rab[p], nqs[p]*numDataToReceivePerPoint[p], MPI_Real, p, tag, OV_COMM, &receiveRequest[p] );
  }

  // ------------------------------------
  // --- un-pack the values received ----
  // ------------------------------------

  RealArray xc,rc,rxc;
  if( numberOfQueries>0 )
  {
    xc.redim(numberOfQueries,rangeDimension);
    rc.redim(numberOfQueries,domainDimension);

    if( computeMapDerivativeAnswer )
      rxc.redim(numberOfQueries,rangeDimension,domainDimension); 
  }

  real * rcp = rc.Array_Descriptor.Array_View_Pointer1;
  const int rcDim0=rc.getRawDataSize(0);
#undef RC
#define RC(i0,i1) rcp[i0+rcDim0*(i1)]
#undef XC
  real * xcp = xc.Array_Descriptor.Array_View_Pointer1;
  const int xcDim0=xc.getRawDataSize(0);
#define XC(i0,i1) xcp[i0+xcDim0*(i1)]
  real * rxcp = rxc.Array_Descriptor.Array_View_Pointer2;
  const int rxcDim0=rxc.getRawDataSize(0);
  const int rxcDim1=rxc.getRawDataSize(1);
#undef RXC
#define RXC(i0,i1,i2) rxcp[i0+rxcDim0*(i1+rxcDim1*(i2))]


#define xrb(i,n,p) rqb[p][(i)+nqr[p]*(n)]
#define rrb(i,m,p) rqb[p][(i)+nqr[p]*(m+rangeDimension)]
  int i=0;
  for( int p=0; p<np; p++ )
  {
    for( int j=0; j<nqr[p]; j++ )
    {
      for( int n=0; n<rangeDimension; n++ )  XC(i,n) = xrb(j,n,p); 
      for( int m=0; m<domainDimension; m++ ) RC(i,m) = rrb(j,m,p);
  
      i++;
    }
  }
  assert( i==numberOfQueries );

  // we can delete rqb[p]  now 
  for( int p=0; p<np; p++ )
    delete [] rqb[p];
  delete [] rqb;

  if( debug & 2 )
  {
    for( int i=0; i<numberOfQueries; i++ )
    {
      fprintf(pDebugFile,"DPM:mapS: myid=%i check point i=%i : xc=(%g,%g,%g), rc=(%g,%g,%g)\n",myid,i,
	      XC(i,0),(rangeDimension>1 ? XC(i,1) : 0.), (rangeDimension>2 ? XC(i,2) : 0.),
	      RC(i,0),(domainDimension>1 ? RC(i,1) : 0.), (domainDimension>2 ? RC(i,2) : 0.));

    }
    fflush(pDebugFile);
    Communication_Manager::Sync();
    MPI_Barrier(OV_COMM);
	
  }
	
  Index J=Range(numberOfQueries);


  // pass base,bound,computeMap,computeMapDerivative 

  // **********************************
  // ****** Evaluate the Mapping ******
  // **********************************
  if( numberOfQueries>0 )
    mapScalar( rc,xc,rxc,params, 0,numberOfQueries-1,computeMapAnswer,computeMapDerivativeAnswer );
//    mapScalar( rc,xc,rxc,params, 0,numberOfQueries-1,computeMapAnswer,computeMapDerivativeAnswer );



  if( debug & 1 )
  {
    for( int i=0; i<numberOfQueries; i++ )
    {
      fprintf(pDebugFile,"DPM:mapS: myid=%i after mapScalar, point i=%i : xc=(%g,%g,%g), rc=(%g,%g,%g)\n",myid,i,
	      XC(i,0),(rangeDimension>1 ? XC(i,1) : 0.), (rangeDimension>2 ? XC(i,2) : 0.),
	      RC(i,0),(domainDimension>1 ? RC(i,1) : 0.), (domainDimension>2 ? RC(i,2) : 0.));
    }
    fflush(pDebugFile);
  }


  // -----------------------------------------------
  // --- Send answers back : pack into a buffer  ---
  // -----------------------------------------------

  real **sab = new real* [np];  // sab = send-answers buffer
  for( int p=0; p<np; p++ )
  {
    // we send back the number we received : nqr[p]
    sab[p] = new real [max(1,nqr[p]*numDataToReceivePerPoint[p])];
  }

#undef rrb
#define rrb(i,m,p) sab[p][(i)+nqr[p]*(m)]
#define rxb(i,m,n,p) sab[p][(i)+nqr[p]*((m)+rangeDimension*(n+1))]
  i=0;
  for( int p=0; p<np; p++ )
  {
    for( int j=0; j<nqr[p]; j++ )
    {
      for( int m=0; m<rangeDimension; m++ ) rrb(j,m,p) = XC(i,m);
      if( info(2,p) ) // true if we are to computeMapDerivative for processor p
      {
	for( int m=0; m<domainDimension; m++ )
	{
	  for( int n=0; n<rangeDimension; n++ )  
	  {
	    rxb(j,m,n,p) = RXC(i,m,n);
	  }
	}
      }
	  
      i++;
    }
  }

  // --------------------
  // --- Send buffers ---
  // --------------------

  for( int p=0; p<np; p++ )
  {
    int tag=tag2+p;
    if( debug & 1 )
    {
      fprintf(pDebugFile,"DPM:mapS: myid=%i, send %i answers to p=%i\n",myid,nqr[p],p);
      fflush(pDebugFile);
    }
	
    MPI_Isend(sab[p], nqr[p]*numDataToReceivePerPoint[p], MPI_Real, p, tag, OV_COMM, &sendRequestAnswers[p] );  
  }

  if( debug & 4 )
    MPI_Barrier(OV_COMM);
      
  // --- wait for all the receives to finish ---
  MPI_Waitall(np,receiveRequest,receiveStatus);

  if( true || debug & 4 )
  { // sanity check:
    for( int p=0; p<np; p++ )
    {
      int num=0;
      MPI_Get_count( &receiveStatus[p], MPI_Real, &num );
      assert( num==nqs[p]*numDataToReceivePerPoint[p] );
    }
  }

  // --- assign results back into the r,xr arrays ---

#undef rrb
#define rrb(i,m,p) rab[p][(i)+nqs[p]*(m)]
#undef rxb 
#define rxb(i,m,n,p) rab[p][(i)+nqs[p]*((m)+rangeDimension*(n+1))]

  real rv[3];
  for( int p=0; p<np; p++ )
  {
    for( int j=0; j<nqs[p]; j++ )
    {
      i = ip(j,p);  // the result goes here 
      assert( i>=base && i<=bound );
	  
      if( computeMap )
      {
	for( int m=0; m<rangeDimension; m++ ) X(i,m) = rrb(j,m,p);

	if( debug & 1 )
	{
	  fprintf(pDebugFile,"DPM:mapS: myid=%i RESULTS for point i=%i : x=(%g,%g,%g), r=(%g,%g,%g)\n",myid,i,
		  X(i,0),(rangeDimension>1 ? X(i,1) : 0.), (rangeDimension>2 ? X(i,2) : 0.),
		  R(i,0),(domainDimension>1 ? R(i,1) : 0.), (domainDimension>2 ? R(i,2) : 0.));
	}
      }
	  
      if( computeMapDerivative )
      {
	for( int m=0; m<domainDimension; m++ )
	{
	  for( int n=0; n<rangeDimension; n++ )  
	  {
	    XR(i,m,n) = rxb(j,m,n,p);
	  }
	}
      }
	  
      i++;
    }
  }

  if( debug>0 )
  {
    fprintf(pDebugFile,"DPM:mapS:DONE: myid=%i\n",myid);
    fflush(pDebugFile);
  }
      
  // wait for ALL sends to finish on this processor before we can clean up
  MPI_Waitall(np,sendRequest,receiveStatus);
  MPI_Waitall(np,sendRequestAnswers,receiveStatus);

  delete [] pinfo;
  delete [] numDataToReceivePerPoint;
  delete [] nqs;
  delete [] nqr;
  delete [] pip;
  delete [] receiveRequest;
  delete [] receiveStatus;
  delete [] sendRequest;
  delete [] sendRequestAnswers;
  for( int p=0; p<np; p++ )
  {
    delete [] sqb[p];
    delete [] sab[p];
    delete [] rab[p];
  }
  delete [] sqb;
  delete [] sab;
  delete [] rab;
    
  return;

#endif
}
