#include "Mapping.h"
#include "ParallelUtility.h"
#include "display.h"
#include "Inverse.h"           // defines global and local inverses
#include "DistributedInverse.h"


// *serial-array version*
void Mapping::
inverseMapS( const RealArray & x, 
             RealArray & r, 
	     RealArray & rx /* =Overture::nullRealArray() */,
	     MappingParameters & params /* =Overture::nullMappingParameters() */ )
// =====================================================================================
// /Description:
//  --- Here is the generic inverse ----
//
// /x (input) : invert these points. The dimensions of this array will determine which
//     points are inverted.
// /r (input/output) : On input this is an initial guess. If you know a reasonable initial
//   guess then supply it, If you don't know an initial guess
//    then set r=-1. for those points that you do not know a guess. If you do not know a guess
//     then do NOT specify some valid value like .5 since this will probably be slower than allowing
//     the value to be automatically generated.
// /rx (output): the derivatives of the inverse mapping.
//  /params (input) :
//   \begin{description}
//     \item[params.computeGlobalInverse] : TRUE means compute a full global inverse,
//       FALSE means only compute a local inverse using the initial guess supplied in r
//     \item[params.periodicityOfSpace] : 
//     \item[params.periodVector] : 
//   \end{description}
//\end{MappingInclude.tex}
//=================================================================================
{
  int debug = 0; // 3;
  // Mapping::debug=3;
  
  if( debug & 8 )
    cout << getName(mappingName) << ": inverseMapS - params.computeGlobalMap=" 
         << params.computeGlobalInverse << endl;

  if( basicInverseOption==canInvert && params.periodicityOfSpace==0 )
  {
    basicInverseS( x,r,rx,params );  // can use user supplied inverse
    return;
  }

#ifndef USE_PPP
  // *serial inverse*
  MappingWorkSpace workSpace; 
  if( params.computeGlobalInverse )
  {
    // first get the initial guess
    approximateGlobalInverse->inverse( x,r,rx,workSpace,params );

    // Now do Newton to Invert:
    exactLocalInverse->inverse( x,r,rx,workSpace,TRUE );   // TRUE means use results found in the
                                                           // workSpace
  }
  else
  {
    exactLocalInverse->inverse( x,r,rx,workSpace,FALSE );
  }
#else

  // ***** parallel inverse *****

//   int myid=max(0,Communication_Manager::My_Process_Number);
//   printf("inverseMapS: myid=%i usesDistributedInverse=%i x=[%i,%i]\n",myid,(int)usesDistributedInverse(),
//           x.getBase(0),x.getBound(0));

  if( !usesDistributedInverse() )
  {
    // ***** this mapping can be inverted with no parallel communication ****

    MappingWorkSpace workSpace; 
    if( params.computeGlobalInverse )
    {
      // first get the initial guess
      
      approximateGlobalInverse->inverse( x,r,rx,workSpace,params );

      // Now do Newton to Invert:
      exactLocalInverse->inverse( x,r,rx,workSpace,TRUE );   // TRUE means use results found in the
                                                             // workSpace
    }
    else
    {
      exactLocalInverse->inverse( x,r,rx,workSpace,FALSE );
    }
  }
  else
  {

    // ***** this Mapping used a distributed grid for the inverse *****

    if( debug >0 )
      openDebugFiles();
    

    MappingWorkSpace workSpace; 
    if( params.computeGlobalInverse )
    {
      approximateGlobalInverse->initialize(); // Warning: this will change the class variables base,bound
      
      int base,bound,computeMap=false,computeMapDerivative=false;
      Index I = getIndex( x,r,rx,base,bound,computeMap,computeMapDerivative );

      const MPI_Comm & OV_COMM = Overture::OV_COMM;
  
      const int np= max(1,Communication_Manager::numberOfProcessors());
      const int myid=max(0,Communication_Manager::My_Process_Number);

      // for each pt x(i,.) find the processors whose local grid may hold the closest point

      real * rp = r.Array_Descriptor.Array_View_Pointer1;
      const int rDim0=r.getRawDataSize(0);
#undef R
#define R(i0,i1) rp[i0+rDim0*(i1)]
#undef X
      const real * xp = x.Array_Descriptor.Array_View_Pointer1;
      const int xDim0=x.getRawDataSize(0);
#define X(i0,i1) xp[i0+xDim0*(i1)]
      real * rxp = rx.Array_Descriptor.Array_View_Pointer2;
      const int rxDim0=rx.getRawDataSize(0);
      const int rxDim1=rx.getRawDataSize(1);
#undef RX
#define RX(i0,i1,i2) rxp[i0+rxDim0*(i1+rxDim1*(i2))]

      if( debug & 1 )
      {
        fprintf(pDebugFile,"mapInverse: myid=%i base=%i bound=%i computeMap=%i computeDer=%i name=%s \n",
             myid,base,bound,computeMap,computeMapDerivative,
             (const char*)getName(mappingName));
      }
      

      int numPoints=bound-base+1;

      int *nqs = new int [np];  // nqs[p] : number of queries to send to processor p
      int *nqr = new int [np];  // nqs[p] : number of queries received from processor p
      for( int p=0; p<np; p++ ) nqs[p]=0;
    
      // ip(j,p) = i, j=0,..,nqs[p] : means x(i,.) is the j'th point sent to processor p 
      int *pip=NULL;
      if( numPoints>0 )
         pip = new int [numPoints*np];    // ********* this is too large in general, use a SparseArray ? 
#define ip(j,p) (pip[(p)+np*(j)])


      const RealArray & globalBoundingBox = getBoundingBox();
      const real extensionFactor=approximateGlobalInverse->stencilWalkBoundingBoxExtensionFactor;

      BoundingBox *& serialBoundingBox = approximateGlobalInverse->serialBoundingBox;
      assert( serialBoundingBox !=NULL );
      // IndexBox pBox;
      for( int p=0; p<np; p++ )
      {
        // Extend the bounding box for this processor in the same manner as findNearestGridPoint

        // CopyArray::getLocalArrayBox( p,grid,pBox );  // pBox : dimensions of gridSerial on processor p 

        real pbb[6]={0.,0.,0.,0.,0.,0.};
        #define bb(side,axis) pbb[(side)+2*((axis))]
	for( int axis=0; axis<rangeDimension; axis++ )
	{
	  for( int side=0; side<=1; side++ )
	  {
	    bb(side,axis)=serialBoundingBox[p].rangeBound(side,axis);
	  }
          real delta =extensionFactor*(globalBoundingBox(1,axis)-globalBoundingBox(0,axis));
          bb(0,axis)-=delta;
          bb(1,axis)+=delta;

//           real eps=REAL_EPSILON*10.*(globalBoundingBox(1,axis)-globalBoundingBox(0,axis));

//           // only increase the bounding box on true boundaries (not internal parallel boundaries )
//           // for parallel boundaries -- just increase/decrease by a very small amount (eps) to 
//           // make sure there is a positive overlap between the boxes.
// 	  if( pBox.base(axis)==gridIndexRange(0,axis) )
//             bb(0,axis)-=delta;
//           else
//             bb(0,axis)-=eps;
//           if( pBox.bound(axis)==gridIndexRange(1,axis) )
//   	    bb(1,axis)+=delta;
//           else
//             bb(1,axis)+=eps;
	}
	if( debug & 1 )
	  fprintf(pDebugFile," myid=%i p=%i extended-boundingBox = [%g,%g][%g,%g][%g,%g]\n",myid,p,
		 bb(0,0),bb(1,0),bb(0,1),bb(1,1),bb(0,2),bb(1,2));
	

	for( int i=base; i<=bound; i++ )
	{
	  // if x(i,.) is inside the bounding box for processor p 
          bool inside = true;
	  for( int axis=0; axis<rangeDimension; axis++ )
	  {
	    inside = inside && X(i,axis)>=bb(0,axis) && X(i,axis)<=bb(1,axis);
	    if( !inside ) break;
	  }
	  if( inside ) // x(i,.) is inside box[p] )
	  {
	    ip(nqs[p],p)=i;  // this point should be sent to processor p for an answer 
	    nqs[p]++;
	  }
	}
        if( debug & 1 ) fprintf(pDebugFile,"mapInverse: myid=%i send nqs[p]=%i queries to p=%i\n",myid,nqs[p],p);
      }

      // --- send/receive number of queries ---
      // we need to send computeMap and computeMapDerivative since these may not be the
      // same on all processors!

      int myInfo[3]={0,computeMap,computeMapDerivative}; //
      // info(0:2,p) : holds (nqr[p],computeMap[p],computeMapDerivative[p])
      int *pinfo = new int [3*np];
      #define info(i,p) pinfo[(i)+3*(p)]

      MPI_Status status;
      const int tag0=309278;
      int numberOfQueries=0;      // total number of queries that we will receieve 
      for( int p=0; p<np; p++ )
      {
        // *** send computeMap computeMapDerivative here too !

	int tags=tag0+p, tagr=tag0+myid;
        myInfo[0]=nqs[p];
	MPI_Sendrecv(myInfo,     3, MPI_INT, p, tags, 
		     &info(0,p), 3, MPI_INT, p, tagr, OV_COMM, &status ); 
// 	MPI_Sendrecv(&nqs[p], 1, MPI_INT, p, tags, 
// 		     &nqr[p], 1, MPI_INT, p, tagr, OV_COMM, &status ); 

        nqr[p]=info(0,p);
	
	if( debug & 1 ) 
	{
	  fprintf(pDebugFile,"mapInverse: myid=%i send %i queries to p=%i (map=%i,der=%i), "
                             "receive %i queries from p=%i, p: (map=%i der=%i)\n",
		 myid,nqs[p],p,computeMap,computeMapDerivative, nqr[p],p,info(1,p),info(2,p));
	  fflush(pDebugFile);
	}
	
	numberOfQueries += nqr[p];
      }

      // For now -- we compute r and rx for the queries if there is any processor wants these results,
      // but we do not return rx if the requesting processor does noy want them.

      int computeMapAnswer=false, computeMapDerivativeAnswer=false;
      for( int p=0; p<np; p++ )
      {
	computeMapAnswer = computeMapAnswer || info(1,p);
	computeMapDerivativeAnswer = computeMapDerivativeAnswer || info(2,p);
      }

//       computeMapAnswer=true;
//       computeMapDerivativeAnswer=true;
      

      // ******* BUFFERS ************
      //  sqb[p] : send queries buffer -- send queries to processor p
      //  rqb[p] : receive queries buffer
      //
      //  sqb[p] : send answers buffer -- send answers to processor p
      //  rqb[p] : receieve answers buffer 

      const int numDataToSendPerPoint=rangeDimension+domainDimension;  // we send x and r 

      // --- Allocate receive buffers ---
      real **rqb = new real* [np];
      for( int p=0; p<np; p++ )
      {
	rqb[p] = new real [max(1,nqr[p]*numDataToSendPerPoint)];
      }
  
      // --- post receives ---
      MPI_Request *receiveRequest     = new MPI_Request[np];  
      MPI_Request *sendRequest        = new MPI_Request[np];  
      MPI_Request *sendRequestAnswers = new MPI_Request[np];  
      const int tag1=2529743;
      for( int p=0; p<np; p++ )
      {  
	int tag=tag1+myid;
	MPI_Irecv( rqb[p], nqr[p]*numDataToSendPerPoint, MPI_Real, p, tag, OV_COMM, &receiveRequest[p] );
      }

      // -- allocate send buffers ---

      // pxs[p](i,m) : holds x and r values m=0,1,...,rangeDimension+domainDimension-1
      real **sqb = new real* [np];
      for( int p=0; p<np; p++ )
      {
	sqb[p] = new real [max(1,nqs[p]*numDataToSendPerPoint)];
      }

      // --- fill in send buffers ---
#define xs(i,n,p) sqb[p][(i)+nqs[p]*(n)]
#define rs(i,m,p) sqb[p][(i)+nqs[p]*(m+rangeDimension)]
      for( int p=0; p<np; p++ )
      {
	for( int j=0; j<nqs[p]; j++ )
	{   
	  int i = ip(j,p);
	  for( int n=0; n<rangeDimension; n++ ) xs(j,n,p) = X(i,n); 
          if( computeMap )
  	    for( int m=0; m<domainDimension; m++ ) rs(j,m,p) = R(i,m);
	}
      }
    
      // --- send buffers ---
      for( int p=0; p<np; p++ )
      {
	if( debug & 1 ) 
	{
	  fprintf(pDebugFile,"mapInverse: Send %i queries from myid=%i to p=%i\n",nqs[p],myid,p);
	  fflush(pDebugFile);
	}
	
	int tag=tag1+p;
	MPI_Isend(sqb[p], nqs[p]*numDataToSendPerPoint, MPI_Real, p, tag, OV_COMM, &sendRequest[p] );  
      }
  
      // --- wait for all the receives to finish ---
      MPI_Status *receiveStatus= new MPI_Status[np];  
      MPI_Waitall(np,receiveRequest,receiveStatus);


      if( true || debug & 4 )
      { // sanity check:

	fflush(0);
	Communication_Manager::Sync();

	for( int p=0; p<np; p++ )
	{
	  int num=0;
	  MPI_Get_count( &receiveStatus[p], MPI_Real, &num );
	  assert( num==nqr[p]*numDataToSendPerPoint );
	}
      }


      // --- post receives for answers (which have not even been computed yet) ---
      //     return r and maybe rx values 
      int *numDataToReceivePerPoint = new int[np];
      real **rab = new real* [np];  // rab = "answers" 
      for( int p=0; p<np; p++ )
      {
        // return r, and optionally rx values

        // something is wrong here (c.f. stir.cmd)
// ***        numDataToReceivePerPoint[p] =domainDimension + info(2,p)*domainDimension*rangeDimension;  

        numDataToReceivePerPoint[p] =domainDimension + domainDimension*rangeDimension;  
        // we receive back the number we sent: nqs[p] 
	rab[p] = new real [max(1,nqs[p]*numDataToReceivePerPoint[p])];
      }

      const int tag2=1329743;
      for( int p=0; p<np; p++ )
      {  
	int tag=tag2+myid;
        if( debug & 1 )
        {
          fprintf(pDebugFile,"inverseMap: myid=%i, expect %i answers from p=%i\n",myid,nqs[p],p);
	  fflush(pDebugFile);
	}
	
	MPI_Irecv( rab[p], nqs[p]*numDataToReceivePerPoint[p], MPI_Real, p, tag, OV_COMM, &receiveRequest[p] );
      }

      // ------------------------------------
      // --- un-pack the values received ----
      // ------------------------------------

      RealArray xc,rc,rxc;
      if( numberOfQueries>0 )
      {
	xc.redim(numberOfQueries,rangeDimension);
	rc.redim(numberOfQueries,domainDimension);

	if( computeMapDerivativeAnswer )
          rxc.redim(numberOfQueries,domainDimension,rangeDimension); 
      }

      real * rcp = rc.Array_Descriptor.Array_View_Pointer1;
      const int rcDim0=rc.getRawDataSize(0);
#undef RC
#define RC(i0,i1) rcp[i0+rcDim0*(i1)]
#undef XC
      real * xcp = xc.Array_Descriptor.Array_View_Pointer1;
      const int xcDim0=xc.getRawDataSize(0);
#define XC(i0,i1) xcp[i0+xcDim0*(i1)]
      real * rxcp = rxc.Array_Descriptor.Array_View_Pointer2;
      const int rxcDim0=rxc.getRawDataSize(0);
      const int rxcDim1=rxc.getRawDataSize(1);
#undef RXC
#define RXC(i0,i1,i2) rxcp[i0+rxcDim0*(i1+rxcDim1*(i2))]


#define xrb(i,n,p) rqb[p][(i)+nqr[p]*(n)]
#define rrb(i,m,p) rqb[p][(i)+nqr[p]*(m+rangeDimension)]
      int i=0;
      for( int p=0; p<np; p++ )
      {
	for( int j=0; j<nqr[p]; j++ )
	{
	  for( int n=0; n<rangeDimension; n++ )  XC(i,n) = xrb(j,n,p); 
	  for( int m=0; m<domainDimension; m++ ) RC(i,m) = rrb(j,m,p);
  
	  i++;
	}
      }
      assert( i==numberOfQueries );

      // we can delete rqb[p]  now 
      for( int p=0; p<np; p++ )
	delete [] rqb[p];
      delete [] rqb;

      if( debug & 2 )
      {
	for( int i=0; i<numberOfQueries; i++ )
	{
	  fprintf(pDebugFile,"inverseMap: myid=%i check point i=%i : xc=(%g,%g,%g), rc=(%g,%g,%g)\n",myid,i,
		 XC(i,0),(rangeDimension>1 ? XC(i,1) : 0.), (rangeDimension>2 ? XC(i,2) : 0.),
		 RC(i,0),(domainDimension>1 ? RC(i,1) : 0.), (domainDimension>2 ? RC(i,2) : 0.));

	}
        fflush(pDebugFile);
	Communication_Manager::Sync();
	MPI_Barrier(OV_COMM);
	
      }
	

      // *********************** ACTUAL INVERSE IS HERE *****************************

      // find the closest point
      if( numberOfQueries>0 )
        approximateGlobalInverse->inverse( xc,rc,rxc,workSpace,params );

      // When the exact local inverse is not a distributed operation we could reduce communication 
      // at this point by sending back just the closest points r -- then doing the exact local inverse
      // at the other processor 

      // For now we invert here -- this will handle the case of the DataPointMapping when the exact local
      // inverse uses the same distributed grid as the approximate global inverse

      // Now do Newton to Invert:
      if( numberOfQueries>0 || usesDistributedMap() ) 
        exactLocalInverse->inverse( xc,rc,rxc,workSpace,true );




      if( debug & 1 )
      {
	for( int i=0; i<numberOfQueries; i++ )
	{
	  fprintf(pDebugFile,"inverseMap: myid=%i after exactInverse, point i=%i : xc=(%g,%g,%g), rc=(%g,%g,%g)\n",myid,i,
		 XC(i,0),(rangeDimension>1 ? XC(i,1) : 0.), (rangeDimension>2 ? XC(i,2) : 0.),
		 RC(i,0),(domainDimension>1 ? RC(i,1) : 0.), (domainDimension>2 ? RC(i,2) : 0.));
	}
        fflush(pDebugFile);
      }


      // -----------------------------------------------
      // --- Send answers back : pack into a buffer  ---
      // -----------------------------------------------

      real **sab = new real* [np];  // sab = send-answers buffer
      for( int p=0; p<np; p++ )
      {
        // we send back the number we received : nqr[p]
	sab[p] = new real [max(1,nqr[p]*numDataToReceivePerPoint[p])];
      }

#undef rrb
#define rrb(i,m,p) sab[p][(i)+nqr[p]*(m)]
#define rxb(i,m,n,p) sab[p][(i)+nqr[p]*((m)+domainDimension*(n+1))]
      i=0;
      for( int p=0; p<np; p++ )
      {
	for( int j=0; j<nqr[p]; j++ )
	{
	  for( int m=0; m<domainDimension; m++ ) rrb(j,m,p) = RC(i,m);
	  if( info(2,p) ) // true if we are to computeMapDerivative for processor p
	  {
	    for( int m=0; m<domainDimension; m++ )
	    {
	      for( int n=0; n<rangeDimension; n++ )  
	      {
		rxb(j,m,n,p) = RXC(i,m,n);
	      }
	    }
	  }
	  
	  i++;
	}
      }

      // --------------------
      // --- Send buffers ---
      // --------------------

      for( int p=0; p<np; p++ )
      {
	int tag=tag2+p;
	if( debug & 1 )
        {
          fprintf(pDebugFile,"inverseMap: myid=%i, send %i answers to p=%i\n",myid,nqr[p],p);
	  fflush(pDebugFile);
	}
	
	MPI_Isend(sab[p], nqr[p]*numDataToReceivePerPoint[p], MPI_Real, p, tag, OV_COMM, &sendRequestAnswers[p] );  
      }

      if( debug & 4 )
	MPI_Barrier(OV_COMM);
      
      // --- wait for all the receives to finish ---
      MPI_Waitall(np,receiveRequest,receiveStatus);

      if( true || debug & 4 )
      { // sanity check:
	for( int p=0; p<np; p++ )
	{
	  int num=0;
	  MPI_Get_count( &receiveStatus[p], MPI_Real, &num );
	  assert( num==nqs[p]*numDataToReceivePerPoint[p] );
	}
      }

      // --- assign results back into the r,rx arrays ---

#undef rrb
#define rrb(i,m,p) rab[p][(i)+nqs[p]*(m)]
#undef rxb 
#define rxb(i,m,n,p) rab[p][(i)+nqs[p]*((m)+domainDimension*(n+1))]

      bool *passigned = new bool [max(1,numPoints)];
#define assigned(i) passigned[(i)-base]
      for( int i=base; i<=bound; i++ )
	assigned(i)=false;
    
      real rv[3];
      for( int p=0; p<np; p++ )
      {
	for( int j=0; j<nqs[p]; j++ )
	{
	  i = ip(j,p);  // the result goes here 
	  assert( i>=base && i<=bound );
	  
	  if( assigned(i) ) 
	  {
	    // this points has already been assigned, we should choose the best inverse

	    // For now: check the distance to the centre of the unit square and choose the
	    //   point with the least distance
	    for( int m=0; m<domainDimension; m++ ) rv[m] = rrb(j,m,p);

	    real dist=0., distNew=0.;
	    for( int m=0; m<domainDimension; m++ )
	    {
	      dist += fabs(r(i,m)-.5);
	      distNew+= fabs(rv[m]-.5);
	    }
	    if( distNew<dist )
	    {
	      assigned(i)=false;  // this will cause the values to be assigned below
	    }
	  }
	  if( !assigned(i) )
	  {
	    assigned(i)=true;
	    if( computeMap )
	    {
	      for( int m=0; m<domainDimension; m++ ) R(i,m) = rrb(j,m,p);

	      if( debug & 1 )
	      {
		fprintf(pDebugFile,"inverseMap: myid=%i RESULTS for point i=%i : x=(%g,%g,%g), r=(%g,%g,%g)\n",myid,i,
		       X(i,0),(rangeDimension>1 ? X(i,1) : 0.), (rangeDimension>2 ? X(i,2) : 0.),
		       R(i,0),(domainDimension>1 ? R(i,1) : 0.), (domainDimension>2 ? R(i,2) : 0.));
	      }
	    }
	  
	    if( computeMapDerivative )
	    {
	      for( int m=0; m<domainDimension; m++ )
	      {
		for( int n=0; n<rangeDimension; n++ )  
		{
		  RX(i,m,n) = rxb(j,m,n,p);
		}
	      }
	    }
	  
	  }
	
	  i++;
	}
      }

      // any points not assigned but have been outside the bounding box. *wdh* 081002
      for( int i=base; i<=bound; i++ )
      {
	if( !assigned(i) )
	{
	  if( computeMap )
	  {
	    for( int m=0; m<domainDimension; m++ ) R(i,m) = Mapping::bogus;
	  }
	}
      }
      

      if( debug>0 )
      {
	fprintf(pDebugFile,"inverseMap:DONE: myid=%i\n",myid);
	fflush(pDebugFile);
      }
      
      // wait for ALL sends to finish on this processor before we can clean up
      MPI_Waitall(np,sendRequest,receiveStatus);
      MPI_Waitall(np,sendRequestAnswers,receiveStatus);

      delete [] pinfo;
      delete [] numDataToReceivePerPoint;
      delete [] nqs;
      delete [] nqr;
      delete [] pip;
      delete [] receiveRequest;
      delete [] receiveStatus;
      delete [] sendRequest;
      delete [] sendRequestAnswers;
      for( int p=0; p<np; p++ )
      {
	delete [] sqb[p];
	delete [] sab[p];
	delete [] rab[p];
      }
      delete [] sqb;
      delete [] sab;
      delete [] rab;
    
      delete [] passigned;
      
    }
    else
    {
      exactLocalInverse->inverse( x,r,rx,workSpace,FALSE );
    }
    
  }
  
#endif

}

