#include "Ogmg.h"
#include "TridiagonalSolver.h"
#include "ParallelUtility.h"
#include "HDF_DataBase.h"
#include "App.h"

#undef ForBoundary
#define ForBoundary(side,axis)   for( axis=0; axis<mgcg[0].numberOfDimensions(); axis++ ) \
                                 for( side=0; side<=1; side++ )

static bool applyFinishBoundaryConditions=false;
  

int Ogmg::debug=0;
FILE *Ogmg::debugFile = NULL;
FILE *Ogmg::infoFile  = NULL;
FILE *Ogmg::checkFile = NULL;
aString Ogmg::infoFileCaption[5]; // for caption on the table
real Ogmg::bogusRealArgument1=0.;
real Ogmg::bogusRealArgument2=0.;


OGFunction *Ogmg::pExactSolution=NULL;


//\begin{>OgmgInclude.tex}{\subsection{constructor}}
Ogmg::
Ogmg()
//==================================================================================
// /Description:
//   Default constructor.
//\end{OgmgInclude.tex} 
//==================================================================================
{
  initialized=FALSE;  // set to TRUE once setup is called
  init();
}

//\begin{>>OgmgInclude.tex}{\subsection{constructor}}
Ogmg::
Ogmg( CompositeGrid & mg, 
      GenericGraphicsInterface *ps_ /* = 0 */ ) 
: parameters(mg)
//==================================================================================
// /Description:
//   
//    Build a multigrid solver.
// /mg (input) : grid to use
// /ps\_ (input) : supply an optional GenericGraphicsInterface object for plotting.
// /Notes: 
//    Here are some notes. See the Ogmg User Guide for further details.
//  \begin{description}
//    \item[Boundary Conditions]: Ogmg looks at the coefficient matrix to determine if the ghost line
//     values are extrapolated or not. If not it assumes there is some sort of neumann or mixed boundary
//     conditions and does a few things differently.
//    \item[Interpolants]  The multigrid solver needs to interpolate at the different levels.
//      If no Interpolant is found to be associated with a CompositeGrid at a given level
//      the {\tt updateToMatchGrid} routine will build an Interpolant the first time (and update it
//      on subsequent calls). Thus if you have built an Interpolant for any level then you are responsible
//      to update it if the grid changes.
//  \end{description}
//\end{OgmgInclude.tex} 
//==================================================================================
{
  initialized=FALSE;  // set to TRUE once setup is called
  init();
  
  ps=ps_;

  updateToMatchGrid(mg);
}


void Ogmg::
init()
// =================================================================================================
// /Description: Initialize variables -- this is a common routine for all constructors
// =================================================================================================
{  
  ps=NULL;

  myid=max(0,Communication_Manager::My_Process_Number);

  orderOfAccuracy=2;
  coarseToFineInterpolationWidth=2;  
  useForcingAsBoundaryConditionOnAllLevels=false;
  
  operatorsForExtraLevels=NULL;
  interpolant=NULL;
  
  numberOfExtraLevels=0;
  assumeSparseStencilForRectangularGrids=true;
  levelZero=0;
  iterationCount=NULL;
  
  v=NULL;
  leftNullVector=NULL;
  leftNullVectorIsComputed=false;
  equationToSolve=OgesParameters::userDefined;
  varCoeff=NULL;
  // for timings
  for( int i=0; i<numberOfThingsToTime; i++ )
    tm[i]=0.;
  
  const int np= max(1,Communication_Manager::numberOfProcessors());
  // Open debug files
  aString name="ogmg";
  char buff[180];
  #ifndef USE_PPP
    debugFile = fopen(sPrintF(buff,"%s.debug",(const char*)name),"w" );      // Here is the debug file
    pDebugFile=debugFile;
    if( infoFile==NULL )
      infoFile  = fopen(sPrintF(buff,"%s.info",(const char*)name),"w" );       // for convergence info
  #else
    debugFile = fopen(sPrintF(buff,"%sNP%i.debug",(const char*)name,np),"w" );      // Here is the debug file
    pDebugFile= fopen(sPrintF(buff,"%s%i.debug",(const char*)name,myid),"w");
    if( infoFile==NULL )
      infoFile  = fopen(sPrintF(buff,"%sNP%i.info",(const char*)name,np),"w" );       // for convergence info
  #endif


  if( checkFile==NULL )
    checkFile = fopen("ogmg.check","w" );      // for regression tests.

  infoFileCaption[0]="";
  infoFileCaption[1]="";
  infoFileCaption[2]="";
  infoFileCaption[3]="";
  infoFileCaption[4]="";

  ogesSmoother=NULL;
  
  nipn=NULL, ndipn=NULL, ipn=NULL;  // for IBS -- could be shared amongst solvers with the same CompositeGrid.
  numberOfIBSArrays=0;
}



Ogmg::
~Ogmg()
{
  if( initialized )
  {
    delete [] operatorsForExtraLevels;

    // delete all the pointers to the tri-diagonal solvers.
    for( int level=0; level<mgcg.numberOfMultigridLevels(); level++ )
    {
      for( int grid=0; grid<mgcg.multigridLevel[level].numberOfComponentGrids(); grid++ )
      {
	for( int dir=0; dir<3; dir++ )
	  delete tridiagonalSolver[level][grid][dir];
      
	delete [] tridiagonalSolver[level][grid];
      }
    
      delete [] tridiagonalSolver[level];
    }
    delete [] tridiagonalSolver;

    delete [] interpolant;
  }

  delete v;
  delete leftNullVector;
  delete varCoeff;
  
  delete [] ogesSmoother;

  // delete arrays used for IBS smoothing
  if( ipn!=NULL )
  {
    for( int i=0; i<numberOfIBSArrays; i++ )
      delete [] ipn[i];

    delete [] ipn;
  }
  delete [] nipn;
  delete [] ndipn;

  if( false )
  { // this fails if multiple Ogmg objects are created at the same time.
    if( Communication_Manager::numberOfProcessors()>1 )
      fclose(pDebugFile);

    fclose(debugFile);
    fclose(infoFile);
    fclose(checkFile);
  }
  
}

//\begin{>>OgmgInclude.tex}{\subsection{setPlotStuff}}
void Ogmg::
set( GenericGraphicsInterface *ps_ )
//==================================================================================
// /Description:
//    Supply a GenericGraphicsInterface object to use for plotting.
// /ps\_ (input) : pointer to a GenericGraphicsInterface object.
//\end{OgmgInclude.tex} 
//==================================================================================
{
  ps=ps_;
}


//! Assign an option.
/*!
  /param option==assumeSparseStencilOnRectangularGrids : assume the operator is a 5-point operator (2D)
     or 7-point operator (3D) on a rectangular grid.
 */
int Ogmg::
setOption(OptionEnum option, bool trueOrFalse )
{
  switch (option)
  {
    case assumeSparseStencilOnRectangularGrids:
      assumeSparseStencilForRectangularGrids=trueOrFalse;
      break;
  default:
    cout << "Ogmg::setOption:ERROR: unknown option = " << option << endl;
    return 1;
  }
  return 0;
}





//\begin{>>OgmgInclude.tex}{\subsection{updateToMatchGrid}}
void Ogmg::
updateToMatchGrid( CompositeGrid & mg_ )
//==================================================================================
// /Description:
//    Update the solver to match this grid.
// /mg (input) : grid to use
//\end{OgmgInclude.tex} 
//==================================================================================
{
//    // for timings
//    for( int i=0; i<numberOfThingsToTime; i++ )
//    {
//      tm[i]=0.;
//    }

  real time0=getCPU();
  
  if( debug & 2 )
  {
    printF("\n ++++++++++++++ Entering Ogmg::updateToMatchGrid ++++++++++++++\n");
  }

  if( mg_.numberOfComponentGrids()>0 )
  { // choose the order of accuracy automatically
    orderOfAccuracy = mg_[0].discretizationWidth(0)==5 ? 4 : 2;
    if( orderOfAccuracy==4 )
      printF("Ogmg::updateToMatchGrid::INFO: Setting orderOfAccuracy=%i\n",orderOfAccuracy);
  }
  
  numberOfExtraLevels=0;
  if( parameters.readMultigridCompositeGrid )
  {
    // read the grid with extra levels from a file
    printF("Reading the multigrid composite grid with extra levels from file=%s\n",
           (const char*)parameters.nameOfMultigridCompositeGrid);

    getFromADataBase(mgcg,parameters.nameOfMultigridCompositeGrid);
    printF("After reading: mgcg.numberOfMultigridLevels=%i\n",mgcg.numberOfMultigridLevels());
    mgcg.update(MappedGrid::THEmask);
    numberOfExtraLevels=mgcg.numberOfMultigridLevels()-1;

//     HDF_DataBase dataFile;
//     dataFile.mount(nameOfMultigridCompositeGrid,"R");
//     mgcg.get(dataFile,"mgcg");
//     dataFile.unmount();
  }
  else 
  {
    if( mg_.numberOfMultigridLevels()==1 )
    {
      // *******************************************
      // ***** Automatically build extra levels ****
      // *******************************************

      numberOfExtraLevels=1;
      buildExtraLevels(mg_);  // This builds the CompositeGrid mgcg


      if( true && parameters.readMultigridCompositeGrid )
      {
	// read the grid with extra levels from a file
	printF("Reading the multigrid composite grid with extra levels from file=%s\n",
	       (const char*)parameters.nameOfMultigridCompositeGrid);

        CompositeGrid mgcg2;
	getFromADataBase(mgcg2,parameters.nameOfMultigridCompositeGrid);
	printF("After reading: mgcg2.numberOfMultigridLevels=%i\n",mgcg2.numberOfMultigridLevels());
	mgcg2.update(MappedGrid::THEmask);

        assert( mgcg.numberOfGrids()==mgcg2.numberOfGrids() );
	assert( mgcg.numberOfMultigridLevels()==mgcg2.numberOfMultigridLevels() );

	assert( max(abs(mgcg.numberOfInterpolationPoints()-mgcg2.numberOfInterpolationPoints()))==0 );
	assert( max(abs(mgcg.interpolationStartEndIndex-mgcg2.interpolationStartEndIndex))==0 );
	display(mgcg.interpolationStartEndIndex,"mgcg.interpolationStartEndIndex");
	display(mgcg2.interpolationStartEndIndex,"mgcg2.interpolationStartEndIndex");
	
	assert( max(abs(mgcg.multigridCoarseningRatio-mgcg2.multigridCoarseningRatio))==0 );
	
        // mgcg2=mgcg;

	for( int l=0; l<mgcg.numberOfMultigridLevels(); l++ )
	{
          printF(" *** Check level=%i\n",l);
	  CompositeGrid & cg1 = mgcg.multigridLevel[l];
	  CompositeGrid & cg2 = mgcg2.multigridLevel[l];
	  
          assert( cg1.numberOfGrids()==cg2.numberOfGrids() );
          assert( cg1.numberOfBaseGrids()==cg2.numberOfBaseGrids() );
	  assert( max(abs(cg1.interpolationWidth()-cg2.interpolationWidth()))==0 );
	  assert( max(abs(cg1.interpolationIsImplicit()-cg2.interpolationIsImplicit()))==0 );
	  assert( abs(cg1.interpolationIsAllExplicit()-cg2.interpolationIsAllExplicit())==0 );
	  assert( abs(cg1.interpolationIsAllImplicit()-cg2.interpolationIsAllImplicit())==0 );
          display(cg1.interpolationStartEndIndex,"cg1.interpolationStartEndIndex");
          display(cg2.interpolationStartEndIndex,"cg2.interpolationStartEndIndex");
	  
	  assert( max(abs(cg1.interpolationStartEndIndex-cg2.interpolationStartEndIndex))==0 );
          // cg2.interpolationStartEndIndex=cg1.interpolationStartEndIndex;
	  
          assert( max(abs(cg1.numberOfInterpolationPoints()-cg2.numberOfInterpolationPoints()))==0 );
          ::display(cg1.numberOfInterpolationPoints(),"cg1.numberOfInterpolationPoints()");
	  ::display(cg2.numberOfInterpolationPoints(),"cg2.numberOfInterpolationPoints()");

          ::display(cg1.numberOfImplicitInterpolationPoints(),"cg1.numberOfImplicitInterpolationPoints()");
	  ::display(cg2.numberOfImplicitInterpolationPoints(),"cg2.numberOfImplicitInterpolationPoints()");
          // assert( max(abs(cg1.numberOfImplicitInterpolationPoints()-cg2.numberOfImplicitInterpolationPoints()))==0 );

	  for( int grid=0; grid<cg1.numberOfComponentGrids(); grid++ )
	  {
	    ::display(cg1.interpolationCoordinates[grid],"cg1.interpolationCoordinates[grid]","%4.1f ");
	    ::display(cg1.interpolationCoordinates[grid],"cg2.interpolationCoordinates[grid]","%4.1f ");
	    
	    if( cg1.numberOfInterpolationPoints(grid)>0 )
	    {
	      assert( max(abs(cg1[grid].discretizationWidth()-cg2[grid].discretizationWidth()))==0 );
	      assert( max(abs(cg1.interpolationPoint[grid]-cg2.interpolationPoint[grid]))==0 );
	      assert( max(abs(cg1.interpoleeGrid[grid]-cg2.interpoleeGrid[grid]))==0 );
	      assert( max(abs(cg1.variableInterpolationWidth[grid]-cg2.variableInterpolationWidth[grid]))==0 );
	      assert( max(fabs(cg1.interpolationCoordinates[grid]-cg2.interpolationCoordinates[grid]))==0 );
	      assert( max(abs(cg1[grid].mask()-cg2[grid].mask()))==0 );
	    }
	    
	  }
	}

	// if( true ) Overture::abort("done");
        mgcg=mgcg2;

      }

    }
    else
    {
      mgcg.reference(mg_);
      mgcg.update( CompositeGrid::THEmultigridLevel);
      mgcg.update(MappedGrid::THEmask);
      numberOfExtraLevels=mgcg.numberOfMultigridLevels()-1;
    }
    
  }
  if( parameters.saveMultigridCompositeGrid && !parameters.readMultigridCompositeGrid )
  {
    // save the grid with extra levels to a file
    printF("Saving the multigrid composite grid with extra levels to file=%s\n",
           (const char*)parameters.nameOfMultigridCompositeGrid);

    mgcg.saveGridToAFile(parameters.nameOfMultigridCompositeGrid,"mgcg");


    if( Ogmg::debug & 4 ) // check that we can read the MG back in
    {
      for( int grid=0; grid<mgcg.numberOfGrids(); grid++ )
      {
	::display(mgcg.interpolationPoint[grid],sPrintF("mgcg.interpolationPoint[%i]",grid));        
      }

      CompositeGrid mgcg2;
      printF("\nReading the multigrid composite grid with extra levels from file=%s\n",
	     (const char*)parameters.nameOfMultigridCompositeGrid);

      getFromADataBase(mgcg2,parameters.nameOfMultigridCompositeGrid);
      
      printF("\n >>>>After reading: mgcg2.numberOfMultigridLevels=%i\n",mgcg2.numberOfMultigridLevels());
      mgcg2.update(MappedGrid::THEmask);
      mgcg2.updateReferences();
      printF("<<< mgcg2.numberOfGrids()=%i, mgcg2.numberOfComponentGrids()=%i, "
	     "mgcg2.interpolationPoint.getLength=%i\n",
	     mgcg.numberOfGrids(),mgcg.numberOfComponentGrids(),mgcg.interpolationPoint.getLength());

      
      for( int grid=0; grid<mgcg2.numberOfGrids(); grid++ )
      {
//	::display(mgcg2.interpolationPoint[grid],sPrintF("mgcg2.interpolationPoint[%i]",grid));        
      } 

      ::display(mgcg.interpolationStartEndIndex(),"mgcg.interpolationStartEndIndex()");
      ::display(mgcg2.interpolationStartEndIndex(),"mgcg2.interpolationStartEndIndex()");

      for( int l=0; l<mgcg2.numberOfMultigridLevels(); l++ )
      {
	CompositeGrid & cgl = mgcg2.multigridLevel[l];
        // cgl.updateReferences();

	printF("*****level=%i cgl.numberOfComponentGrids()=%i, "
	       "cgl.interpolationPoint.getLength=%i\n",
	       l,cgl.numberOfComponentGrids(),cgl.interpolationPoint.getLength());

        ::display(mgcg.multigridLevel[l].interpolationStartEndIndex(),"mgcg[l].interpolationStartEndIndex()");
        ::display(cgl.interpolationStartEndIndex(),"cgl.interpolationStartEndIndex()");

	::display(cgl.numberOfInterpolationPoints(),"cgl.numberOfInterpolationPoints()");
	for( int grid=0; grid<cgl.numberOfComponentGrids(); grid++ )
	{
// 	  ::display(cgl.interpoleeGrid[grid],"cgl.interpoleeGrid[grid]");
// 	  ::display(cgl.variableInterpolationWidth[grid],"cgl.variableInterpolationWidth[grid]");
//	  ::display(cgl.interpolationPoint[grid],"cgl.interpolationPoint[grid]");        
// 	  ::display(cgl.interpoleeLocation[grid],"cgl.interpoleeLocation[grid]");        
// 	  ::display(cgl.interpolationCoordinates[grid],"cgl.interpolationCoordinates[grid]");  
	}
      }
      
    }
    
  }
  

  CompositeGrid & mg =  mgcg; 

  // displaySmoothers("Ogmg::updateToMatchGrid(cg) before parameters.updateToMatchGrid(mg)");

  parameters.updateToMatchGrid(mg);

  // displaySmoothers("Ogmg::updateToMatchGrid(cg) after parameters.updateToMatchGrid(mg)");

  if( !initialized )
    setup(mg);
  
  bool numberOfMultigridLevelsHasChanged= 
                          mgcg.numberOfMultigridLevels() > 0 &&
                          mg.numberOfMultigridLevels() > 0   &&
           (mgcg.numberOfMultigridLevels()!=mg.numberOfMultigridLevels());
  
  bool numberOfComponentGridsHasChanged = 
                          mgcg.numberOfMultigridLevels() > 0 &&
                          mg.numberOfMultigridLevels() > 0   &&
                     (mgcg.numberOfComponentGrids()!=mg.numberOfComponentGrids());

  if( numberOfMultigridLevelsHasChanged )
  {
    interpolantWasCreated.resize(mg.numberOfMultigridLevels());
    if( mg.numberOfMultigridLevels()>mgcg.numberOfMultigridLevels() )
    {
      Range R(mgcg.numberOfMultigridLevels(),mg.numberOfMultigridLevels()-1);
      interpolantWasCreated(R)=FALSE;
    }
    
    workUnits.redim(mg.numberOfMultigridLevels());
    workUnits=0.;       

    // number of times to iterate on each level (=1 -> V-cycle, 2=W-cycle)
//     numberOfIterations.resize(mg.numberOfMultigridLevels());
//     numberOfIterations(R)=1;

  }
  if( numberOfMultigridLevelsHasChanged || numberOfComponentGridsHasChanged )
  {
    lineSmoothIsInitialized.resize(3,mg.numberOfComponentGrids(),mg.numberOfMultigridLevels());

    // create arrays of pointers for the tridiagonal solvers   ****** delete these *****
    //      tridiagonalSolver[level][grid][axis]
    typedef TridiagonalSolver* TRIP;
    typedef TRIP* TRIPP;
    typedef TRIPP* TRIPPP;
    typedef TRIPPP* TRIPPPP;
  
    tridiagonalSolver = new TRIPPP[mg.numberOfMultigridLevels()];
    for( int level=0; level<mgcg.numberOfMultigridLevels(); level++ )
    {
      tridiagonalSolver[level] = new TRIPP[mg.numberOfComponentGrids()];
      for( int grid=0; grid<mg.numberOfComponentGrids(); grid++ )
	tridiagonalSolver[level][grid]= new TRIP[3];
    }

  }  

  lineSmoothIsInitialized=false;  // assume that all grids have changed

  mgcg.reference(mg);


  defectMG.updateToMatchGrid(mgcg);  // allocate space for the defect

  // Holds defect ratios for auto-subSmooth
  defectRatio.redim(mgcg.numberOfComponentGrids(),mgcg.numberOfMultigridLevels());  
  defectRatio=-1.;  // This means the ratios have not been computed yet
  
// *** move   uOld.updateToMatchGrid(mgcg.multigridLevel[0]);  // used in solve **** why is this in the class ? ****
  
  // Make sure there is an interpolant at each level (except the last if we use a Oges)
  const int mgLevels = mgcg.numberOfMultigridLevels();
  delete [] interpolant;
  interpolant = new Interpolant[mgLevels];
  for( int level=0; level<mgLevels; level++ )
  {
    if( mgcg.multigridLevel[level]->interpolant==NULL || interpolantWasCreated(level) )
    {
      if( debug & 4 )
        printF("***********Ogmg: creating the Interpolant at level=%i \n",level);
      interpolantWasCreated(level)=TRUE;
      interpolant[level].updateToMatchGrid(mgcg.multigridLevel[level]);    // object used for interpolation
      // display(mgcg.interpolationPoint[0],"mgcg.interpolationPoint[0]");
      // display(mgcg.multigridLevel[level].interpolationPoint[0],"mgcg.multigridLevel[level].interpolationPoint[0]");
      
      // 40904: bug found with interpolate -- opt implicit iteration was not use the variableInterpolationWidth
      //   and caused bad values to appear. -- fixed made to gf/interpOpt.bf etc. 
      if( true ) // ***********************
        interpolant[level].setImplicitInterpolationMethod(Interpolant::iterateToInterpolate);
    }
  }
  
  // count the total number of grid points.
  numberOfGridPoints=0;
  for( int grid=0; grid<mgcg.numberOfComponentGrids(); grid++ )
    numberOfGridPoints+=mgcg[grid].mask().elementCount();
  
  if( mgcg.numberOfMultigridLevels()==1 && !parameters.useDirectSolverForOneLevel )
  {
    // If there is only one level then iterate with a line smoother
    parameters.useDirectSolverOnCoarseGrid=false;
    parameters.setSmootherType(OgmgParameters::alternatingLineJacobi);
    // parameters.setSmootherType(OgmgParameters::redBlack);
  }
  
  // defectMG.updateToMatchGrid(mgcg);  // allocate space for the defect

  InterpolateParameters interpParams(mgcg.numberOfDimensions());
//  int coarseToFineInterpolationWidth=2;  // ***********
  interpParams.setInterpolateOrder(coarseToFineInterpolationWidth);
  interp.initialize(interpParams);  


  active.redim(mgcg.numberOfComponentGrids());
  active=true;  // set active(grid)=false if we do not need to solve for the solution on this grid.

  tm[timeForInitialize]+=getCPU()-time0;
}


void Ogmg::
setup(CompositeGrid & mg )
// ================================================================================================
// /Description:
//   Initialize Ogmg for the first time a CompositeGrid is available. This routine must only be called
//  once.
// ================================================================================================
{
  assert( !initialized );
  initialized=TRUE;
  
  // Here are guesses at the relative times to perform operations
  timeForAddition=1.;
  timeForMultiplication=1.;
  timeForDivision=4.;
  // workUnit = number of operations for a Jacobi Smooth
  if( mg.numberOfDimensions()==2 )
    workUnit=10*timeForMultiplication+9*timeForAddition+timeForDivision;    // non-rectangular
  else
    workUnit=28*timeForMultiplication+27*timeForAddition+timeForDivision;


  numberOfIterations=0;
  numberOfCycles=0;
  
  workUnits.redim(mg.numberOfMultigridLevels());
  workUnits=0.;       

  width1=orderOfAccuracy+1;
  width2 = mg.numberOfDimensions()>1 ? 3 : 1;
  width3 = mg.numberOfDimensions()>2 ? 3 : 1;
  halfWidth1=width1/2;
  halfWidth2=width2/2;
  halfWidth3=width3/2;

  lineSmoothIsInitialized.redim(3,mg.multigridLevel[0].numberOfComponentGrids(),mg.numberOfMultigridLevels());
  lineSmoothIsInitialized=FALSE;

  // we keep track if we have built Interpolants so that we can update them in updateToMatchGrid
  interpolantWasCreated.redim(mg.numberOfMultigridLevels());
  interpolantWasCreated=FALSE;
  interpolant=NULL;
  
  // create arrays of pointers for the tridiagonal solvers   ****** delete these *****
  //      tridiagonalSolver[level][grid][axis]
  typedef TridiagonalSolver* TRIP;
  typedef TRIP* TRIPP;
  typedef TRIPP* TRIPPP;
  typedef TRIPPP* TRIPPPP;
  
  tridiagonalSolver = new TRIPPP[mg.numberOfMultigridLevels()];
  for( int level=0; level<mg.numberOfMultigridLevels(); level++ )
  {
    tridiagonalSolver[level] = new TRIPP[(int&)mg.multigridLevel[level].numberOfComponentGrids()];
    for( int grid=0; grid<mg.multigridLevel[level].numberOfComponentGrids(); grid++ )
    {
      tridiagonalSolver[level][grid]= new TRIP[3];
      for( int dir=0; dir<3; dir++ )
        tridiagonalSolver[level][grid][dir]=NULL;
    }
    
  }

  if( mg.numberOfDimensions()==3 )
  {
    // In 3D choose an iterative solver for the coarse grid 
    printF("Ogmg::set(OgesParameters::THEbestIterativeSolver)\n");
    
    parameters.ogesParameters.set(OgesParameters::THEbestIterativeSolver);

    real relativeTolerance=1.e-3;  // reduce residual by this amount.
    parameters.ogesParameters.set(OgesParameters::THErelativeTolerance,relativeTolerance);
    
  }

}

//\begin{>>OgmgInclude.tex}{\subsection{getMaximumResidual}} 
real Ogmg::
getMaximumResidual() const
//===================================================================================
// /Description: 
//   Return the maximum residual from the last solve
//\end{OgmgInclude.tex}
//===================================================================================
{
  return maximumResidual;
}

//\begin{>>OgmgInclude.tex}{\subsection{getNumberOfIterations}} 
int Ogmg::
getNumberOfIterations() const
//===================================================================================
// /Description: 
//   Return the number of multigrid iterations (cycles).
// /Return value: the number of iterations.
//\end{OgmgInclude.tex}
//===================================================================================
{
  return numberOfIterations;
}



//\begin{>>OgmgInclude.tex}{\subsection{sizeOf}} 
real Ogmg:: 
sizeOf( FILE *file /* =NULL */ ) const 
//===================================================================================
// /Description: 
//   Return number of bytes allocated by Ogmg; Optionally print detailed info to a file
//
// An estimate of space requirements is  $20N$ (2D) or $40N$ (3D) where $N$ is the number of
//  grid points on the finest grid (assumes maximum number of levels so that $.5+.25+.125+...=1$.
// For line smoothers there is addition space required.
//
// /file (input) : optionally supply a file to write detailed info to. Choose file=stdout to
// write to standard output.
// /Return value: the number of bytes.
//\end{OgmgInclude.tex}
//===================================================================================
{
  real size=0.;

  size+=sizeof(*this);

  real uMGSize,fMGSize,cMGSize,defectMGSize,uOldSize,mgcgSize,operatorsSize,interpolantSize,
    tridSize,directSize;
  

  // In some grid functions, don't count finest level since we share
  if( uMG.numberOfComponentGrids()>0 )
  {
    uMGSize=uMG.sizeOf()-uMG.multigridLevel[0].sizeOf(); // 1N
    size+=uMGSize;
  }
  if( fMG.numberOfComponentGrids()>0 )
  {
    fMGSize=fMG.sizeOf()-fMG.multigridLevel[0].sizeOf(); // 2N
    size+=fMGSize;
  }
  if( cMG.numberOfComponentGrids()>0 )
  {
    // If user supplies the coeff matrix then we just keep a ref to the finest level
    if( equationToSolve==OgesParameters::userDefined && cMG.numberOfMultigridLevels()>1 )
      cMGSize=cMG.sizeOf()-cMG.multigridLevel[0].sizeOf(); // 9N or 27N
    else
      cMGSize=cMG.sizeOf();
    size+=cMGSize;
  }
  
  defectMGSize=defectMG.sizeOf();                           // 2N
  size+=defectMGSize;
  uOldSize=uOld.sizeOf();       // uOld only lives on the fine grid     // 1N
  size+=uOldSize;

  if( v!=NULL )
    size+=v->sizeOf();

  if( numberOfExtraLevels>=0 )
  {
    // count the size in mgcg -- the grids on level 0 are referenced.
    real sizeOfReference=0.;
    for( int grid=0; grid<mgcg.numberOfComponentGrids(); grid++ )
    {
      sizeOfReference+=mgcg[grid].sizeOf();
    }
    mgcgSize=mgcg.sizeOf()-sizeOfReference;  // subtract off the duplicate space
    size+=mgcgSize;  // mask, center, jacobian =    2*7N or 2*13N  // don't count finest level if we can share
                           //                             total =  29N  59N
  }
  int level;
  if( operatorsForExtraLevels!=NULL )
  {
    operatorsSize=0.;
    for( int extraLevel=0; extraLevel<numberOfExtraLevels; extraLevel++ ) // is this correct?
      operatorsSize+=operatorsForExtraLevels[extraLevel].sizeOf();  
    size+=operatorsSize;
    
  }
  for( level=0; level<mgcg.numberOfMultigridLevels(); level++ )
  {
    if( interpolantWasCreated(level) )
      interpolantSize=interpolant[level].sizeOf(); 
    size+=interpolantSize;
    tridSize=0.;
    for( int grid=0; grid<mgcg.numberOfComponentGrids(); grid++ )
    {
      for( int axis=0; axis<mgcg.numberOfDimensions(); axis++ )
      {
	if( lineSmoothIsInitialized(axis,grid,level) ) 
	  tridSize+=tridiagonalSolver[level][grid][axis]->sizeOf();
      }
    }
    size+=tridSize;
  }
  if( parameters.useDirectSolverOnCoarseGrid )
  {
    directSize=directSolver.sizeOf();
    size+=directSize;
  }
  
  // count arrays used for IBS smoothing
  real sizeIBS=0.;
  if( ipn!=NULL )
  {
    for( int i=0; i<numberOfIBSArrays; i++ )
      sizeIBS+=ndipn[i];
    sizeIBS*=mgcg.numberOfDimensions()*sizeof(int);

    size+=sizeIBS;
  }

  if( file!=NULL && myid==0 )
  {
    const real meg=1024.*1024.; // 1e6;
    fprintf(file," Ogmg::sizeOf: uMG=%6.1f M, fMG=%6.1f M, cMG=%6.1f M, defectMG=%6.1f M, uOld=%6.1f M, mgcg=%6.1f M, \n"
	    "                    operators=%6.1f M, interpolant=%6.1f M, trid=%6.1f M, direct=%6.1f M, IBS=%6.1f M\n"
	    "                 ** total = %6.1f M \n",
	    uMGSize/meg,fMGSize/meg,cMGSize/meg,defectMGSize/meg,uOldSize/meg,mgcgSize/meg,
	    operatorsSize/meg,interpolantSize/meg,
	    tridSize/meg,directSize/meg,sizeIBS/meg,
            size/meg);
  }
  
  return size;
}

//\begin{>>OgmgInclude.tex}{\subsection{setOgmgParameters}}
int  Ogmg::
setOgmgParameters(OgmgParameters & parameters_ )
//==================================================================================
// /Description:
// set parameters equal to another parameter object.
//
//\end{OgmgInclude.tex} 
//==================================================================================
{
//  assert( initialized );
//  parameters_.initializeGridDependentParameters(cg);
  parameters=parameters_;
  return 0;
}

//\begin{>>OgmgInclude.tex}{\subsection{fullMultiGrid}}
int Ogmg::
fullMultigrid()
//==================================================================================
// /Description:
//    Perform the first full multigrid cycle (nested iteration) in order to obtain an initial guess.
//
//
//                        X
//                       /
//                  X   X
//                 / \ /
//                X   X
//
//\end{OgmgInclude.tex} 
//==================================================================================
{
  real time=getCPU();
//    if( debug & 2 ) 
//    {
//      int level=0;
//      defect(level);
//      real maximumDefect=maxNorm(defectMG.multigridLevel[level]);
//      printF("Entering fullMultigrid(), maximum defect=%8.2e...\n",maximumDefect);
//    }
  
  const int iteration=0;
  int level=0;
  fullMultigridWorkUnits=0.;  // count work used in fullMultigrid

  // average the RHS to all coarser levels
  bool transferForcing=true;
  for( level=0; level<parameters.numberOfMultigridLevels()-1; level++ )
  {
    defectMG.multigridLevel[level]=fMG.multigridLevel[level];  // we could avoid this copy with an option in fineToCoarse
    fineToCoarse(level,transferForcing); 

    fullMultigridWorkUnits+=.25/(level==0? 1. : pow(pow(2.,double(mgcg.numberOfDimensions())),double(level)));
  }
  
  if( Ogmg::debug & 4 )
  {
    level=parameters.numberOfMultigridLevels()-1;
    fMG.multigridLevel[level].display(sPrintF(buff,"fullMultigrid rhs for coarse level=%i",level),
        debugFile,"%9.1e");
  }

  real maximumDefect=-1.;
  for( level=parameters.numberOfMultigridLevels()-1; level>0; level-- )
  {
    if( !parameters.useDirectSolverOnCoarseGrid || directSolver.isSolverIterative() )
      uMG.multigridLevel[level]=0.;  // initial guess for iterative solvers on the coarse level.

    const int numberOfSubCycles=1;
    cycle(level,iteration,maximumDefect,numberOfSubCycles);  // solve on this level using MG
    fullMultigridWorkUnits+=workUnits(level)/pow(pow(2.,double(mgcg.numberOfDimensions())),double(level));
    if( false && level==1 )
    {
      cycle(level,iteration,maximumDefect,numberOfSubCycles);  // solve on this level using MG
      fullMultigridWorkUnits+=workUnits(level)/pow(pow(2.,double(mgcg.numberOfDimensions())),double(level));
    }
    
    
    if( ps!=0 && (debug & 8) && ps->isGraphicsWindowOpen() )
    {
      psp.set(GI_TOP_LABEL,sPrintF(buff,"level=%i u during Full MG.",level)); 
      ps->erase();
      PlotIt::contour(*ps,uMG.multigridLevel[level],psp);
      defect(level);
      psp.set(GI_TOP_LABEL,sPrintF(buff,"level=%i defect during Full MG.",level)); 
      ps->erase();
      PlotIt::contour(*ps,defectMG.multigridLevel[level],psp);
      psp.set(GI_TOP_LABEL,sPrintF(buff,"level=%i RHS during Full MG.",level)); 
      ps->erase();
      PlotIt::contour(*ps,fMG.multigridLevel[level],psp);
    }

    uMG.multigridLevel[level-1]=0.;

    levelZero=level-1;          // for boundary conditions at level-1
    coarseToFineInterpolationWidth=3; // increase the accuracy of the interpolation
    coarseToFine(level-1);      // transfer coarse grid to fine grid at level-1
    coarseToFineInterpolationWidth=2;  // reset 

    fullMultigridWorkUnits+=.125/pow(pow(2.,double(mgcg.numberOfDimensions())),double(level));
    

    if( Ogmg::debug & 4 )
    {
      uMG.multigridLevel[level].display(sPrintF(buff,"fullMultigrid u from cycle for level=%i",level),
        debugFile,"%9.1e");

      uMG.multigridLevel[level-1].display(sPrintF(buff,"fullMultigrid u from coarseToFine for level=%i",level-1),
        debugFile,"%9.1e");
    }
  }    
  levelZero=0;
  if( false )
  { // could do a final smooth but is it worth it?
    level=0;
    int numberOfSmoothingSteps=2;
    smooth(level,numberOfSmoothingSteps,iteration);
  }
  if( debug & 2 )
  {
    level=0;
    defect(level);
    real maximumDefect=maxNorm(defectMG.multigridLevel[level]);
    printF("...leaving fullMultigrid(). maximum defect=%8.2e (work units=%8.2e)\n\n",
                maximumDefect,fullMultigridWorkUnits);
  }
  if( ps!=0 && (debug & 8) && ps->isGraphicsWindowOpen() )
  {
    level=0;
    psp.set(GI_TOP_LABEL,sPrintF(buff,"level=%i u after Full MG.",level)); 
    ps->erase();
    PlotIt::contour(*ps,uMG.multigridLevel[level],psp);
    defect(level);
    psp.set(GI_TOP_LABEL,sPrintF(buff,"level=%i defect after Full MG.",level)); 
    ps->erase();
    PlotIt::contour(*ps,defectMG.multigridLevel[level],psp);
  }
  
  tm[timeForFullMultigrid]+=getCPU()-time;
  return 0;
}



//\begin{>>OgmgInclude.tex}{\subsection{solve}}
int Ogmg::
solve( realCompositeGridFunction & u, realCompositeGridFunction & f)
//==================================================================================
// /Description:
//   Solve Au=f with multigrid
//
// /u (input/output) : initial guess on input, answer on output. It is NOT necessary
//    that u be defined on all multigrid levels. u may only live on finest level.
//    It is best if u satisfies the boundary conditions on input although this is not required.
// /f (input) : right hand side for the problem. f should be defined for the finest level.
//    As with u, f can only be defined on the finest level if desired.
//\end{OgmgInclude.tex} 
//==================================================================================
{
  real time=getCPU();

  if( debug & 2 )
    printF(" ******** Ogmg::solve ********\n");
  
  assert(parameters.gridDependentParametersInitialized);

  directSolver.setOgesParameters(parameters.ogesParameters);  // why is this done here ? do only once?

  CompositeGrid & m = *u.getCompositeGrid();
  if( m.numberOfMultigridLevels()>1 )
  {
    uMG.reference(u);   // reference to local versions
    fMG.reference(f);
  }
  else
  {
    // input grid functions only live on the finest level
    uMG.updateToMatchGrid(mgcg);
    fMG.updateToMatchGrid(mgcg);
 
    // uMG.multigridLevel[0].dataCopy(u);  // **** could avoid a copy *****
    // fMG.multigridLevel[0].dataCopy(f);

    for( int grid=0; grid<mgcg.numberOfComponentGrids(); grid++ )
    {
      uMG[grid].reference(u[grid]);
      uMG.multigridLevel[0][grid].reference(u[grid]);  // need to do this too ?
      fMG[grid].reference(f[grid]);
      fMG.multigridLevel[0][grid].reference(f[grid]);  // need to do this too ?
    }
  }
  
  if( parameters.useErrorEstimateForConvergence )
    uOld.updateToMatchGrid(mgcg.multigridLevel[0]);  // used in solve **** why is this in the class ? ****

  
  if( parameters.numberOfMultigridLevels() < mgcg.numberOfMultigridLevels() ||
      parameters.numberOfComponentGrids()  < mgcg.numberOfComponentGrids()  )
  {
    printF("Ogmg::solve: update parameters -- CompositeGrid has apparently changed\n");
    parameters.updateToMatchGrid(mgcg);
  }

//   if( FALSE && parameters.problemIsSingular )
//   {
//     cout << "Ogmg::solve: problem is singular-- adding 1 to f for testing...\n";
//     f+=1.;
//   }
  if( parameters.problemIsSingular )
  {
    printF("***Ogmg::solve: problem is singular, directSolver.setCompatibilityConstraint \n");
    directSolver.set(OgesParameters::THEcompatibilityConstraint,true);
    parameters.ogesParameters.set(OgesParameters::THEcompatibilityConstraint,true);

    createNullVector();

    // alpha and v are used for trying out a different way to project singular problems
    // in addAdjustmentForSingularProblem
    if( alpha.getLength(0)<m.numberOfMultigridLevels() )
    {
      alpha.redim(m.numberOfMultigridLevels());
      alpha=0.;
    }
    
    if( false )
    {

      if( v==NULL )
      {
	// v is used to project the singular problem -- it should approximately satisfy Lv = - r
	CompositeGrid & cg = *u.getCompositeGrid();
      
	v = new realCompositeGridFunction(cg);
	realCompositeGridFunction & vv=*v;
	for( int grid=0; grid<cg.numberOfComponentGrids(); grid++ )
	{
	  realArray & x = cg[grid].center();

	  // this is v for a square: Delta v = -1
	  Range all;
	  vv[grid]= .25*( x(all,all,all,0)*(1.-x(all,all,all,0)) + x(all,all,all,1)*(1.-x(all,all,all,1)) );
	
	}
      }
    }
    
  }

  if( parameters.outputMatlabFile )
   cycleResults.redim(parameters.maximumNumberOfIterations+2,numberOfCycleResults+mgcg.numberOfComponentGrids());
    
  if( parameters.cycleType==OgmgParameters::cycleTypeF )
  { // initialize the iterationCount[level] for an F-cycle
    if( iterationCount==NULL )
    {
      iterationCount= new int[mgcg.numberOfMultigridLevels()];
    }
    
    for( int l=0; l<mgcg.numberOfMultigridLevels(); l++ )
    {
      iterationCount[l]=0;
    }
  }

  RealArray error(parameters.maximumNumberOfIterations+2);
  real maximumDefect=-1., maximumDefectOld=1.;

  // Get initial defect here and save to outputResults as "iteration==0"
  if( // parameters.useFullMultigrid ||   // why was this here ??
      debug & 2 || parameters.outputMatlabFile ) 
  {
    int level=0;
    defect(level);
    maximumDefect=maxNorm(defectMG.multigridLevel[level]);

    if( !parameters.useFullMultigrid )
      computeDefectRatios(level);
    
    if( debug & 2 )
      printF("Solve: maximum defect at start=%8.2e...\n",maximumDefect);
    
    if( debug & 2 || parameters.outputMatlabFile ) 
    {
      // output initial results
      int iteration=0;
      real defectNew=maximumDefect, defectOld=maximumDefect;
      outputResults(level,iteration,maximumDefect,defectNew,defectOld );
    }

  }

  // For fourth-order we need to initially set some values at ghost points.
  applyInitialConditions();

  int level=0;
  real defectConvergenceRate,errorConvergenceRate,errorEstimate;
  for( int iteration=0; iteration<parameters.maximumNumberOfIterations; iteration++ )
  {
    real timeI=getCPU();

    if( parameters.useErrorEstimateForConvergence )
    {
      uOld.dataCopy(uMG.multigridLevel[0]);   // why does uOld have to be in the class ??????????????????????????????
      tm[timeForMiscellaneous]+=getCPU()-timeI;
    }
    
    if( iteration==0 && parameters.useFullMultigrid )
    {
      fullMultigrid();   // obtain an initial guess through a nested iteration from the coarse level
    }
    else
      fullMultigridWorkUnits=0.;  // count work used in fullMultigrid

    cycle(level,iteration,maximumDefect,parameters.maximumNumberOfIterations);

    if( maximumDefect<0. )
    {
      printF("******Ogmg:solve:ERROR: maximumDefect not set by cycle! ******\n");
    }

    real time0=getCPU();
    if( parameters.useErrorEstimateForConvergence )
    {
      error(iteration)=l2Error(uMG.multigridLevel[0],uOld);  
      tm[timeForMiscellaneous]+=getCPU()-time0;
    }
    else
    {
      error(iteration)=0.;
    }
    
    numberOfCycles++;
    
    // In order to be able to test the convergence criteria on the first iteration
    // we make a rough guess for the errorConvergenceRate.
    real errorScale=REAL_EPSILON*100.;  // should scale by uNorm
    
    if( parameters.useErrorEstimateForConvergence )
    {
      if( iteration>0 )
	errorConvergenceRate=error(iteration)/max(error(iteration-1), errorScale);
      else
	errorConvergenceRate=.9;  // this is a guess. Could do better here.
      errorEstimate=error(iteration)*errorConvergenceRate/(1.-errorConvergenceRate);
    }
    else
    {
      errorEstimate=0.;
    }
    
    if( iteration>0 )
    {
	
      defectConvergenceRate=maximumDefect/maximumDefectOld;
      if( Ogmg::debug & 4 )
      {
        timeI=getCPU()-timeI;  // cpu time for this iteration.
	
        printF("Ogmg:solve: e%i=%7.1e, e%i/e%i=%5.3f, err est=%7.1e, "
               " d%i=%7.1e, d%i/d%i=%6.4f, ECR=%5.2f ***\n",
	       iteration,error(iteration),iteration,iteration-1,errorConvergenceRate,errorEstimate,
	       // -timeI/log(max(DBL_MIN,defectConvergenceRate)),
	       iteration,maximumDefect,iteration,iteration-1,defectConvergenceRate,
	       pow(defectConvergenceRate,1./(max(1.,workUnits(0)))));
      }
    }
    else
    {
      if( Ogmg::debug & 4)
      {
	printF("Ogmg:solve: iteration=%i, maxDefect=%8.2e (tol=%8.2e), errorEstimate=%8.2e (tol=%8.2e)\n",
	       iteration,maximumDefect,parameters.residualTolerance*numberOfGridPoints,
	       fabs(errorEstimate),parameters.errorTolerance);
      }
    }
    

    if( debug & 2 ) 
      printF("   ->solve: (level=0) it=%i, maxDefect=%8.2e <? resTol*pts"
             "=(%7.1e)*(%i)=%7.1e diff=%7.1e err est=%7.1e <? %7.1e\n",
	     iteration+1, maximumDefect, parameters.residualTolerance,numberOfGridPoints,
                parameters.residualTolerance*numberOfGridPoints,error(iteration),fabs(errorEstimate),parameters.errorTolerance);

    numberOfIterations=iteration+1;
    if( maximumDefect < parameters.residualTolerance*numberOfGridPoints && 
        ( fabs(errorEstimate) < parameters.errorTolerance || iteration==0 ) )
    {
      break;
    }
    
    maximumDefectOld=maximumDefect;
  }

  removeAdjustmentForSingularProblem(0,0);

  // For fourth-order we need to set some values on the 2nd ghost line at the end
  applyFinalConditions();

  if( m.numberOfMultigridLevels()==1 )
  {
    u=uMG.multigridLevel[0];
    f=fMG.multigridLevel[0];
  }
  maximumResidual=maximumDefect;
  
  outputCycleInfo();

  tm[timeForSolve]+=getCPU()-time;
  return 0;
}

//\begin{>>OgmgInclude.tex}{\subsection{cycle}}
int Ogmg::
cycle(const int & level, const int & iteration, real & maximumDefect, const int & numberOfCycleIterations )
//==================================================================================
// /Description:
//   Perform a multigrid cycle. This routine is called recursively.
//
// /maximumDefect (input/output) : on input (if non-negative) 
//     this is the current maximum defect, on output this is the new  maximum defect.
//
//\end{OgmgInclude.tex} 
//==================================================================================
{

  workUnits(level)=0.;      // counts total work on this level and higher levels for this cycle.
  if( level==0 && iteration==0 )
    workUnits(level)=fullMultigridWorkUnits;
  
  real defectOld,defectNew=REAL_MAX*.1;

  defectOld=maximumDefect;
  if( iteration==0 ) defectRatio(Range(),level)=-1.;  // this means it is not computed
  
  if( defectOld<0. && ( (level==0 && (debug & 2)) || debug & 4 || parameters.showSmoothingRates) )
  {
    // compute the initial defect if we are printing convergence rates:
    defect(level);
    defectOld= maxNorm(defectMG.multigridLevel[level]);
    computeDefectRatios(level);  // for auto-sub-smooth
  }

  addAdjustmentForSingularProblem(level,iteration);
  
  bool useTrueCycle=false;  // if false we sometimes do pre+post smooths all at once
  if( level<mgcg.numberOfMultigridLevels()-1 )
  {


    if( iteration==0 || useTrueCycle )
    {
      // On the first iteration we apply a pre-smooth, otherwise we 
      // combine pre and post smooths in the post smooth.

      // printF("level=%i, iteration=%i, number of presmooths=%i \n",level,iteration,parameters.numberOfSmooths(0,level));
      
      // *wdh* 030429 smooth(level,parameters.numberOfSmooths(0,level),iteration);
      // for testing we sometimes smooth lots of times on the first step:
      int numberOfSmoothingSteps=parameters.numberOfSmooths(0,level);
      if( iteration==0 && level==0 ) 
        numberOfSmoothingSteps = max(numberOfSmoothingSteps,parameters.minimumNumberOfInitialSmooths);
      smooth(level,numberOfSmoothingSteps,iteration);

      // printF(">>>>>>>>>>>level=%i, iteration=%i, number of PRE  smooths=%i, numberOfCycles=%i\n",
      //       level,iteration,numberOfSmoothingSteps,parameters.numberOfCycles(level));

      if( ps!=0 && (debug & 8 ) && ps->isGraphicsWindowOpen() )
      {
	psp.set(GI_TOP_LABEL,sPrintF(buff,"cycle: it=%i, level=%i, after initial smooth.",iteration,level)); 
	ps->erase();
	PlotIt::contour(*ps,uMG.multigridLevel[level],psp);
      }
    }
    

    // uMG.multigridLevel[level].applyBoundaryCondition(0,BCTypes::extrapolateInterpolationNeighbours);
    defect(level);
    workUnits(level)+=1.; 
    computeDefectRatios(level);  // for auto-sub-smooth

    fineToCoarse(level);   // compute f[level+1] = Restriction of the defect
    workUnits(level)+=.25; 

    #ifdef USE_PPP
    for( int grid=0; grid<mgcg.numberOfComponentGrids(); grid++ )
    {
      const realSerialArray & uLocal = uMG.multigridLevel[level+1][grid].getLocalArrayWithGhostBoundaries();
      realSerialArray & uLocalnc = (realSerialArray &) uLocal;
      uLocalnc=0.;
    }
    #else
      uMG.multigridLevel[level+1]=0.;       // set initial guess for next level
    #endif

    int numberOfSubCycles=parameters.numberOfCycles(level+1);
    if( parameters.cycleType==OgmgParameters::cycleTypeF )
    { // here is an F cycle:
      //  level+1== 1 : perform 2,2,2,2,......  subcycles
      //            2 : perform 2,1,2,1,......  subcycles
      //            3 : perform 2,1,1,2,1,1,... subcycles
      
      numberOfSubCycles=(iterationCount[level+1] % (level+1)) == 0 ? 2 : 1;
      iterationCount[level+1]++;
    }
    if( (level+1)==mgcg.numberOfMultigridLevels()-1 ) 
      numberOfSubCycles=1;  // only 1 iteration on the coarset level

    if( debug & 16 && parameters.cycleType==OgmgParameters::cycleTypeF )
      printF("Before cycle(level+1=%i): iterationCount[level+1]=%i numberOfSubCycles=%i\n",level+1,iterationCount[level+1],numberOfSubCycles);
    
    for( int subIteration=0; subIteration<numberOfSubCycles; subIteration++)
    {
      cycle(level+1,subIteration,maximumDefect,numberOfSubCycles);      // solve for u[level+1] (approximately)
      workUnits(level)+=workUnits(level+1)/pow(2.,double(mgcg.numberOfDimensions()));
    }

    coarseToFine(level);   // correct u[level]
    workUnits(level)+=.125; 

    // ** if this is not the last iteration at this level then do both pre and post smooths at once.
    int numberOfSmoothingSteps=useTrueCycle || (level>0 && iteration>=(numberOfCycleIterations-1)) ? 
      parameters.numberOfSmooths(1,level) :  parameters.numberOfSmooths(0,level)+parameters.numberOfSmooths(1,level);
//      int numberOfSmoothingSteps=useTrueCycle || (level>0 && iteration>=(parameters.numberOfCycles(level)-1)) ? 
//        parameters.numberOfSmooths(1,level) :  parameters.numberOfSmooths(0,level)+parameters.numberOfSmooths(1,level);

    // printF("<<<<<<<<<<<level=%i, iteration=%i, number of post smooths=%i, numberOfCycles=%i\n",
    //        level,iteration,numberOfSmoothingSteps,parameters.numberOfCycles(level));
    
    smooth(level,numberOfSmoothingSteps,iteration+1);

    if( ps!=0 && (debug & 8 ) && ps->isGraphicsWindowOpen() )
    {
      psp.set(GI_TOP_LABEL,sPrintF(buff,"cycle: it=%i, level=%i, after smooth.",iteration,level)); 
      ps->erase();
      PlotIt::contour(*ps,uMG.multigridLevel[level],psp);
    }

  }
  else
  { // Coarse grid solution: 
    real time=getCPU();
    if( debug & 8 )
      fMG.multigridLevel[level].display("RHS for coarse grid solve",debugFile,"%8.1e ");
    
    // Oges::debug=63;
    if( parameters.useDirectSolverOnCoarseGrid )
    {
      if( debug & 4 )
	printF("%*.1s  ***direct solve on level %i, iteration=%i\n",level*4," ",level,iteration);
        
//    if( directSolver.isSolverIterative() ) 
//    {
//      real tol=REAL_EPSILON*100.;
//      directSolver.set(OgesParameters::THEpreconditioner,OgesParameters::incompleteLUPreconditioner);
//      directSolver.set(OgesParameters::THErelativeTolerance,max(tol,REAL_EPSILON*10.));
//      directSolver.set(OgesParameters::THEmaximumNumberOfIterations,10000);
//    } 

//       if( directSolver.isSolverIterative() ) 

      if( false )
      {
        // *wdh* 061208 -- try this
	// uMG.multigridLevel[level]=0.;  
        assign(uMG.multigridLevel[level],0.);
	
// 	for( int grid=0; grid<mgcg.numberOfComponentGrids(); grid++ )
// 	  fMG.multigridLevel[level][grid].updateGhostBoundaries();
//         fMG.multigridLevel[level]=0.;
	
      }

      // Oges::debug=63;
      

      directSolver.solve( uMG.multigridLevel[level],fMG.multigridLevel[level] );
      workUnits(level)+=orderOfAccuracy+1;  // estimate work required for back-substitution ?????

      if( debug & 8 )
	uMG.multigridLevel[level].display(sPrintF("u after coarse grid solve, level=%i",level),debugFile,"%8.1e ");
    }
    else
    {
      if( debug & 4 )
	printF(" ***iterate to solve coarse grid problem on level %i\n",level);
      int numberOfSmoothingSteps=parameters.numberOfSmooths(0,level)+parameters.numberOfSmooths(1,level);
      smooth(level,numberOfSmoothingSteps,iteration);
    }
    tm[timeForDirectSolver]+=getCPU()-time;
    

    if( ps!=0 && (debug & 4 ) && ps->isGraphicsWindowOpen() )
    {
      psp.set(GI_TOP_LABEL,sPrintF(buff,"cycle: it=%i, level=%i, solution on coarsest level.",iteration,level)); 
      ps->erase();
      PlotIt::contour(*ps,uMG.multigridLevel[level],psp);
    }

  }

  
  if( parameters.problemIsSingular )
  {
    // realCompositeGridFunction & uu =(realCompositeGridFunction &)uMG.multigridLevel[level];
    if( level==0 )
      setMean(uMG.multigridLevel[level],parameters.meanValueForSingularProblem,level);
    else
      setMean(uMG.multigridLevel[level],0.,level);
  }


//   if( false && parameters.problemIsSingular && level==0 ) // level<mgcg.numberOfMultigridLevels()-1 )
//   {
//     // reset the right hand side
//     if( alpha(level) != 0. )
//       fMG.multigridLevel[level]+=alpha(level)*rightNullVector.multigridLevel[level];
//   }


  if( level==0 )
  { // we need to return the maximum defect for the convergence test -- if we are not close we maybe could 
    // just use the value from the next to last smooth?
    defectNew=0.;
    const int option=1;  // max norm
    for( int grid=0; grid<mgcg.multigridLevel[level].numberOfComponentGrids(); grid++ )
    {
      defectNew=max(defectNew,defectNorm(level,grid,option));  // this is approximate but fast
    }
    maximumDefect=defectNew;
    if( maximumDefect < parameters.residualTolerance*numberOfGridPoints )
    {
      // If we are close to convergence we compute the true max-defect
  
      defectNew=defectMaximumNorm(level);

//         real newDefectNew=defectMaximumNorm(level);
//         defect(level);   // work units ***
//         // workUnits(level)+=1;
//         defectNew= maxNorm(defectMG.multigridLevel[level]);
//         printF("@@@@ cycle: level=%i, maxDefect(approx)=%8.2e , actual=%8.2e, new-actual=%8.2e\n",
//               level,maximumDefect,defectNew,newDefectNew);

      if( debug & 2 )
        printF("@@@@ cycle: level=%i, maxDefect(approx)=%8.2e , actual=%8.2e\n",
               level,maximumDefect,defectNew);

    }
    // printF(" cycle: level=%i, maxDefect(approx)=%8.2e , actual=%8.2e\n",level,maximumDefect,defectNew);
    maximumDefect=defectNew;
  }

  // This next function computes the defect somtimes which is needed for some reason ***fix this***
  outputResults(level, iteration+1, maximumDefect, defectNew, defectOld);  // iteration->iteration+1 030710
  

  return 0;
}
  


// =================================================================================================================
//! Output results for this multigrid cycle
// =================================================================================================================
int Ogmg::
outputResults(const int & level, const int & iteration, real & maximumDefect, real & defectNew, real & defectOld )
{

  real time0=getCPU();  // start timeForMiscellaneous

  const bool print = myid==0;

  defectNew=maximumDefect;
  if( (level==0 && (debug & 2)) || debug & 4 || parameters.showSmoothingRates )
  {
    real time1=getCPU();

    defect(level);  // is this needed ??? *******************************************************************

    tm[timeForMiscellaneous]-=getCPU()-time1;     // don't count time for defect here 

    defectNew= maxNorm(defectMG.multigridLevel[level]);
    const real defectRatio=defectNew/max(defectOld,REAL_MIN*100.);
    if( iteration>0 )
    {
      printF("%*.1s Ogmg::cycle:level=%i, it=%i, WU=%5.2f, defect=%7.2e, defect/defectOld=%6.3f,"
	     "  ECR%i=%8.3f %s\n",level*4," ",level,iteration,workUnits(level),defectNew,defectRatio,
	     level,pow(defectRatio,1./(max(1.,workUnits(level)))),
	     (level==0 ? "***" : " "));
    }
    else
    {
      printF("%*.1s Ogmg::cycle:level=%i, it=%i, Initial defect =%7.2e\n",level*4," ",level,iteration,defectNew);
    }
  }
  else if( level==0 )
  {
    
/* ----
    // **** this needs to be computed if debug is not turned on *** fix this ***

    real time1=getCPU();
    defect(level);              // is this needed ???  **************************************************************
    tm[timeForMiscellaneous]-=getCPU()-time1;     // don't count time for defect here 


    defectNew= maxNorm(defectMG.multigridLevel[level]);   // ** this may not have been computed!
    ---- */
  }
  maximumDefect=defectNew;

  if( !(level==0 && (Ogmg::debug & 2 || parameters.outputMatlabFile) ) )
  {
    tm[timeForMiscellaneous]+=getCPU()-time0;
    return 0;
  }
  
  
  RealArray maxDefect(mgcg.numberOfComponentGrids());
  RealArray l2Defect(mgcg.numberOfComponentGrids());
  
  if( parameters.showSmoothingRates )
  {
    // NOTE: These computations require communication
    int grid;
    for( grid=0; grid<mgcg.multigridLevel[level].numberOfComponentGrids(); grid++ )
    {
      maxDefect(grid)=maxNorm(defectMG.multigridLevel[level][grid]);
      l2Defect(grid)=l2Norm(defectMG.multigridLevel[level][grid]);
    }
  }
   
  if( !print )
  {
    tm[timeForMiscellaneous]+=getCPU()-time0;
    return 0;
  }
  
  if( level==0 && Ogmg::debug & 2 )
  { 
    // write results to a file for easy insertion into the documentation
    if( iteration==0 )
    {
      fprintf(infoFile,
	      "\\begin{table}[hbt]\n"
	      "\\begin{center}\n"
	      "\\begin{tabular}{|c|c|c|c|c|} \\hline \n"
              " $i$   & $\\vert\\vert\\mbox{res}\\vert\\vert_\\infty$  &  CR     &  WU    & ECR  \\\\   \\hline \n");
      if( parameters.showSmoothingRates )
      {
        // output the residual on each grid if smoothing rates are being displayed
	fprintf(infoFile,"\\begin{tabular}{");
        int grid;
	for( grid=0; grid<mgcg.multigridLevel[level].numberOfComponentGrids(); grid++ )
          fprintf(infoFile,"|c");
	fprintf(infoFile,"|} \\hline \n"
        		 " $i$  &");
	for( grid=0; grid<mgcg.multigridLevel[level].numberOfComponentGrids(); grid++ )
	  fprintf(infoFile," r_%i(i)   &",grid+1);
        fprintf(infoFile,"	\\\\   \\hline \n");

        // fprintf(infoFile,"e=[\n  for matlab\n");  
      }
    }
    else 
    {
      if( defectNew/defectOld>1.e-3 )
      {
	fprintf(infoFile," $%2i$  & $%8.1e$ & $%5.3f$ & $%4.1f$ & $%4.2f$ \\\\ \n",iteration,
		defectNew,defectNew/defectOld,workUnits(level),
		pow(defectNew/defectOld , 1./(max(1.,workUnits(level)))));
	
      }
      else
      {
	fprintf(infoFile," $%2i$  & $%8.1e$ & $%6.5f$ & $%4.1f$ & $%4.2f$ \\\\ \n",iteration,
		defectNew,defectNew/defectOld,workUnits(level),
		pow(defectNew/defectOld , 1./(max(1.,workUnits(level)))));
      }
      

      if( parameters.showSmoothingRates )
      {
	// output the maximum defect on each grid
	fprintf(infoFile," $%2i$  &",iteration);
        int grid;
	for( grid=0; grid<mgcg.multigridLevel[level].numberOfComponentGrids(); grid++ )
	{
	  fprintf(infoFile," $%8.1e$ &",maxDefect(grid));
	}
	fprintf(infoFile," \\\\ %% max defects per grid\n");

	fprintf(infoFile," $%2i$  &",iteration);
	for( grid=0; grid<mgcg.multigridLevel[level].numberOfComponentGrids(); grid++ )
	{
	  fprintf(infoFile," $%8.1e$ ($n_s=%i$) &",l2Defect(grid),parameters.numberOfSubSmooths(grid,0));
	}
	fprintf(infoFile," \\\\ %% L2 defects per grid\n");
      }
    }
    
    // write info to a check file for regression testing
    if( iteration==0 )
    {
      fprintf(checkFile,"numberOfMultigridLevels = %i\n",mgcg.numberOfMultigridLevels());
      fprintf(checkFile," iteration  res      CR     WU   ECR\n");
    }
    fprintf(checkFile,"   %3i   %8.1e  %5.3f  %4.1f   %4.2f  \n",iteration,
	    defectNew,defectNew/defectOld,workUnits(level),
	    pow(defectNew/defectOld , 1./(max(1.,workUnits(level)))));

    // here we keep track of the total work units and total reduction in the residual
    // so we can output an average ECR.
    if( iteration==0 ) // do not count iteration==0
    {
      totalWorkUnits=0.;
      totalResidualReduction=1.;
    }
    else
    {
      totalWorkUnits+=workUnits(level);
      totalResidualReduction*=defectNew/defectOld;
    }
  }

  if( level==0 && parameters.outputMatlabFile )
  {
    // save results for later output to a matlab file
    cycleResults(iteration,defectPerCycle)=maximumDefect;
    cycleResults(iteration,workUnitsPerCycle)=workUnits(level);
    for( int grid=0; grid<mgcg.multigridLevel[level].numberOfComponentGrids(); grid++ )
    {
      cycleResults(iteration,grid0DefectPerCycle+grid)=l2Defect(grid);
    }
    
  }

  tm[timeForMiscellaneous]+=getCPU()-time0;
  return 0;
}


//! Output results about the cycle convergence rates etc. in a form suitable for matlab
int Ogmg::
outputCycleInfo( )
{
  if( parameters.outputMatlabFile )
  {
    const int startCycle=0;

    FILE *matlabFile = fopen("ogmg.m","w" ); 
    int i;
    fprintf(matlabFile,"t=[");
    for( i=startCycle; i<=numberOfIterations; i++ ) // *note* start at iteration 1
      fprintf(matlabFile,"%i ",i);
    fprintf(matlabFile,"];\n");

    fprintf(matlabFile,"defect=[");
    for( i=startCycle; i<=numberOfIterations; i++ )  // *note* start at iteration 1
      fprintf(matlabFile,"%9.3e ",cycleResults(i,defectPerCycle));
    fprintf(matlabFile,"];\n");

    
    fprintf(matlabFile,"clf\nset(gca,'FontSize',16);\n");

    fprintf(matlabFile,"plot(t,defect,'r-o');\n");
    fprintf(matlabFile,"title('Multigrid Convergence','FontSize',18);\n");
    
    fprintf(matlabFile,"ylabel('maximum residual');\n");

    fprintf(matlabFile,"xlabel('multigrid cycle');\n");
    fprintf(matlabFile,"set(gca,'YScale','log');\n");
    fprintf(matlabFile,"grid on\n");

//      real averageCR=pow(totalResidualReduction,1./max(1.,numberOfCycles));
//      real averageECR=pow(totalResidualReduction,1./max(1.,totalWorkUnits));
    // Compute convergence rates without the first cycle -- 
    // the first cycle usually skews the results to be too good.
    int cycleStart=2, cycleEnd=numberOfCycles;
    real residualReduction=cycleResults(cycleEnd,defectPerCycle)/cycleResults(cycleStart-1,defectPerCycle);
    real wu=sum(cycleResults(Range(cycleStart,cycleEnd),workUnitsPerCycle));

    real aveCR=pow(residualReduction,1./(cycleEnd-cycleStart+1));
    real aveECR= pow(residualReduction,1./max(1.,wu));

    aString buffer;

    int level=0;
//     aString cycleLabel = (parameters.cycleType==OgmgParameters::cycleTypeF ? "F" :
// 			  parameters.numberOfCycles(level)==1 ? "V" : 
// 			  parameters.numberOfCycles(level)==2 ? "W" : 
// 			  sPrintF(buffer,"C%i",parameters.numberOfCycles(level)));
    aString cycleLabel;
    if( parameters.cycleType==OgmgParameters::cycleTypeF )
      cycleLabel="F";
    else if( parameters.numberOfCycles(level)==1 )
      cycleLabel = "V"; 
    else if( parameters.numberOfCycles(level)==2 )
      cycleLabel = "W"; 
    else
      cycleLabel = sPrintF(buffer,"C%i",parameters.numberOfCycles(level));
    
    fprintf(matlabFile,sPrintF(buffer,"legend('%s','%s','%s','%7.1e points. %i levels.','CR=%3.3f, ECR=%3.2f, %s[%i,%i]');\n",
                               (const char*)infoFileCaption[0],
			       (const char*)infoFileCaption[1],(const char*)infoFileCaption[2],
                               (real)numberOfGridPoints,mgcg.numberOfMultigridLevels(),aveCR,aveECR,
                               (const char*)cycleLabel,parameters.numberOfSmooths(0,level),parameters.numberOfSmooths(1,level)));


    fprintf(matlabFile,"%% pause\n");
    
    fprintf(matlabFile,"%% print -deps2 residual.eps\n\n\n");
    fprintf(matlabFile,"%% print -depsc2 residual.eps\n\n\n");
    

    if( parameters.showSmoothingRates )
    {
      // output defects for each component grid
      int grid;
      for( grid=0; grid<mgcg.numberOfComponentGrids(); grid++ )
      {
	fprintf(matlabFile,"defect%i=[",grid);
	for( i=1; i<=numberOfIterations; i++ )
	  fprintf(matlabFile,"%9.3e ",cycleResults(i,grid0DefectPerCycle+grid));
	fprintf(matlabFile,"];\n");
      }
      const int numberOfSymbols=6;
      aString symbol[numberOfSymbols]={"r-o","g-x","b-s","c-<","m->","r-+"}; //

      fprintf(matlabFile,"plot(");
      for( grid=0; grid<mgcg.numberOfComponentGrids(); grid++ )
      {
	fprintf(matlabFile,sPrintF(buff,"t,defect%i,'%s'",grid,(const char*)symbol[(grid%numberOfSymbols)]));
        if( grid< mgcg.numberOfComponentGrids()-1 )
	  fprintf(matlabFile,",");
	else
          fprintf(matlabFile,");\n");
      }
      
      if( parameters.autoSubSmoothDetermination )
        fprintf(matlabFile,"title('Residuals by component grid, variable smoothing');\n");
      else
	fprintf(matlabFile,"title('Residuals by component grid, fixed smoothing');\n");
      
      fprintf(matlabFile,"ylabel('L_2 norm');\n");

      fprintf(matlabFile,"legend(");
      for( grid=0; grid<mgcg.numberOfComponentGrids(); grid++ )
      {
        fprintf(matlabFile,"'grid%i'",grid+1);
	if( grid< mgcg.numberOfComponentGrids()-1 )
	  fprintf(matlabFile,",");
	else
          fprintf(matlabFile,");\n");
      }
      fprintf(matlabFile,"xlabel('iteration');\n");
      fprintf(matlabFile,"set(gca,'YScale','log');\n");
//    fprintf(matlabFile,"grid on;\n");
      fprintf(matlabFile,"%% pause\n");
      fprintf(matlabFile,"%% print -deps2 residualByGrid.eps\n");
      fprintf(matlabFile,"%% print -depsc2 residual.eps\n\n\n");

    }
    

    fclose(matlabFile);
  }
  return 0;
}


void Ogmg::
displaySmoothers(const aString & label, FILE *file /* =stdout */ )
//==================================================================================
// /Description:
//    Show the smoothers currently being used on each grid and level
//
//\end{OgmgInclude.tex} 
//==================================================================================
{
  fPrintF(file,"\n -------- %s: smoothers --------\n",(const char*)label);
  int numberOfComponentGrids=mgcg.numberOfComponentGrids();
  int numberOfMultigridLevels=mgcg.numberOfMultigridLevels();
  if( mgcg.numberOfMultigridLevels()<=1 )
  {
    numberOfComponentGrids=parameters.smootherType.getLength(0);
    numberOfMultigridLevels=parameters.smootherType.getLength(1);
  }
  
  for( int grid=0; grid<numberOfComponentGrids; grid++ )
  {
    fPrintF(file,"grid %i : ",grid);
    for( int level=0; level<numberOfMultigridLevels; level++ )
    {
      int smooth = parameters.smootherType(grid,level);
      fPrintF(file,"%s[%i,%i] ",
	      (const char *)parameters.smootherName[smooth],
	      parameters.numberOfSmooths(0,level),parameters.numberOfSmooths(1,level));
    }
    fPrintF(file,"\n");
  }
  fPrintF(file," -----------------------------------------\n");
}



// extern real timeForApplyBCneumann;  // from neumann.C

// These are from boundaryConditions.C:
extern real timeForNeumannBC;
extern real timeForBC;
extern real timeForFinishBC;
extern real timeForGeneralNeumannBC;
extern real timeForExtrapolationBC;

extern real timeForSetupBC;
extern real timeForBCWhere;
extern real timeForBCOpt;
extern real timeForBC4Extrap;
extern real timeForBCFinal;

//\begin{>>OgmgInclude.tex}{\subsection{printStatistics}}
void Ogmg::
printStatistics(FILE *file_ /* =stdout */) const
//==================================================================================
// /Description:
//   Print performance statistics such as the cpu time required by various routines.
//
//\end{OgmgInclude.tex} 
//==================================================================================
{
  const int numberOfFiles = debug & 2 ? 2 : 1;
  const int cycles = max(1,numberOfCycles);    // number of cycles on level 1

  real *tm2 = (real*)tm;
  for( int i=0; i<numberOfThingsToTime; i++ )
    ParallelUtility::getMaxValues(tm2,tm2,numberOfThingsToTime);
  
  const real total = tm[timeForSolve];
  const int np= max(1,Communication_Manager::numberOfProcessors());
  
  real averageCR=pow(totalResidualReduction,1./max(1.,numberOfCycles));
  real averageECR=pow(totalResidualReduction,1./max(1.,totalWorkUnits));

  real aveCR=averageCR;
  real aveECR=averageECR;
  if( parameters.outputMatlabFile )
  {
    // Compute convergence rates without the first cycle -- 
    // the first cycle usually skews the results to be too good.
    int cycleStart=2, cycleEnd=numberOfCycles;
    real residualReduction=cycleResults(cycleEnd,defectPerCycle)/cycleResults(cycleStart-1,defectPerCycle);
    real wu=sum(cycleResults(Range(cycleStart,cycleEnd),workUnitsPerCycle));

    aveCR=pow(residualReduction,1./(cycleEnd-cycleStart+1));
    aveECR= pow(residualReduction,1./max(1.,wu));
  }
  
  real *tmnc = (real*) tm; // remove const
  int i;
  for(i=0; i<numberOfThingsToTime; i++ )
  {
    tmnc[i]=ParallelUtility::getMaxValue(tm[i]);
  }

  // here is the total time:
  const real sumTotal = tm[timeForSmooth]+tm[timeForDefect]+tm[timeForFineToCoarse]
      +tm[timeForCoarseToFine]+tm[timeForDirectSolver]+tm[timeForMiscellaneous];


  // *************Header label for the info file********************************************

  // Make a label for the info file to define the smoother -- make a list of the different smoothers used.
  aString buffer,smootherLabel;

  int level=0, grid, smooth=-1;
//   aString cycleLabel = (parameters.cycleType==OgmgParameters::cycleTypeF ? "F" :
//                         parameters.numberOfCycles(level)==1 ? "V" : 
// 			parameters.numberOfCycles(level)==2 ? "W" : 
//                         sPrintF(buffer,"C%i",parameters.numberOfCycles(level)));
  aString cycleLabel;
  if( parameters.cycleType==OgmgParameters::cycleTypeF )
    cycleLabel="F";
  else if( parameters.numberOfCycles(level)==1 )
    cycleLabel = "V"; 
  else if( parameters.numberOfCycles(level)==2 )
    cycleLabel = "W"; 
  else
    cycleLabel = sPrintF(buffer,"C%i",parameters.numberOfCycles(level));

  // Label the cycle: (e.g. V[2,1]
  smootherLabel= sPrintF(buffer,"%s[%i,%i]: ",(const char*)cycleLabel,
			     parameters.numberOfSmooths(0,level),parameters.numberOfSmooths(1,level));

  for( grid=0; grid<mgcg.numberOfComponentGrids(); grid++ )
  {
    int newSmooth=parameters.smootherType(grid,level);
    if( grid==0 || min(abs(parameters.smootherType(Range(0,grid-1),level)-newSmooth))!=0 )
    { // this is a new smoother (not equal to any of the previous)
      smooth=parameters.smootherType(grid,level);
      if( grid!=0 ) smootherLabel+=", ";
      smootherLabel+= sPrintF(buffer,"%s",(const char *)parameters.smootherName[smooth]);
      if( smooth==OgmgParameters::Jacobi )
        smootherLabel+= sPrintF(buffer," $\\omega=%3.2f$",parameters.omegaJacobi);
      if( smooth==OgmgParameters::GaussSeidel )
        smootherLabel+= sPrintF(buffer," $\\omega=%3.2f$",parameters.omegaGaussSeidel);
      if( smooth==OgmgParameters::redBlack || smooth==OgmgParameters::redBlackJacobi )
        smootherLabel+= sPrintF(buffer," $\\omega=%3.2f$",parameters.omegaRedBlack);
      if( smooth==OgmgParameters::lineJacobiInDirection1 ||
          smooth==OgmgParameters::lineJacobiInDirection2 ||
          smooth==OgmgParameters::lineJacobiInDirection3 ||
          smooth==OgmgParameters::alternatingLineJacobi )
        smootherLabel+= sPrintF(buffer," $\\omega=%3.2f$",parameters.omegaLineJacobi);
      if( smooth==OgmgParameters::lineZebraInDirection1 ||
          smooth==OgmgParameters::lineZebraInDirection2 ||
          smooth==OgmgParameters::lineZebraInDirection3 ||
          smooth==OgmgParameters::alternatingLineZebra )
        smootherLabel+= sPrintF(buffer," $\\omega=%3.2f$",parameters.omegaLineZebra);

    }
  }

  real mem=Overture::getCurrentMemoryUsage();
  real maxMem=ParallelUtility::getMaxValue(mem);  // max over all processors
  real minMem=ParallelUtility::getMinValue(mem);  // min over all processors
  real totalMem=ParallelUtility::getSum(mem);  // min over all processors
  real aveMem=totalMem/np;
  real maxMemRecorded=ParallelUtility::getMaxValue(Overture::getMaximumMemoryUsage());

  const bool print = myid==0;

  if( print )
  {
    fprintf(infoFile,
	    "\\hline \n");
    for( int n=0; n<5; n++ )
    {
      if( infoFileCaption[n]!="" )
	fprintf(infoFile,"\\multicolumn{5}{|c|}{%s}  \\\\\n",(const char*)infoFileCaption[n]);
    }

    if( parameters.useFullMultigrid )
    {
      fprintf(infoFile,"\\multicolumn{5}{|c|}{Full Multigrid.}  \\\\\n");
    }
  
    fprintf(infoFile,
	    "\\multicolumn{5}{|c|}{%s}  \\\\\n"
	    "\\multicolumn{5}{|c|}{%8.2e grid-points. %i levels. np=%i.}  \\\\\n"
	    "\\multicolumn{5}{|c|}{Average CR=$%4.3f$, ECR=$%3.2f$.}  \\\\\n"
	    "\\multicolumn{5}{|c|}{time/cycle = %6.2e s.}  \\\\\n"
	    "\\hline \n"
	    "\\end{tabular}\n"
	    "\\end{center}\n"
	    "\\caption{Multigrid convergence rates.}\n"
	    "%% \\label{fig:square} \n"
	    "\\end{table}\n" 
	    "{\\footnotesize\n"
	    "\\begin{verbatim}\n",
	    (const char*)smootherLabel,
	    (real)numberOfGridPoints,
	    mgcg.numberOfMultigridLevels(),np,
	    aveCR,aveECR,tm[timeForSolve]/cycles);
    
  }
  
  for( int io=0; io<numberOfFiles; io++ )
  {
    // write info to the screen and the info file
    FILE *file = io==0 ? file_ : infoFile;

    if( file==NULL ) continue;
    
    if( print ) fprintf(file,"\n ========================Ogmg Summary=========================\n\n");
    
    real size=sizeOf(file);

    if( !print ) continue;
      
    fprintf(file," storage allocated = %6.2e MBytes, %7.1f bytes/(grid point) or "
          "%6.1f reals/(grid point)\n\n",size/1.e6,
          size/numberOfGridPoints,size/numberOfGridPoints/sizeof(real));

    fprintf(file," number of levels = %i \n",mgcg.numberOfMultigridLevels());
    fprintf(file," interpolate defect = %i\n",(int)parameters.interpolateTheDefect);
    if( parameters.cycleType==OgmgParameters::cycleTypeF )
    {
       fprintf(file," F-cycle used.\n");
    }
    else
    {
      fprintf(file," number of cycles per level=");
      for( level=0; level<mgcg.numberOfMultigridLevels(); level++ )
	fprintf(file," %i ",parameters.numberOfCycles(level));
      fprintf(file,"\n");
    }
    fprintf(file," number of smooths (global) per level=");
    for( level=0; level<mgcg.numberOfMultigridLevels(); level++ )
      fprintf(file," [%i,%i] ",parameters.numberOfSmooths(0,level),parameters.numberOfSmooths(1,level));
    fprintf(file,"\n");
    aString coarseGridSolver;
    if( parameters.useDirectSolverOnCoarseGrid )
      coarseGridSolver=directSolver.parameters.getSolverName();
    else
      coarseGridSolver="smoother";
    int length=coarseGridSolver.length();
    if( length > 70 )
    { // name is too long, split it.
      aString a=coarseGridSolver,b;
      int i;
      for( i=50; i<length; i++ )
      {
	if( coarseGridSolver[i]==' ' )
	{
          a=coarseGridSolver(0,i-1); b=coarseGridSolver(i+1,length-1);
	  break;
	}
        else if( i==(length-1) )
	{ // no split point found at a blank, split at char 60:
          int ii=60;
	  a=coarseGridSolver(0,ii-1)+"-"; b=coarseGridSolver(ii,length-1);
	}
      }
      fprintf(file,
              " coarse grid solver : %s\n"
	      "                      %s\n",(const char*)a,(const char*)b);
    }
    else
      fprintf(file," coarse grid solver : %s \n",(const char*)coarseGridSolver);

    fprintf(file," grid ordering in smooth is %s.\n",(parameters.gridOrderingForSmooth==0 ? "1..ng" :
						    parameters.gridOrderingForSmooth==1 ? "ng..1" : "alternating"));
    
    fprintf(file," number of iterations for implicit interpolation is %i\n",
            parameters.maximumNumberOfInterpolationIterations);
    

// 	   "   grid  smoother               iterations  grid name\n"
// 	   "   ----  --------               ----------  ---------\n");
//     level=0;
//     for( grid=0; grid<mgcg.numberOfComponentGrids(); grid++ )
//     {
//       int smooth = parameters.smootherType(grid,level);
//       smooth = smooth<0 || smooth>=numberOfSmoothers ? numberOfSmoothers : smooth;
//       fprintf(file,
//               " %5i %-20s   %8i      %s\n",
// 	     grid,
// 	     (const char *) parameters.smootherName[smooth],
// 	     parameters.numberOfSmooths(level),
// 	     (const char *) mgcg[grid].mapping().getName(Mapping::mappingName));
//     }

    fprintf(file,"\n");
    for( grid=0; grid<mgcg.numberOfComponentGrids(); grid++ )
    {
      fprintf(file,"grid %i : ",grid);
      for( level=0; level<mgcg.numberOfMultigridLevels(); level++ )
      {
        int smooth = parameters.smootherType(grid,level);
        fprintf(file,"%s[%i,%i] ",
                (const char *)parameters.smootherName[smooth],
                 parameters.numberOfSmooths(0,level),parameters.numberOfSmooths(1,level));
      }
      fprintf(file,
	      " : %s \n"
	      "         bc=",(const char*)mgcg[grid].getName());
      for( int axis=0; axis<mgcg.numberOfDimensions(); axis++ )
      {
	fprintf(file,"[");
        for( int side=0; side<=1; side++ )
	{
	  fprintf(file,"%s",(boundaryCondition(side,axis,grid)==OgmgParameters::extrapolate ? "dirichlet" :
				   boundaryCondition(side,axis,grid)==OgmgParameters::equation ? "neumann" : 
                                   mgcg[grid].boundaryCondition(side,axis)<0 ? "periodic" :
                                    mgcg[grid].boundaryCondition(side,axis)==0 ? "interp" :  "other"));
          if( side==0 )
            fprintf(file,",");
          else
            fprintf(file,"] ");

	}
      }
      fprintf(file,"\n         ave no. of subSmooths: ");
      for( level=0; level<mgcg.numberOfMultigridLevels(); level++ )
      {
        fprintf(file,"l%i=%4.1f, ",level,parameters.totalNumberOfSubSmooths(grid,level)/max(1.,parameters.totalNumberOfSmoothsPerLevel(level)));
      }
      
      fprintf(file,"\n");
      
    }  // end for grid

    if( parameters.numberOfBoundarySmoothIterations>0 && 
        parameters.numberOfBoundaryLayersToSmooth>0 && parameters.numberOfLevelsForBoundarySmoothing>0 )
    {
      fprintf(file," Boundary smoothing: local iterations=%i, layers=%i for %i levels.\n",
	      parameters.numberOfBoundarySmoothIterations,
              parameters.numberOfBoundaryLayersToSmooth,parameters.numberOfLevelsForBoundarySmoothing);
    }
    if( parameters.numberOfIBSIterations>0 && parameters.numberOfInterpolationSmoothIterations>0 && 
        parameters.numberOfInterpolationLayersToSmooth>0 && parameters.numberOfLevelsForInterpolationSmoothing>0 )
    {
      fprintf(file," IBS: global iterations=%i, local iterations=%i, layers=%i for %i levels.\n",
	      parameters.numberOfIBSIterations,parameters.numberOfInterpolationSmoothIterations,
              parameters.numberOfInterpolationLayersToSmooth,parameters.numberOfLevelsForInterpolationSmoothing);
    }
    

    // ************************************************************************8
    // ************Output INFO on Discrete BC's********************************8
    // ************************************************************************8
    if( orderOfAccuracy==2 )
    {
    }
    else if( orderOfAccuracy==4 )
    {
      fprintf(file,"\nDiscrete boundary conditions: B=boundary, G1=ghost-1, G2=ghost-2, extrapN=extrap to order N\n");
      
      aString ghost1,ghost2;
      int grid=0,level,ghostLine,orderOfExtrapolation;
      // ***** Dirichlet *****
      for( level=0; level<=min(1,mgcg.numberOfMultigridLevels()-1); level++ )
      {
	ghostLine=1;
	getGhostLineBoundaryCondition( OgmgParameters::extrapolate, ghostLine, grid, level, 
				       orderOfExtrapolation, &ghost1 );
	ghostLine=2;
	getGhostLineBoundaryCondition( OgmgParameters::extrapolate, ghostLine, grid, level, 
				       orderOfExtrapolation, &ghost2 );
      
        if( level==0 )
  	  fprintf(file,"  Dirichlet: l=0 : B=D,    G1=%8s, G2=%8s\n",(const char*)ghost1,(const char*)ghost2);
        else
          fprintf(file,"           : l>0 : B=D,    G1=%8s, G2=%8s\n",(const char*)ghost1,(const char*)ghost2);
      }
      // ***** Neumann or Mixed BC *****
      for( level=0; level<=min(1,mgcg.numberOfMultigridLevels()-1); level++ )
      {
	ghostLine=1;
	getGhostLineBoundaryCondition( OgmgParameters::equation, ghostLine, grid, level, 
				       orderOfExtrapolation, &ghost1 );
	ghostLine=2;
	getGhostLineBoundaryCondition( OgmgParameters::equation, ghostLine, grid, level, 
				       orderOfExtrapolation, &ghost2 );
      
        if( level==0 )
	  fprintf(file,"  Neumann  : l=0 : B=PDE,  G1=%8s, G2=%8s\n",(const char*)ghost1,(const char*)ghost2);
        else
	  fprintf(file,"           : l>0 : B=PDE,  G1=%8s, G2=%8s\n",(const char*)ghost1,(const char*)ghost2);
      }

    }
    aString corner="extrap  ", lowerLevelCorner="extrap  ";
    if( parameters.useSymmetryCornerBoundaryCondition )
    {
      sPrintF(corner,"taylor%i  ",orderOfAccuracy); 
      lowerLevelCorner=corner;
    }
    
    if( ogesSmoother!=NULL )
      printF("\n Oges smoother: %s\n",(const char*)ogesSmoother[0].parameters.getSolverName());
   
    fprintf(file," Corner BC: l=0 : %s, l>0 %8s\n",(const char*)corner,(const char*)lowerLevelCorner);
    

    fprintf(file,"\n"
	    "    Ogmg, Statistics  grids=%4i, gridPoints=%8i, processors=%i\n"
	    "    ----------------                  time (s)  time/cycle  percentage\n"
	    " smooth..(includes bc's)...............%6.2e  %6.2e   %6.2f%% \n"
	    " defect.(excluding those in smooth)....%6.2e  %6.2e   %6.2f%% \n"
	    " fine to coarse........................%6.2e  %6.2e   %6.2f%% \n"
	    " coarse to fine........................%6.2e  %6.2e   %6.2f%% \n"
	    " direct solve on coarsest level........%6.2e  %6.2e   %6.2f%% \n"
	    " miscellaneous.........................%6.2e  %6.2e   %6.2f%% \n"
	    " sum of above..........................%6.2e  %6.2e   %6.2f%%  \n"
            " Details:\n"
            "    defect called from smooth..........%6.2e  %6.2e  (%6.2f%%) (already counted)\n"
            "    relaxation part of smooth..........%6.2e  %6.2e  (%6.2f%%) (already counted)\n"
            "    extra boundary smoothing...........%6.2e  %6.2e  (%6.2f%%) (already counted)\n"
            "    extra interpolation smoothing......%6.2e  %6.2e  (%6.2f%%) (already counted)\n"
            "    tridiagonal factor part of smooth..%6.2e  %6.2e  (%6.2f%%) (already counted)\n"
            "    tridiagonal solve part of smooth...%6.2e  %6.2e  (%6.2f%%) (already counted)\n"
	    "    interpolation......................%6.2e  %6.2e  (%6.2f%%) (already counted)\n"
	    "    boundary conditions................%6.2e  %6.2e  (%6.2f%%) (already counted)\n"
	    "    initial guess with FMG.............%6.2e  %6.2e  (%6.2f%%) (already counted)\n"
	    "    fine to coarse BC's................%6.2e  %6.2e  (%6.2f%%) (already counted)\n"
	    "    compute norms of defect............%6.2e  %6.2e  (%6.2f%%) (already counted)\n"
	    "    ghost boundary update..............%6.2e  %6.2e  (%6.2f%%) (already counted)\n\n"
	    " total.................................%6.2e  %6.2e   %6.2f%% \n\n"
            " initialize............................%6.2e  %6.2e   %6.2f%%  (not counted above)\n"
            "    build extra levels.................%6.2e  %6.2e   %6.2f%%  (already counted)\n"
            "    operator averaging.................%6.2e  %6.2e   %6.2f%%  (already counted)\n"
            "    build predefined equations.........%6.2e  %6.2e   %6.2f%%  (already counted)\n"
            "    oges smoother init.................%6.2e  %6.2e   %6.2f%%  (already counted)\n\n"
            " TOTAL (solve+initialize)..............%6.2e  %6.2e\n"
	    " total number of grid points = %i \n"
	    " total number of cycles      = %i \n",
            mgcg.numberOfComponentGrids(),numberOfGridPoints,np,
	    tm[timeForSmooth],       tm[timeForSmooth]/cycles,       100.*tm[timeForSmooth]/total,
	    tm[timeForDefect],       tm[timeForDefect]/cycles,       100.*tm[timeForDefect]/total,
	    tm[timeForFineToCoarse], tm[timeForFineToCoarse]/cycles, 100.*tm[timeForFineToCoarse]/total,
	    tm[timeForCoarseToFine], tm[timeForCoarseToFine]/cycles, 100.*tm[timeForCoarseToFine]/total,
	    tm[timeForDirectSolver], tm[timeForDirectSolver]/cycles, 100.*tm[timeForDirectSolver]/total,
	    tm[timeForMiscellaneous],tm[timeForMiscellaneous]/cycles,100.*tm[timeForMiscellaneous]/total,
	    sumTotal,                sumTotal/cycles,                100.*sumTotal/total,
	    tm[timeForDefectInSmooth],tm[timeForDefectInSmooth]/cycles,100.*tm[timeForDefectInSmooth]/total,
	    tm[timeForRelaxInSmooth],tm[timeForRelaxInSmooth]/cycles,100.*tm[timeForRelaxInSmooth]/total,
	    tm[timeForBoundarySmooth],tm[timeForBoundarySmooth]/cycles,100.*tm[timeForBoundarySmooth]/total,
	    tm[timeForInterpolationSmooth],tm[timeForInterpolationSmooth]/cycles,100.*tm[timeForInterpolationSmooth]/total,
	    tm[timeForTridiagonalFactorInSmooth],tm[timeForTridiagonalFactorInSmooth]/cycles,
                    100.*tm[timeForTridiagonalFactorInSmooth]/total,
	    tm[timeForTridiagonalSolverInSmooth],tm[timeForTridiagonalSolverInSmooth]/cycles,
                    100.*tm[timeForTridiagonalSolverInSmooth]/total,
	    tm[timeForInterpolation],tm[timeForInterpolation]/cycles,100.*tm[timeForInterpolation]/total,
	    tm[timeForBoundaryConditions],tm[timeForBoundaryConditions]/cycles,
	    100.*tm[timeForBoundaryConditions]/total,
	    tm[timeForFullMultigrid],tm[timeForFullMultigrid]/cycles,
	    100.*tm[timeForFullMultigrid]/total,
	    tm[timeForFineToCoarseBC], tm[timeForFineToCoarseBC]/cycles, 100.*tm[timeForFineToCoarseBC]/total,
	    tm[timeForDefectNorm], tm[timeForDefectNorm]/cycles, 100.*tm[timeForDefectNorm]/total,
            tm[timeForGhostBoundaryUpdate],tm[timeForGhostBoundaryUpdate]/cycles,
                 100.*tm[timeForGhostBoundaryUpdate]/total,
	    tm[timeForSolve],        tm[timeForSolve]/cycles,        100.*tm[timeForSolve]/total,
	    tm[timeForInitialize],tm[timeForInitialize]/cycles,100.*tm[timeForInitialize]/total,
	    tm[timeForBuildExtraLevels],tm[timeForBuildExtraLevels]/cycles,100.*tm[timeForBuildExtraLevels]/total,
	    tm[timeForOperatorAveraging],tm[timeForOperatorAveraging]/cycles,100.*tm[timeForOperatorAveraging]/total,
	    tm[timeForBuildPredefinedEquations],tm[timeForBuildPredefinedEquations]/cycles,
                         100.*tm[timeForBuildPredefinedEquations]/total,
	    tm[timeForOgesSmootherInit],tm[timeForOgesSmootherInit]/cycles,100.*tm[timeForOgesSmootherInit]/total,
            tm[timeForSolve]+tm[timeForInitialize],(tm[timeForSolve]+tm[timeForInitialize])/cycles,
	    numberOfGridPoints,cycles
      );
    
    fprintf(file,"==== memory per-proc: [min=%g,ave=%g,max=%g](Mb), max-recorded=%g (Mb), total=%g (Mb)\n",
	    minMem,aveMem,maxMem,maxMemRecorded,totalMem);
    

    fprintf(file,"*** timeForNeumannBC=%8.2e timeForBC=%8.2e "
	    " timeForFinishBC=%8.2e timeForBCFinal=%8.2e\n"
	    "   timeForGeneralNeumannBC=%8.2e timeForExtrapolationBC=%8.2e \n     "
	    " timeForSetupBC=%8.2e, timeForBCWhere=%8.2e, timeForBCOpt=%8.2e timeForBC4Extrap=%8.2e\n",
	    timeForNeumannBC,timeForBC,timeForFinishBC,timeForBCFinal,timeForGeneralNeumannBC,
	    timeForExtrapolationBC,timeForSetupBC,timeForBCWhere,timeForBCOpt,timeForBC4Extrap );

    if( Ogmg::debug & 4 )
    {
      fprintf(file,"***time timeForNeumannBC=%8.2e timeForBC=%8.2e "
	      " timeForFinishBC=%8.2e timeForGeneralNeumannBC=%8.2e timeForExtrapolationBC=%8.2e "
	      " timeForSetupBC=%8.2e, timeForBCWhere=%8.2e\n",
              timeForNeumannBC,timeForBC,timeForFinishBC,timeForGeneralNeumannBC,
	      timeForExtrapolationBC,timeForSetupBC,timeForBCWhere );
    }
    
    
    // the average time/cycle for baseLineNumberOfGridPoints should be about 1.
    real baseLineNumberOfGridPoints=100000;   // choose this so the scaleTime is about 1 -- valid for the sun ultra 10
    real gridPointsNormalization=baseLineNumberOfGridPoints/numberOfGridPoints;
    real scaledTime= (tm[timeForSolve]/cycles)*gridPointsNormalization;
    
    if( parameters.outputMatlabFile )
    {
      // Output convergence rates with and without the first cycle -- 
      // the first cycle usually skews the results to be too good.
      for( int ii=1; ii<=2; ii++ )
      {
	int cycleStart=ii, cycleEnd=numberOfCycles;
	real residualReduction=cycleResults(cycleEnd,defectPerCycle)/cycleResults(cycleStart-1,defectPerCycle);
	real aveCR=pow(residualReduction,1./(cycleEnd-cycleStart+1));
	real wu=sum(cycleResults(Range(cycleStart,cycleEnd),workUnitsPerCycle));
	real aveECR= pow(residualReduction,1./max(1.,wu));
	fprintf(file,"\nIteration=%i..%i : Total WU=%8.2e, total res reduction=%8.2e, ave CR=%5.4f ave ECR=%5.3f\n",
		cycleStart,cycleEnd,wu,residualReduction,aveCR,aveECR);
      
      }
    }
    else
    {
	fprintf(file,"\nIteration=%i..%i : Total WU=%8.2e, total res reduction=%8.2e, ave CR=%5.4f ave ECR=%5.3f\n",
		1,numberOfCycles,totalWorkUnits,totalResidualReduction,averageCR,averageECR);
    }
    
    fprintf(file,"\n");
    

//    real averageTCR=pow(averageCR,1./(max(REAL_EPSILON,scaledTime)));
//      fprintf(file,"\nIteration=1..%i : Total WU=%8.2e, total res reduction=%8.2e, ave CR=%5.4f ave ECR=%5.3f"
//                   " aver TCR=%7.2f (scaledtime=%8.1e)\n",
//  	   numberOfCycles-1,totalWorkUnits,totalResidualReduction,
//                  averageCR,averageECR,averageTCR,scaledTime);

  } // end for io
  fprintf(infoFile,
          "\\end{verbatim}\n"
          "} %% end footnotesize\n");

}





//! Once the coefficient matrix is known on the finest level we can automatically
//! build the coeff matrices on coarser levels.
int Ogmg::
buildCoefficientArrays()
// ===================================================================================
// ===================================================================================
{
  real time0=getCPU();
  
  if( debug & 4 )
    printF("\n ------------ Entering Ogmg::buildCoefficientArrays -----\n\n");

  lineSmoothIsInitialized=FALSE;  // assume that all grids have changed

  realCompositeGridFunction & coeff = cMG.numberOfMultigridLevels()>1 ? cMG.multigridLevel[0] : cMG;

  // This next is no longer needed -- the boundaryConditionArray should already be set *wdh* 070216
  // initializeBoundaryConditions(coeff);

  if( debug & 4 ) printF("**** time for initializeBoundaryConditions=%e\n",getCPU()-time0);

  // initialize arrays used in the case of constant coefficients for predefined equations:
  initializeConstantCoefficients();
  
  real timeForOperators=0.;

  const int width = orderOfAccuracy+1;  // 3 or 5
  const int stencilSize=int( pow(double(width),double(mgcg.numberOfDimensions()))+1.5 );
  const int orderOfExtrapolation = orderOfAccuracy==2 ? 3 : 4;  // 5
  
  if( numberOfExtraLevels>0 )
  {
     // we need operators to apply boundary conditions. *** can we fix this ? ****
    if( false 
        // true  // +++++++++++++++++++++++++++++++++++++++++++++++++++++++++ *wdh* 040830 +++++++++++++++
        && operatorsForExtraLevels!=NULL )
    {
      delete [] operatorsForExtraLevels;     // **** don't have to delete if we have the same number of extra levels!
      operatorsForExtraLevels=NULL;
    }
    if( operatorsForExtraLevels==NULL )
      operatorsForExtraLevels= new CompositeGridOperators [numberOfExtraLevels];
 
    int level0=0;
    int l;
    for( l=0; l<numberOfExtraLevels; l++ )
    {
      int level=level0+l;

      // we need operators to apply boundary conditions.
      RealCompositeGridFunction & cl = cMG.multigridLevel[level+1];
      CompositeGrid & cgl = *cl.getCompositeGrid();

      if( debug & 16 )
        printF("+++++buildCoefficientArrays: level+1=%i cgl.rcData=%i \n",level+1,cgl.rcData);

      CompositeGridOperators & op = operatorsForExtraLevels[l];
      op.setStencilSize(stencilSize);
      op.setOrderOfAccuracy(orderOfAccuracy);

      real time1=getCPU();
      op.updateToMatchGrid(cgl);
      cl.setOperators(op);
      timeForOperators+=getCPU()-time1;

      // *************************************
      // **** Build the averaged operator ****
      // *************************************
      operatorAveraging(cMG,level);

      if( false && orderOfAccuracy==4 )
      { 
        // **** for testing *****

        printF("\n ++++++++++++++++ over-write averaged operator with laplacian from operators ++++++++++++++\n\n");
        // cgl.update(MappedGrid::THEinverseVertexDerivative );
	
        int stencilSize=cl[0].getLength(0);
	if( stencilSize==0 )  // for rectangular grids
	{
	   const int width = orderOfAccuracy+1;  // 3 or 5
           stencilSize=int(pow(double(width),double(cgl.numberOfDimensions()))+1.5);
	}
        if( Ogmg::debug & 4 ) cl.display("*************Coefficients from averaging ***********************","%5.1f ");
	
	
	op.setStencilSize(stencilSize);
        cl.destroy();
        cl.updateToMatchGrid(cgl,stencilSize);
        const int numberOfGhostLines=2;
        cl.setIsACoefficientMatrix(true,stencilSize,numberOfGhostLines);  
	cl.setOperators(op);

	cl=op.laplacianCoefficients();   // get the coefficients for the Laplace operator 
        // fill in the coefficients for the boundary conditions
	cl.applyBoundaryConditionCoefficients(0,0,BCTypes::dirichlet,BCTypes::allBoundaries);
	cl.applyBoundaryConditionCoefficients(0,0,BCTypes::extrapolate,BCTypes::allBoundaries);
        if( orderOfAccuracy==4 )
	{
	  BoundaryConditionParameters bcParams;
	  bcParams.ghostLineToAssign=2;
          bcParams.orderOfExtrapolation=orderOfExtrapolation; // orderOfAccuracy+1;
	  cl.applyBoundaryConditionCoefficients(0,0,BCTypes::extrapolate,BCTypes::allBoundaries,bcParams); // extrap 2nd ghost line
	}
	
        if( Ogmg::debug & 4 ) 
  	  cl.display(sPrintF(buff,"buildCoefficientArray: coeff from operators on level=%i",level+1),debugFile,"%9.1e");

      }
      
    }

    bool nullVectorNeedsToBeComputed=parameters.problemIsSingular && parameters.projectRightHandSideForSingularProblem;
    if( nullVectorNeedsToBeComputed )
    {
      // if the null vector needs to be computed, check to see if we can read it from a file (i.e. it was
      // saved from a previous computation)
      nullVectorNeedsToBeComputed= readLeftNullVector()!=0;
    }

    // do not finishBoundaryConditions until all levels have been generated since we don't
    // want the interpolation points to be inserted until the end.
    for( l=0; l<numberOfExtraLevels; l++ )
    {
      int level=level0+l;

      RealCompositeGridFunction & cl = cMG.multigridLevel[level+1];
      CompositeGrid & cgl = *cl.getCompositeGrid();

      CompositeGridOperators & opc = *cl.getOperators();
      opc.setStencilSize(stencilSize);
      opc.setOrderOfAccuracy(orderOfAccuracy);

      // We only need to apply finish boundary conditions on the coarsest level if 
      //    we are calling a direct solver *wdh* 011107
      // We need to finish BC's for singular problems if we need to compute the leftNullVectors
      if( (l==(numberOfExtraLevels-1) && parameters.useDirectSolverOnCoarseGrid) || 
          applyFinishBoundaryConditions || nullVectorNeedsToBeComputed )
      {
	if( debug & 4 ) 
	{
	  
          printF("*****buildCoefficientArray:  cl.finishBoundaryConditions(); for coarsest level %i\n",level+1);
          printF(" cornerBC: %i\n",bcParams.getCornerBC(0,0,0));
	}
	
        // bcParams.orderOfExtrapolation=2;
	
	if( false && parameters.useSymmetryCornerBoundaryCondition ) // these need to be implemented
	{
	  BoundaryConditionParameters::CornerBoundaryConditionEnum cornerBC=
	    orderOfAccuracy==2 ? BoundaryConditionParameters::taylor2ndOrderEvenCorner :
	    BoundaryConditionParameters::taylor4thOrderEvenCorner;

	  // cornerBC=BoundaryConditionParameters::evenSymmetryCorner; // ********************************
	  // cornerBC=BoundaryConditionParameters::oddSymmetryCorner; // ********************************
    
    
	  bcParams.setCornerBoundaryCondition(cornerBC);  // this sets all corners and edges
	}
	
	cl.finishBoundaryConditions(bcParams);

      }
      else
      {
        // we only call finishBoundaryConditions on each grid to fill in corner points
        // we do not want the interpolation equations filled in
        for( int grid=0; grid<cgl.numberOfComponentGrids(); grid++ )
	{
          // bcParams.orderOfExtrapolation=3; // ***** test for mixed BC ***
          if( !(equationToSolve!=OgesParameters::userDefined && cgl[grid].isRectangular() && 
		(level<mgcg.numberOfMultigridLevels()-1 || !parameters.useDirectSolverOnCoarseGrid)) )
	  {
	    if( debug & 4 ) 
              printF("*****buildCoefficientArray:  cl[grid].finishBoundaryConditions() for level %i\n",level+1);
   	    cl[grid].finishBoundaryConditions(bcParams);
	  }
	  
	}
	
      }
      

      // *** a corner with two adjacent neumann BC's had best results with orderOfExtrap==1 for
      // coarser levels. This didn't seem to hard dirichlet BC's --> probably the
      // best would be u(-1,-1)=u(1,1) ?? 
      if( FALSE )
      {
	bcParams.orderOfExtrapolation=2;
	cl.finishBoundaryConditions(bcParams);
	
	cl[0](0,-1,-1,0)=1.;
	cl[0](1,-1,-1,0)=0.;
	cl[0](2,-1,-1,0)=-1.;

        int n1 = cgl[0].gridIndexRange(End,axis1);
	cl[0](0,n1+1,-1,0)= 1.;
	cl[0](1,n1+1,-1,0)= 0.;
	cl[0](2,n1+1,-1,0)=-1.;

        int n2 = cgl[0].gridIndexRange(End,axis2);
	cl[0](0,-1,n2+1,0)=1.;
	cl[0](1,-1,n2+1,0)=0.;
	cl[0](2,-1,n2+1,0)=-1.;
	cl[0](0,n1+1,n2+1,0)= 1.;
	cl[0](1,n1+1,n2+1,0)= 0.;
	cl[0](2,n1+1,n2+1,0)=-1.;

      }
      
      if( debug & 64 )
      {
	cl.display(sPrintF(buff,"buildCoefficientArray: coeff on level=%i",level+1),debugFile,"%9.1e");
      }
    }
    // ******** do this here after we have averaged  

    if( applyFinishBoundaryConditions )
    {
      if( debug & 4 ) printF("*****buildCoefficientArray:  coeff.finishBoundaryConditions(); for finest level\n");
      coeff.finishBoundaryConditions();
    }
    else
    {
      // printF("*****buildCoefficientArray:  DO NOT call coeff.finishBoundaryConditions() for finest level\n");
      // we only call finishBoundaryConditions on each grid to fill in corner points
      // we do not want the interpolation equations filled in
      if( parameters.problemIsSingular && nullVectorNeedsToBeComputed )
      {
	if( debug & 4 ) printF("*****buildCoefficientArray:problemIsSingular  coeff.finishBoundaryConditions() for finest level\n");
	coeff.finishBoundaryConditions(bcParams);
      }
      else
      {
	for( int grid=0; grid<mgcg.numberOfComponentGrids(); grid++ )
	{
          if( !(equationToSolve!=OgesParameters::userDefined && mgcg[grid].isRectangular()) )
	  {
	    if( debug & 4 ) printF("*****buildCoefficientArray:  coeff[grid=%i].finishBoundaryConditions() "
                           "for finest level\n",grid);
	    coeff[grid].finishBoundaryConditions(bcParams);
	  }
	  
	}
      }
      
    }
    

    if( debug & 64 )
    {
      coeff.display(sPrintF(buff,"buildCoefficientArray: coeff on finest grid"),debugFile,"%9.1e");
    }
    // cMG.multigridLevel[0].finishBoundaryConditions();  // *** ???
    
  }
  else
  {
    // number of extra levels is zero 
    if( parameters.useDirectSolverForOneLevel )
    {
      // *wdh* 040830 cMG.multigridLevel[0].finishBoundaryConditions(bcParams);
      cMG.finishBoundaryConditions(bcParams);
    }
    
  }
  
  if( debug & 4 ) printF(" *** timeForOperators=%8.2e\n",timeForOperators);
  

  if( debug & 64 )
  {
    cMG.multigridLevel[mgcg.numberOfMultigridLevels()-1].display(
                 sPrintF(buff,"After finishBC:coeff on coarsest level=%i",mgcg.numberOfMultigridLevels()-1),
            debugFile,"%8.2e ");
  }

  directSolver.setGrid(mgcg.multigridLevel[mgcg.numberOfMultigridLevels()-1]); // direct solver for coarse grid

  // printF("*** coeff.multigridLevel[mgcg.numberOfMultigridLevels()-1].getIsACoefficientMatrix= %i \n",
  //        coeff.multigridLevel[mgcg.numberOfMultigridLevels()-1].getIsACoefficientMatrix());
  
  // supply coefficients to coarse grid solver
  if( parameters.useDirectSolverOnCoarseGrid )
  {
    if( mgcg.numberOfMultigridLevels()>1 )
      directSolver.setCoefficientArray( cMG.multigridLevel[mgcg.numberOfMultigridLevels()-1] );   
    else
      directSolver.setCoefficientArray( cMG );
  }
  
  if( parameters.problemIsSingular )
  {
    // parameters.setProblemIsSingular(TRUE);
    printF("*** problem is singular, directSolver.setCompatibilityConstraint \n");
    parameters.ogesParameters.set(OgesParameters::THEcompatibilityConstraint,TRUE);
  }
  // directSolver.initialize();                  // initialize oges (assigns classify array used below)

  tm[timeForInitialize]+=getCPU()-time0;
  
  if( debug & 4 ) printF("******time for buildCoefficientArray= %8.2e\n",getCPU()-time0);
  

//    if( true )
//    {
//      for( int l=0; l<numberOfExtraLevels; l++ )
//      {
//        int level=l;

//        // we need operators to apply boundary conditions.
//        RealCompositeGridFunction & cl = cMG.multigridLevel[level+1];
//        CompositeGrid & cgl = *cl.getCompositeGrid();
//        printF("+++++buildCoefficientArrays:END level+1=%i cgl.rcData=%i \n",level+1,cgl.rcData);
      
//      }
//    }
  
  return 0;
}





//\begin{>>OgmgInclude.tex}{\subsection{setCoefficientArray}}
int Ogmg::
setCoefficientArray( realCompositeGridFunction & coeff,
                     const IntegerArray & bc /* =Overture::nullIntArray() */ )
//==================================================================================
// /Description:
//    Supply the coefficient matrix, and optionally supply boundary conditions.
// /matrix (input) : a coefficient matrix defined on all levels.
//
// /bc(0:1,0:2,numberOfComponentGrids) (input): boundary conditions, Ogmg::dirichlet, neumann or mixed.
//    If this array is NOT supplied then the boundary conditions, dirichlet, neumann, mixed should be
//  given in cg[grid].boundaryCondition(side,axis)
//
// Ogmg::updateToMatchGrid should have already been called at this point so that
// the the multigrid-composite grid with extra levels has already been built.
//\end{OgmgInclude.tex} 
//==================================================================================
{
  real time0=getCPU();
  if( equationToSolve!=OgesParameters::userDefined )
    return 1;

  if( numberOfExtraLevels>0 )
  {
    Range all;
    int stencilSize=coeff[0].getLength(0);

    cMG.destroy();  // *wdh* 040830 -- this was needed by moveAndSolve --- ***** FIX THIS RIGHT ****
    
    cMG.updateToMatchGrid(mgcg,stencilSize,all,all,all);
    cMG.setIsACoefficientMatrix(TRUE,stencilSize); 
    
    cMG.multigridLevel[0].setOperators(*coeff.getOperators()); // *wdh* 011213
    
    if( debug & 4 )
    {
      printF("Ogmg::setCoefficientArray: cMG: numberOfComponentGrids=%i, numberOfGrids=%i, sizeOf=%12.0f\n",
	     cMG.numberOfComponentGrids(),cMG.numberOfGrids(),cMG.sizeOf());
      for( int ll=0; ll<mgcg.numberOfMultigridLevels(); ll++ )
      {
	printf("cMG.multigridLevel[%i].sizeOf()=%12.0f\n",ll,cMG.multigridLevel[ll].sizeOf());
      }
      printF("cMG.sizeOf()=%12.0f\n",cMG.sizeOf());
    }
    
    for( int grid=0; grid<coeff.numberOfComponentGrids(); grid++ )
    {
      if( FALSE )
       cMG[grid]=coeff[grid];  // copy existing levels  *** reference ??? *********** ********** fix this ****
      else
      {
	cMG[grid].reference(coeff[grid]);
	cMG.multigridLevel[0][grid].reference(cMG[grid]);  // need to do this too
      }
      
      cMG[grid].setOperators(*coeff[grid].getOperators());
      // cMG.multigridLevel[0][grid].setOperators(*coeff[grid].getOperators());
    }

  }
  else
  {
    cMG.reference(coeff);
  }

  // assign boundary conditions
  setBoundaryConditions( bc );

  // ::display(boundaryCondition,"setCoefficientArray: boundaryCondition after setBoundaryConditions( bc );");

  tm[timeForInitialize]+=getCPU()-time0;

  buildCoefficientArrays();
  
  return 0;
}


//\begin{>>OgmgInclude.tex}{\subsection{setBoundaryConditions}}
int Ogmg::
setBoundaryConditions(const IntegerArray & bc_,
		      const RealArray & bcData /* =Overture::nullRealArray() */ )
//==================================================================================
// /Description:
//    Set the boundary conditions. This is a private routine.
//
// This function assigns the internal arrays:
//
//   boundaryCondition(side,axis,grid) = extrapolate
//                                     = equation
//   bc(side,axis,grid) = dirichlet
//                      = neumann
//                      = mixed
// 
//\end{OgmgInclude.tex} 
//==================================================================================
{
  const int numberOfComponentGrids=mgcg.numberOfComponentGrids();
  
  boundaryCondition.redim(2,3,numberOfComponentGrids);
  boundaryCondition=0;
  bc.redim(2,3,numberOfComponentGrids);

  bool bcSupplied=false;
  if( bc_.getLength(0)>0 )
  {
    bcSupplied=true;
    bc=bc_;
  }

  bool bcDataSupplied=bcSupplied==true && bcData.getLength(0)>0;
  
  const int numData=2;
  boundaryConditionData.redim(numData,2,3,numberOfComponentGrids);
  int grid;
  for( grid=0; grid<numberOfComponentGrids; grid++ )
  {
    MappedGrid & mg = mgcg[grid];  

    for(int axis=axis1; axis<mg.numberOfDimensions(); axis++ )
    {
      for( int side=Start; side<=End; side++ )
      {
	if( !bcSupplied )
	{
	  bc(side,axis,grid)=mg.boundaryCondition(side,axis);
	}
	if( mg.boundaryCondition(side,axis) > 0 )
	{
	  if( bc(side,axis,grid)==OgmgParameters::dirichlet )
	  {
	    boundaryCondition(side,axis,grid)=OgmgParameters::extrapolate;  // default for now
            // mg.boundaryCondition()(side,axis)=bc(side,axis,grid);
	  }
	  else if( bc(side,axis,grid)==OgmgParameters::neumann || bc(side,axis,grid)==OgmgParameters::mixed )
	  {
	    boundaryCondition(side,axis,grid)=OgmgParameters::equation;
            // mg.boundaryCondition()(side,axis)=bc(side,axis,grid);
            if( bcDataSupplied )
	    {
	      if( bc(side,axis,grid)==OgmgParameters::neumann )
	      {
		boundaryConditionData(0,side,axis,grid)=0.;
		boundaryConditionData(1,side,axis,grid)=1.;
	      }
	      else
	      {
		boundaryConditionData(0,side,axis,grid)=bcData(0,side,axis,grid);
		boundaryConditionData(1,side,axis,grid)=bcData(1,side,axis,grid);

                if( Ogmg::debug & 2 )
		  printF("Ogmg::setBoundaryConditions: (side,axis,grid)=(%i,%i,%i) "
			 " mixed BC: a0,a1=%8.2e,%8.2e\n",side,axis,grid,bcData(0,side,axis,grid),
			 bcData(1,side,axis,grid)  );

		if( bcData(0,side,axis,grid)!=0. && bcData(1,side,axis,grid)==0. )
		{
		  // this is really a Dirichlet BC
                  if( Ogmg::debug & 4 )
  		    printF("Ogmg::setBoundaryConditions: (side,axis,grid)=(%i,%i,%i) "
			   " mixed BC is really dirichlet\n",side,axis,grid);
		
		  boundaryCondition(side,axis,grid)=OgmgParameters::extrapolate; 
                  bc(side,axis,grid)=OgmgParameters::dirichlet; // *wdh* 040831
		  
		  // mg.boundaryCondition()(side,axis)=OgesParameters::dirichlet;
		}
		else if( bcData(0,side,axis,grid)==0. && bcData(1,side,axis,grid)==0. )
		{
		  printF("Ogmg::setBoundaryConditions:ERROR: bcData==0 for a mixed bc, bc(%i,%i,%i)=%i\n",
			 side,axis,grid,bc(side,axis,grid));
		  Overture::abort();
		}
	      }
	    } // end if bcDataSupplied
	  }
	  else 
	  {
	    printF("Ogmg::setBoundaryConditions:ERROR: unknown bc(%i,%i,%i)=%i\n",
		   side,axis,grid,bc(side,axis,grid));
	    throw "error";
	  }
	}
	else
	{
          bc(side,axis,grid)=mg.boundaryCondition(side,axis);  // set to zero or -1 *wdh* 030605
	  
	  boundaryCondition(side,axis,grid)=0;
	}
      }
    }
  }



  return 0;
}



//\begin{>>OgmgInclude.tex}{\subsection{setOrderOfAccuracy}}
int Ogmg::
setOrderOfAccuracy(const int & orderOfAccuracy_)
//==================================================================================
// /Description:
//    Set the order of accuracy (2 or 4).
// /orderOfAccuracy\_ (input) : 
//
//\end{OgmgInclude.tex} 
//==================================================================================
{
  if( orderOfAccuracy_==2 || orderOfAccuracy_==4 )
  {
    orderOfAccuracy=orderOfAccuracy_;
    // no longer needed directSolver.setOrderOfAccuracy(orderOfAccuracy);
  }
  else
  {
    printF("Ogmg::setOrderOfAccuracy: ERROR: orderOfAccuracy must be 2 or 4\n");
    return 1;
  }
  return 0;
}


int Ogmg:: 
createNullVector()
// =====================================================================
// /Description:
//   Create the right null vector. This grid function is 1 at all interior
// and boundary points, but zero at interpolation and unused points.
// The vector is normalized to have l2 norm equal to one.
//
// =====================================================================
{
  rightNullVector.updateToMatchGrid(mgcg);
  Index I1,I2,I3;
  for( int l=0; l<mgcg.numberOfMultigridLevels(); l++ )  // often not needed on the coarsets level***
  {
    real norm=0.;
    int grid;
    for( grid=0; grid<mgcg.multigridLevel[l].numberOfComponentGrids(); grid++ )
    {
      MappedGrid & c = mgcg.multigridLevel[l][grid];
      realArray & nullVector = rightNullVector.multigridLevel[l][grid];
      const intArray & mask = c.mask();
      getIndex( c.gridIndexRange(),I1,I2,I3 );

      nullVector=0.;
      where( mask(I1,I2,I3) >0 )
        nullVector(I1,I2,I3)=1.;
      norm+=sum(nullVector);
    }
    if( norm==0. )
    {
      printF("Ogmg::createNullVector:WARNING: The norm of the null vector is zero on level=%i\n",l);
      norm=1.;
    }
    // normalize so that l2 norm is 1
    if( false )   // ********************************** turn this off *****
      rightNullVector.multigridLevel[l]/=sqrt(norm);
  }
  return 0;
}

real Ogmg::
rightNullVectorDotU( const int & level, const RealCompositeGridFunction & u )
// =====================================================================
// /Description:
//     Determine the inner product of the right null vector times u
//       ccValue =   (rightNullVector, u ) 
// /level (input) : right null vector and u live on this level.
// =====================================================================
{
  RealCompositeGridFunction & nullVector = rightNullVector.multigridLevel[level];
  real ccValue=0.;
  for( int grid=0; grid<mgcg.multigridLevel[level].numberOfComponentGrids(); grid++ )
    ccValue+=sum( nullVector[grid]*u[grid] );

  return ccValue;
}




//\begin{>>OgmgInclude.tex}{\subsection{interpolate}}
int Ogmg::
interpolate(realCompositeGridFunction & u, const int & grid /* =-1 */, int level /* =-1 */ )
// =====================================================================================
// /Description:
//
// Interpolate here so we can keep track of the cpu time used.
//
// /grid (input) : interpolate this grid only (if possible)

//\end{OgmgInclude.tex} 
//==================================================================================
{
  real time=getCPU();


  if( parameters.maximumNumberOfInterpolationIterations>0 )
    u.getInterpolant()->setMaximumNumberOfIterations(parameters.maximumNumberOfInterpolationIterations);

  if( !parameters.decoupleCoarseGridEquations || level==0 )
  {
    if( grid==-1 )
      u.interpolate();  // interpolate all grids
    else
      u.getInterpolant()->interpolate(grid,u);
    // u.display("u after interpolate",debugFile,"%5.1f ");
    
  }
  else 
  {
    // this is for testing the ICMG algorithm where we do not interpolate coarser levels.
    CompositeGrid & cg = *u.getCompositeGrid();
    const int gridStart = grid==-1 ? 0 : grid;
    const int gridEnd   = grid==-1 ? cg.numberOfComponentGrids()-1 : grid;
    for( int gg=gridStart; gg<=gridEnd; gg++ )
    {
      realArray & uu = u[gg];
      const intArray & ip = cg.interpolationPoint[gg]; 
      Range R=cg.numberOfInterpolationPoints(gg);
      if( R.getLength() > 0 )
      {
	if( cg.numberOfDimensions()==2 )
	  uu(ip(R,0),ip(R,1),0)=0.;
	else
	  uu(ip(R,0),ip(R,1),ip(R,2))=0.;
      }
    }
  }
  
  
  tm[timeForInterpolation]+=getCPU()-time;
  return 0;
}




//\begin{>>OgmgInclude.tex}{\subsection{update}}
int Ogmg::
update( GenericGraphicsInterface & gi )
// =====================================================================================
// /Description:
//   Update parameters interactively.
//\end{OgmgInclude.tex} 
//==================================================================================
{
  parameters.update(gi,mgcg);
  directSolver.setOgesParameters(parameters.ogesParameters);
  return 0;
}

//\begin{>>OgmgInclude.tex}{\subsection{update}}
int Ogmg::
update( GenericGraphicsInterface & gi, CompositeGrid & cg  )
// =====================================================================================
// /Description:
//   Update parameters interactively. Use this update if you have not already given a
// CompositeGrid to Ogmg (through a constructor or with the updateToMatchGrid function).
//
//\end{OgmgInclude.tex} 
//==================================================================================
{
  if( cg.numberOfMultigridLevels()> mgcg.numberOfMultigridLevels() ||
      cg.numberOfComponentGrids()!= mgcg.numberOfComponentGrids() )
    parameters.update(gi,cg);
  directSolver.setOgesParameters(parameters.ogesParameters);
  return 0;
}

