#include "Ogmg.h"
#include "Oges.h"
#include "SparseRep.h"
#include "Integrate.h"
#include "HDF_DataBase.h"

int Ogmg::
getSingularParameter(int level)
// ======================================================================================
// /Description:
//   Return some values for a singular problem
// ======================================================================================
{
  if( alpha.getLength(0)<mgcg.numberOfMultigridLevels() )
  {
    alpha.redim(mgcg.numberOfMultigridLevels());
    alpha=0.;
  }
  
  CompositeGrid & cg = mgcg.multigridLevel[level];
  realCompositeGridFunction & f = fMG.multigridLevel[level];
  realCompositeGridFunction & leftNull = (*leftNullVector).multigridLevel[level];
  realCompositeGridFunction & rightNull = rightNullVector.multigridLevel[level];
  Index I1,I2,I3;
  alpha(level)=0.;
  real leftDotRight=0.;
  for( int grid=0; grid<cg.numberOfComponentGrids(); grid++ )
  {
    getIndex(cg[grid].dimension(),I1,I2,I3);
    alpha(level)+=sum(leftNull[grid](I1,I2,I3)*f[grid](I1,I2,I3));
    leftDotRight+=sum(leftNull[grid](I1,I2,I3)*rightNull[grid](I1,I2,I3));  // could scale so this is 1.
  }
  alpha(level)/=leftDotRight;

  printF("%*.1s **Compatibility value: level=%i alpha = l.f/l.l = %8.2e **\n",level*4," ",level,alpha(level));

  return 0;
}

int Ogmg::
saveLeftNullVector()
// ======================================================================================
// /Description:
//    Save the left-null-vector in a file.
// ======================================================================================
{
  assert( leftNullVector!=NULL );
  
  HDF_DataBase dataFile;
  dataFile.mount(parameters.nullVectorFileName,"I");

  int streamMode=0;  // 1=save in compressed form.
  dataFile.put(streamMode,"streamMode");
  if( !streamMode )
    dataFile.setMode(GenericDataBase::noStreamMode); // this is now the default
  else
  {
    dataFile.setMode(GenericDataBase::normalMode); // need to reset if in noStreamMode
  }

  // *** trouble using this next line ***
  //  leftNullVector->put(dataFile,"leftNullVector");

  int numLevels = parameters.useDirectSolverOnCoarseGrid ? mgcg.numberOfMultigridLevels()-1 : 
    mgcg.numberOfMultigridLevels();
  
  for( int level=0; level<numLevels; level++ )  
  {
    RealCompositeGridFunction & nullVector= (*leftNullVector).multigridLevel[level];

    nullVector.put(dataFile,sPrintF("leftNullVectorLevel%i",level));
  }
  

  dataFile.unmount();

  if( false && Ogmg::debug & 1 )
    printF("Ogmg::saveLeftNullVector:left null vector was saved in the file [%s]\n",
	   (const char*)parameters.nullVectorFileName);

  return 0;
}

int Ogmg::
readLeftNullVector()
// ======================================================================================
// /Description:
//    Attempt to read the left null vector from a file.
//
// /return value: 0=success, 1=file was not found
// ======================================================================================
{
  if( !(parameters.nullVectorOption==OgmgParameters::readOrComputeNullVector ||
        parameters.nullVectorOption==OgmgParameters::readOrComputeAndSaveNullVector) )
  {
    // we cannot read the null vector in this case
    return 1;  
  }
  
  if( leftNullVector==NULL )
    leftNullVector=new realCompositeGridFunction(mgcg);
  
  if( leftNullVectorIsComputed ) return 0;  // null vector has already been read or computed directly


  FILE *file = fopen(parameters.nullVectorFileName,"r");
  if( file==0 )
  {
    return 1;  // file was not found
  }
  fclose(file);
   
  HDF_DataBase dataFile;

  bool found = dataFile.mount(parameters.nullVectorFileName, "R")==0;
  assert( found );

  // check to see if this file was saved in streamMode (compressed)
  int streamMode=true; // use this as a default since this is the old way.
  dataFile.get(streamMode,"streamMode");

  if( !streamMode )
    dataFile.setMode(GenericDataBase::noStreamMode);

  // *** trouble using this next line ***
  // leftNullVector->get(dataFile,"leftNullVector");

  int numLevels = parameters.useDirectSolverOnCoarseGrid ? mgcg.numberOfMultigridLevels()-1 : 
    mgcg.numberOfMultigridLevels();
  
  for( int level=0; level<numLevels; level++ )  
  {
    RealCompositeGridFunction & nullVector= (*leftNullVector).multigridLevel[level];

    nullVector.get(dataFile,sPrintF("leftNullVectorLevel%i",level));
  }

  dataFile.unmount();

  if( Ogmg::debug & 1 ) printF("Ogmg::readLeftNullVector:left null vector was successfully read from a file\n");

  if( false )
  {
    for( int level=0; level<numLevels; level++ )  
    {
      RealCompositeGridFunction & nullVector= (*leftNullVector).multigridLevel[level];

      nullVector.display(sPrintF("Here is the leftNullVector level=%i",level),"%5.1e ");
    }
  }
  
  leftNullVectorIsComputed=true;

  return 0;
}



int Ogmg::
computeLeftNullVector()
// ======================================================================================
// /Description:
//    Compute the left-null-vector for a singular problem
// ======================================================================================
{
  if( leftNullVector==NULL )
    leftNullVector=new realCompositeGridFunction(mgcg);

  // attempt to read the null vector from a file
  bool found = readLeftNullVector()==0;
  if( found ) 
  {
    leftNullVectorIsComputed=true;
    return 0;
  }

  int numLevels = parameters.useDirectSolverOnCoarseGrid ? mgcg.numberOfMultigridLevels()-1 : 
              mgcg.numberOfMultigridLevels();
  
  for( int level=0; level<numLevels; level++ )  
  {
    Oges solver;
    if( parameters.nullVectorParameters!=NULL )
    {
      solver.setOgesParameters(*parameters.nullVectorParameters);
    }
     CompositeGrid & cg = mgcg.multigridLevel[level];
     RealCompositeGridFunction & coeff = cMG.multigridLevel[level];
     RealCompositeGridFunction & nullVector= (*leftNullVector).multigridLevel[level];

     if( false  )
     {
      Integrate integrate(cg);
      RealCompositeGridFunction & leftNull = integrate.leftNullVector();
      nullVector=leftNull;
      
     }
     else
     {
       RealCompositeGridFunction f(cg);   // **** could probablu use some other work space

       const int numberOfDimensions = cg.numberOfDimensions();
       int grid;

       Range all;

       // *** we need to finish BC's --- could do this elsewhere
//      int stencilSize=10, orderOfAccuracy=2;
//      CompositeGridOperators op(cg);                            // create some differential operators
//      op.setStencilSize(stencilSize);
//      op.setOrderOfAccuracy(orderOfAccuracy);
//      coeff.setOperators(op);

//      op.finishBoundaryConditions(coeff);

//      coeff.finishBoundaryConditions();
     
       solver.updateToMatchGrid(cg);
  
       // count the total number of grid points.
       int numberOfGridPoints=0;
       for( grid=0; grid<cg.numberOfComponentGrids(); grid++ )
	 numberOfGridPoints+=cg[grid].mask().elementCount();

       Index I1,I2,I3;
       bool useIterativeSolver= false || numberOfDimensions==3;
       if( useIterativeSolver )
       {
	 solver.set(OgesParameters::THEbestIterativeSolver);
	 // solver.set(OgesParameters::THEpreconditioner,OgesParameters::incompleteLUPreconditioner);
	 // solver.setConjugateGradientPreconditioner(Oges::diagonal);
	 // Oges::debug=7;
	 solver.set(OgesParameters::THEtolerance,REAL_EPSILON*numberOfGridPoints);
       }    

       // solver.set(OgesParameters::THErescaleRowNorms,false);  // turn this off for debugging

       solver.set(OgesParameters::THEsolveForTranspose,true); // solve the transpose system (we want the left null vector)
       solver.set(OgesParameters::THEfixupRightHandSide,false);     // no need to zero out equations at special points.

       bool solveSingularProblem=true;


     // assign the rhs: f=0 except for the rhs to the compatibility equation which we set to the numberOfGridPoints.
     // (this will cause the sum of the interior weights to be numberOfGridPoints so that each value should be about 1)
       f=0.;
       solver.setCoefficientArray( coeff );     // supply coefficients
       if( solveSingularProblem )
	 solver.set(OgesParameters::THEcompatibilityConstraint,true); // system is singular so add an extra equation
       solver.initialize();

       // find the equation where the compatibility constraint is put (some unused point)
       int n,i1e,i2e,i3e,gride;
       solver.equationToIndex(solver.extraEquationNumber(0),n,i1e,i2e,i3e,gride);
       f[gride](i1e,i2e,i3e,n)=numberOfGridPoints;

       printF(" ***Extra equation: n,i1e,i2e,i3e,gride=%i %i %i %i %i \n",n,i1e,i2e,i3e,gride);
       
     
       // real value=numberOfGridPoints;
       // solver.setExtraEquationValues( f,&value );  // new way

       // nullVector.updateToMatchGrid(cg);
       if( solver.isSolverIterative() )
	 printF(" $$$$$$$$$$$$ iterative solver for null vector $$$$$$$$$$$$\n");
       if( true || solver.isSolverIterative() )
       {
	 
	 nullVector=1.;
	 if( false )
	 {
	   for( grid=0; grid<cg.numberOfComponentGrids(); grid++ )
	   {
	     MappedGrid & c = cg[grid];
	     const IntegerDistributedArray & classify = coeff[grid].sparse->classify;

	     getIndex(c.dimension(),I1,I2,I3);
	     where( classify(I1,I2,I3)!=(int)SparseRepForMGF::interior && 
		    classify(I1,I2,I3)!=(int)SparseRepForMGF::boundary &&
		    classify(I1,I2,I3)!=(int)SparseRepForMGF::ghost1 )
	     {
	       nullVector[grid](I1,I2,I3)=0.;
	     }
	   }
	 }
	 
       }
       
	 
       // coeff.display("coeff");

       // Oges::debug=63;
     
       printf("**** Solve for leftNullVector on level=%i *******\n",level);
       real time=getCPU();
       solver.solve( nullVector,f );   // solve for the (unscaled weights)
       time=getCPU()-time;
       printf("**** ...done cpu=%8.2e (s)\n",time);

       if( Ogmg::debug & 1 && solver.isSolverIterative() )
       {
	 real absoluteTolerance,relativeTolerance;
	 solver.get(OgesParameters::THEabsoluteTolerance,absoluteTolerance);
	 solver.get(OgesParameters::THErelativeTolerance,relativeTolerance);
	 real maxResidual=solver.getMaximumResidual();
    
	 printF("Ogmg::computeLeftNullVector: iter's to solve = %i (max res=%8.2e "
		"rel-tol=%7.1e, abs-tol=%7.1e)\n",
		solver.getNumberOfIterations(),maxResidual,relativeTolerance,absoluteTolerance);
       }
       
     // zero out the nullVector at unused points. 
     // nullVector values are non-zero at interpolation points -- the RHS should be zero at these
     // points anyway.
       for( grid=0; grid<cg.numberOfComponentGrids(); grid++ )
       {
	 MappedGrid & c = cg[grid];
	 const IntegerDistributedArray & classify = coeff[grid].sparse->classify;

	 getIndex(c.dimension(),I1,I2,I3);
	 where( classify(I1,I2,I3)!=(int)SparseRepForMGF::interior && 
		classify(I1,I2,I3)!=(int)SparseRepForMGF::boundary &&
		classify(I1,I2,I3)!=(int)SparseRepForMGF::ghost1 )
	 {
	   nullVector[grid](I1,I2,I3)=0.;
	 }
       }
       if( false )
	 nullVector.display(sPrintF("Here is the leftNullVector level=%i",level),"%5.1e ");
     
     }
     
  }

  if( parameters.nullVectorOption==OgmgParameters::computeAndSaveNullVector ||
      parameters.nullVectorOption==OgmgParameters::readOrComputeAndSaveNullVector )
  {
    // save the null vector in a file (so it can be used in future computations)
    saveLeftNullVector();
  }

  leftNullVectorIsComputed=true;

  return 0;
}


//! Adjust the right-hand-side for a singular problem
int Ogmg::
addAdjustmentForSingularProblem(int level, int iteration )
{

  if( !parameters.problemIsSingular || !parameters.projectRightHandSideForSingularProblem )
    return 0;

  // fMG.multigridLevel[level]+=alpha(level)*rightNullVector.multigridLevel[level];

  if( true )
  {
    // determine alpha for the equation  
    //           L u + alpha*r = f
    //           r.u = g

    if( parameters.useDirectSolverOnCoarseGrid && level==mgcg.numberOfMultigridLevels()-1 )
      return 0;


    if( !leftNullVectorIsComputed )
    {
      computeLeftNullVector();
    }

    if( level>0 || iteration==0 )
    {
      getSingularParameter(level);
        
      fMG.multigridLevel[level]-=alpha(level)*rightNullVector.multigridLevel[level];

	
    }
  }
  else
  {
    // Here we play around with different ways to compute alpha      

//     if( !( (level==0 && (debug & 2)) || debug & 4) )
//     {
//       defect(level);   
//       defectOld= max(fabs(defectMG.multigridLevel[level]));
//     }
      

    // real alphaOld=0.;
//     if( true || defectOld<.1 )
//     {
//       // alphaOld=alpha(level);
//       alpha(level)=rightNullVectorDotU( level, defectMG.multigridLevel[level] );
//     }
    
//     real rDotU = rightNullVectorDotU( level, uMG.multigridLevel[level] );
//     printf(" *** level=%i : defectOld=%8.2e, alpha=%8.2e, rDotU=%8.2e \n",level,defectOld,alpha(level),rDotU);
    
//  const real beta=1.;
//  alpha(level)-=rDotU*beta;

    // fMG.multigridLevel[level]+=1.*rightNullVector.multigridLevel[level]; // shift for testing
    
    if( iteration==0 )
    {
      for( int it=0; it<200; it++ )
      {

	smooth(level,parameters.numberOfSmooths(0,level),it);

	real rDotU=parameters.meanValueForSingularProblem-rightNullVectorDotU( level, uMG.multigridLevel[level] );
	real rDotV=rightNullVectorDotU(level,*v);
	real beta = rDotU/rDotV;
    
	alpha(level)+=beta;
	uMG.multigridLevel[level]+=beta*(*v);
    
      
	fMG.multigridLevel[level]-=beta*rightNullVector.multigridLevel[level];

	if( it % 5 == 0 )
	  printf(" >>>>level=%i: it=%i: rDotU=%8.2e, rDotV=%8.2e, beta=%8.2e, alpha=%8.2e \n",level,
		 it,rDotU,rDotV,beta,alpha(level));

      }
      uMG.multigridLevel[level]=0.;
      
    }

//     real rDotU=parameters.meanValueForSingularProblem-rightNullVectorDotU( level, uMG.multigridLevel[level] );
//     real rDotV=rightNullVectorDotU(level,*v);
//     real beta = rDotU/rDotV;
    
//     alpha(level)+=beta;
//     uMG.multigridLevel[level]+=beta*(*v);
    
      
//     fMG.multigridLevel[level]-=beta*rightNullVector.multigridLevel[level];

//     printf(" >>>>level=%i: it=%i: rDotU=%8.2e, rDotV=%8.2e, beta=%8.2e, alpha=%8.2e \n",level,
// 	   iteration,rDotU,rDotV,beta,alpha(level));
      


    // fMG.multigridLevel[level]-=alpha(level)*rightNullVector.multigridLevel[level];

//     rDotU=parameters.meanValueForSingularProblem-rightNullVectorDotU( level, uMG.multigridLevel[level] );
//     printf(">>>>AFTER:  rDotU=%8.2e (=0 ?) \n", rDotU);
      
  }

  return 0;
}


//! Remove the adjustment to the right-handside for a singular problem
int Ogmg::
removeAdjustmentForSingularProblem(int level, int iteration )
{
   if( !parameters.problemIsSingular )
     return 0;
   
   if( level==0 ) 
   {
     // reset the right hand side
     if( alpha(level) != 0. )
       fMG.multigridLevel[level]+=alpha(level)*rightNullVector.multigridLevel[level];
   }

  return 0;
}

