#include "Overture.h"
#include "ParallelUtility.h"
#include "ParallelGridUtility.h"
#include "Ogen.h"

// Macro to extract a local array with ghost boundaries
//  type = int/float/double/real
//  xd = distributed array
//  xs = serial array 
#ifdef USE_PPP
 #define GET_LOCAL(type,xd,xs)\
    type ## SerialArray xs; getLocalArrayWithGhostBoundaries(xd,xs)
 #define GET_LOCAL_CONST(type,xd,xs)\
    type ## SerialArray xs; getLocalArrayWithGhostBoundaries(xd,xs)
#else
 #define GET_LOCAL(type,xd,xs)\
    type ## SerialArray & xs = xd
 #define GET_LOCAL_CONST(type,xd,xs)\
    const type ## SerialArray & xs = xd
#endif

#define ForBoundary(side,axis)   for( int axis=0; axis<mg.numberOfDimensions(); axis++ ) \
                                 for( int side=0; side<=1; side++ )

#define  FOR_3(i1,i2,i3,I1,I2,I3)\
  I1Base=I1.getBase(); I2Base=I2.getBase(); I3Base=I3.getBase();\
  I1Bound=I1.getBound(); I2Bound=I2.getBound(); I3Bound=I3.getBound();\
  for( i3=I3Base; i3<=I3Bound; i3++ )  \
  for( i2=I2Base; i2<=I2Bound; i2++ )  \
  for( i1=I1Base; i1<=I1Bound; i1++ )

#define  FOR_3D(i1,i2,i3,I1,I2,I3)\
  int I1Base,I2Base,I3Base;\
  int I1Bound,I2Bound,I3Bound;\
  I1Base=I1.getBase(); I2Base=I2.getBase(); I3Base=I3.getBase();\
  I1Bound=I1.getBound(); I2Bound=I2.getBound(); I3Bound=I3.getBound();\
  for( i3=I3Base; i3<=I3Bound; i3++ )  \
  for( i2=I2Base; i2<=I2Bound; i2++ )  \
  for( i1=I1Base; i1<=I1Bound; i1++ )

int 
checkOverlappingGrid( const CompositeGrid & cg, const int & option /* =0 */, bool onlyCheckBaseGrids /* =true */ )
// ===================================================================================
// Check the validity of the Overlapping Grid
//
//   o check interpolation points
//     check for valid explicit or implicit interpolation
//
// /option (input) : a bit flag to turn out various options and outputs:
//       2 : check that discretization points are valid 
//       4 : print a list of interpolation points and where they interpolate from
//      64 : print the mask
// /return value: 0 if no errors where found, otherwise the number of errors found.
//
// /onlyCheckBaseGrids (input) : if true only check interpolation points on base grids. If false, also
//   check interpolation points on AMR grids.
// ===================================================================================
{
  int ok=0;
  const int myid=max(0,Communication_Manager::My_Process_Number);
  
//   #ifdef USE_PPP
//       // Copy the grid and all interpolation data to a single processor:
//     CompositeGrid cg;
// //     ParallelGridUtility::redistribute( cg_, cg, Range(0,0) );
//   #else
//     const CompositeGrid & cg = cg_;
//   #endif
  
  int iv[3]={0,0,0}, &i1=iv[0], &i2=iv[1], &i3=iv[2];
  int iiv[3]={0,0,0};  // holds lower corner of interpolation stencil
  Index Iv[3], &I1=Iv[0], &I2=Iv[1], &I3=Iv[2];
  
  const int numberOfGrids = onlyCheckBaseGrids ? cg.numberOfBaseGrids() : cg.numberOfComponentGrids();

  realSerialArray rr(1,3); rr=0.;
  intSerialArray interpolates(1), useBackupRules(1);
  bool checkForOneSided=TRUE;
  useBackupRules=FALSE;

  int l;   // multigrid level
  const int numberOfDimensions = cg.numberOfDimensions();
  int grid,axis;
  for( l=0; l<cg.numberOfMultigridLevels(); l++ ) 
  {
    const CompositeGrid & cgl = cg.numberOfMultigridLevels()==1 ? cg : cg.multigridLevel[l];
    
    // compute the allowable tolerance for interpolating near the boundary of a grid:
    Range all;
    RealArray epsilonForInterpolation(numberOfDimensions,numberOfGrids);
    for( grid=0; grid<numberOfGrids; grid++ )
      for( int axis=0; axis<numberOfDimensions; axis++ )
	epsilonForInterpolation(axis,grid)=max(cg.epsilon()*2.1/cg[grid].gridSpacing(axis),
					       max(cg[grid].sharedBoundaryTolerance()(all,axis)));
    

    for( grid=0; grid<numberOfGrids; grid++ )
    {
      MappedGrid & mg = (MappedGrid&)cgl[grid]; // cast away const
      const intArray & maskgd = mg.mask();
      GET_LOCAL_CONST(int,maskgd,maskg);

      if( option & 64 )
	mg.mask().display("checkOverlappingGrid: Here is the mask");
      // printf("CompositeGrid::ISusedPoint=%i, CompositeGrid::ISdiscretizationPoint=%i \n",
      //	   CompositeGrid::ISusedPoint,CompositeGrid::ISdiscretizationPoint);
    
      // *wdh* 070828 -- check that points on interpolation boundaries are unused or interpolation points
      const IntegerArray & eir = mg.extendedIndexRange(); // extendedGridIndexRange(mg);
      ForBoundary(side,axis)
      {
	if( mg.boundaryCondition(side,axis)==0 )
	{
	  getBoundaryIndex(eir,side,axis,I1,I2,I3);
	  // the number of interpolation layers is the half the discrization width
          const int numInterpLayers = (mg.discretizationWidth(axis)-1)/2;
          int ia=eir(side,axis);
	  int ib = ia + (numInterpLayers -1 )*(1-2*side);
          Iv[axis] = side==0 ? Range(ia,ib) : Range(ib,ia);

	  bool hasPoints=ParallelUtility::getLocalArrayBounds(maskgd,maskg,I1,I2,I3);  
	  if( !hasPoints ) continue; 

	  FOR_3D(i1,i2,i3,I1,I2,I3)
	  {
	    if( maskg(i1,i2,i3)>0 )
	    {
              printf("checkOverlappingGrid:ERROR: grid=%i (i1,i2,i3)=(%i,%i,%i) near the interp. "
                     "boundary (side,axis)=(%i,%i) is NOT valid, mask>0 !\n",grid,i1,i2,i3,side,axis);
	      ok++;
	      
	    }
	  }
	}
      }
      

      #ifdef USE_PPP
       const intSerialArray & numberOfInterpolationPoints = cgl->numberOfInterpolationPointsLocal;
       const intSerialArray & ip = cgl->interpolationPointLocal[grid];
       const intSerialArray & il = cgl->interpoleeLocationLocal[grid];
       const intSerialArray & ig = cgl->interpoleeGridLocal[grid];
       const realSerialArray & ci = cgl->interpolationCoordinatesLocal[grid];
       const intSerialArray & viw = cgl->variableInterpolationWidthLocal[grid];
      #else
       const intSerialArray & numberOfInterpolationPoints = cgl.numberOfInterpolationPoints;
       const intSerialArray & ip = cgl.interpolationPoint[grid];
       const intSerialArray & il = cgl.interpoleeLocation[grid];
       const intSerialArray & ig = cgl.interpoleeGrid[grid];
       const realSerialArray & ci = cgl.interpolationCoordinates[grid];
       const intSerialArray & viw = cgl.variableInterpolationWidth[grid];
      #endif

      iv[2]=0;
      for( int i=0; i<numberOfInterpolationPoints(grid); i++ )
      {
	for( int axis=0; axis<numberOfDimensions; axis++ )
	{
	  iv[axis]=ip(i,axis);
	  iiv[axis]=il(i,axis);
	}
	// here is the donor grid:
        
	int gridi = ig(i);
	assert( gridi>=0 && gridi<numberOfGrids );
	
	MappedGrid & gi = (MappedGrid&)cgl[gridi];
	const intArray & maskd = gi.mask();
    
	GET_LOCAL_CONST(int,maskd,mask);
	

	if( !(maskg(iv[0],iv[1],iv[2]) & CompositeGrid::ISinterpolationPoint) )
	{
	  cout << " checkOverlappingGrid: ERROR mask wrong at an interpolation point" << endl;
	  printf(" -> grid=%i, donor=%i, interp pt i=%i iv=[%i,%i,%i] mask=%i\n",
		 grid,gridi,i,iv[0],iv[1],iv[2],maskg(iv[0],iv[1],iv[2]));
	  
	  ok++;
	}

	// width(Range(0,2))=cgl.interpolationWidth(Range(0,2),grid,gridi);  
	const int width = viw(i);

	if( option & 4 )
	{
	  printf("grid=%i, i=%i, p=(%i,%i,%i) from (%i,%i,%i) on grid %i, width=%i\n",
		 grid,i,iv[0],iv[1],iv[2],iiv[0],iiv[1],iiv[2],gridi,width);
	}
	

	for( int axis=axis1; axis<numberOfDimensions; axis++ ) 
	{
	  real rsb=ci(i,axis)/gi.gridSpacing(axis)+gi.indexRange(Start,axis);
	  real px= gi.isCellCentered(axis)  ? rsb-iiv[axis]-.5 : rsb-iiv[axis];

	  if( fabs( px - 0.5*( width-1)) >  0.5 + REAL_EPSILON*10. )
	  {
	    // Interpolation weight is too large, this could be an error
	    //..............okay if extrapolating near a boundary
	    //..............okay if we are exactly on the end points. (for refinement grids, 981030)
	    real pxmax=gi.isCellCentered(axis)  ? .5 : 0.; 
//	    const real eps=REAL_EPSILON*max(500.,100./gi.gridSpacing(axis));
//	    const real eps=cg.epsilon()*2.1/gi.gridSpacing(axis);
	    const real eps=epsilonForInterpolation(axis,gridi);
	    if( ! (
		  ( ( il(i,axis) < gi.indexRange(Start,axis)+width ) ||
		    ( il(i,axis) > gi.indexRange(End  ,axis)-width )  
		    ) &&
		  (px >= -eps-pxmax && px <= width-1.+eps+pxmax ) 
		  ) 
		&& fabs(px)>REAL_EPSILON*10. && fabs(px-width+1.)>REAL_EPSILON*10. )
	    {
	      ok++;
	      printf( " checkOG:ERROR: Invalid interp pt (%i,%i,%i) grid %i (%s) from (%i,%i,%i) grid %i (%s)\n"
		      "    r=(%6.2e,%6.2e,%6.2e)  interp weight =%7.3e (not in [%7.1e,%7.1e]) (eps=%6.3e) width=%i\n",
		      iv[0],iv[1],iv[2],grid,(const char*)mg.getName(),
		      il(i,axis1),il(i,axis2),(numberOfDimensions==3 ? il(i,axis3) : 0),
		      gridi,(const char*)gi.getName(),
		      ci(i,axis1),ci(i,axis2),(numberOfDimensions==3 ? ci(i,axis3) : 0.),
		      px,0.5*( width-1)-.5,0.5*( width-1)+.5,eps,width);
	    }
	  }
	}
	const int validMask = cgl.interpolationIsImplicit(grid,gridi) ? CompositeGrid::ISusedPoint
	  : CompositeGrid::ISdiscretizationPoint;
	bool pointOk=TRUE;
	const int width2 = numberOfDimensions>1 ? width : 1;
	const int width3 = numberOfDimensions>2 ? width : 1;

#ifndef USE_PPP	
        // ---- serial check ----
	for( int w3=0; w3<width3 && pointOk; w3++ )    
	{
	  for( int w2=0; w2<width2 && pointOk; w2++ )    
	  {
	    for( int w1=0; w1<width; w1++ )    
	    {
	      const int & m = mask(iiv[0]+w1,iiv[1]+w2,iiv[2]+w3);
	      if( !(m & validMask) )
	      {
		pointOk=FALSE;
		printf("checkOverlappingGrid:ERROR: Some of the donor points cannot be used for interpolation\n");
		printf(" grid=%s (%i), point %i = (%i,%i,%i), donor grid=%s (%i), donor=(%i,%i,%i), "
		       "offset=(%i,%i,%i), width=%i, mask=%i, "
		       "isImplicit=%i, mask&discretization=%i, mask&interpolation=%i mask&ghost=%i\n",
		       (const char*)mg.getName(),grid,
		       i,iv[0],iv[1],iv[2],
		       (const char*)gi.getName(),gridi,
		       iiv[0]+w1,iiv[1]+w2,iiv[2]+w3,w1,w2,w3,width,m,
		       cgl.interpolationIsImplicit(grid,gridi),m & CompositeGrid::ISdiscretizationPoint,
		       m & CompositeGrid::ISinterpolationPoint, m & CompositeGrid::ISghostPoint );

                //              printf("Here are the bits of the mask[31...0]=");
                // 		for( int b=31; b>=0; b-- )
                // 		  printf("%i",m & (1<<b) ? 1 : 0);
                // 		printf("\n");

		for( int axis=axis1; axis<numberOfDimensions; axis++ ) 
		  rr(0,axis)=ci(i,axis);
		interpolates=TRUE;  useBackupRules=FALSE;
		cg.rcData->canInterpolate(cg.gridNumber(grid),cg.gridNumber(gridi), rr, interpolates,
					  useBackupRules, checkForOneSided );
 		printf(" canInterpolate: interpolate=%i, ISdiscretizationPoint=%i "
 		       "ISinterpolationPoint=%i\n",interpolates(0),
 		       MappedGrid::ISdiscretizationPoint,CompositeGrid::ISinterpolationPoint);


// #ifndef USE_PPP
// 		cg.rcData->canInterpolate(grid,gridi, rr, interpolates, useBackupRules, checkForOneSided );
// #endif

		ok++;
		break;
	      }
	    }
	  }
	}
#endif
      } // end for i 

#ifdef USE_PPP
      
      // --- parallel version : double check interpolation ---
      IntegerArray interpolates, useBackupRules;
      RealArray rr;
      const IntegerArray & ise = cg.interpolationStartEndIndex;
      for( int gridi=0; gridi<numberOfGrids; gridi++ )
      {
	if( grid==gridi ) continue;
	
	MappedGrid & gi = (MappedGrid&)cgl[gridi];
	
	// printf("checkOverlappingGrid: myid=%i grid=%i gridi=%i interpolationStartEndIndex()=%i %i \n",
	//        myid,grid,gridi,
	//        ise(0,grid,gridi),
	//        ise(1,grid,gridi));
	const int numi = ise(0,grid,gridi)>=0 ? ise(1,grid,gridi)-ise(0,grid,gridi)+1 : 0;

	if( numi>0 )
	{
	  Range R=numi;
	  interpolates.redim(R);  interpolates=false;
	  useBackupRules.redim(R); useBackupRules=true;
          rr.redim(R,numberOfDimensions); 
          Range Ri=Range(ise(0,grid,gridi),ise(1,grid,gridi));
          Range Rx=numberOfDimensions;
          rr=ci(Ri,Rx);
	}
	else
	{
	  interpolates.redim(0);
	  useBackupRules.redim(0);
	  rr.redim(0);
	}

	Ogen::checkCanInterpolate((CompositeGrid&)cg,cg.gridNumber(grid),cg.gridNumber(gridi), rr, interpolates,
				  useBackupRules);
	for( int j=0; j<numi; j++ )
	{
	  if( !interpolates(j) )
	  {
            const int i=ise(0,grid,gridi)+j;
	    for( int axis=0; axis<numberOfDimensions; axis++ )
	    {
	      iv[axis]=ip(i,axis);
	      iiv[axis]=il(i,axis);
	    }
	    const int width = viw(i);
	    printf("checkOverlappingGrid:ERROR: Unable to interpolate a point, myid=%i\n",myid);
	    printf(" grid=%s (%i), point %i = (%i,%i,%i), donor grid=%s (%i), donor=(%i,%i,%i),"
		   " width=%i, isImplicit=%i.\n",
		   (const char*)mg.getName(),grid,
		   i,iv[0],iv[1],iv[2],
		   (const char*)gi.getName(),gridi,
		   width,cgl.interpolationIsImplicit(grid,gridi));
	    ok++;
	  }
	  else
	  {
	    if( false )
	    {
	      const int i=ise(0,grid,gridi)+j;
	      printf(" checkOverlappingGrid: myid=%i, i=%i grid=%i gridi=%i interpolates=%i\n",myid,i,
		     grid,gridi,interpolates(j));
	    }
	  }
	}
      }
      
#endif
      if( option & 2 )
      {
        // check that discretization points are valid

	getIndex(mg.gridIndexRange(),I1,I2,I3);
	bool hasPoints=ParallelUtility::getLocalArrayBounds(maskgd,maskg,I1,I2,I3);  
	if( hasPoints ) 
	{
          // NOTE: it is very slow to check every pt so only check most likely problem pts.
	  if( numberOfDimensions==2 )
	  {
	    FOR_3D(i1,i2,i3,I1,I2,I3)
	    {
	      if( maskg(i1,i2,i3)>0 && 
		  (maskg(i1-1,i2-1,i3)<=0 || maskg(i1  ,i2-1,i3)<=0 || maskg(i1+1,i2-1,i3)<=0 ||
		   maskg(i1-1,i2  ,i3)<=0 ||                           maskg(i1+1,i2  ,i3)<=0 ||
		   maskg(i1-1,i2+1,i3)<=0 || maskg(i1  ,i2+1,i3)<=0 || maskg(i1+1,i2+1,i3)<=0 ) )
	      {
		if( !Ogen::canDiscretize(mg,iv) )
		{
		  printf("checkOverlappingGrid:ERROR: grid=%i (i1,i2,i3)=(%i,%i,%i) has mask>0 "
			 "but point can NOT be discretized\n",grid,i1,i2,i3);
		  // ok++;
		}
	      }
	    }
	  }
	  else
	  {
	    FOR_3D(i1,i2,i3,I1,I2,I3)
	    {
	      if( maskg(i1,i2,i3)>0 && 
		  (maskg(i1-1,i2-1,i3-1)<=0 || maskg(i1  ,i2-1,i3-1)<=0 || maskg(i1+1,i2-1,i3-1)<=0 ||
		   maskg(i1-1,i2  ,i3-1)<=0 || maskg(i1  ,i2  ,i3-1)<=0 || maskg(i1+1,i2  ,i3-1)<=0 ||
		   maskg(i1-1,i2+1,i3-1)<=0 || maskg(i1  ,i2+1,i3-1)<=0 || maskg(i1+1,i2+1,i3-1)<=0 || 
		   maskg(i1-1,i2-1,i3  )<=0 || maskg(i1  ,i2-1,i3  )<=0 || maskg(i1+1,i2-1,i3  )<=0 ||
		   maskg(i1-1,i2  ,i3  )<=0 ||                             maskg(i1+1,i2  ,i3  )<=0 ||
		   maskg(i1-1,i2+1,i3  )<=0 || maskg(i1  ,i2+1,i3  )<=0 || maskg(i1+1,i2+1,i3  )<=0 || 
		   maskg(i1-1,i2-1,i3+1)<=0 || maskg(i1  ,i2-1,i3+1)<=0 || maskg(i1+1,i2-1,i3+1)<=0 ||
		   maskg(i1-1,i2  ,i3+1)<=0 || maskg(i1  ,i2  ,i3+1)<=0 || maskg(i1+1,i2  ,i3+1)<=0 ||
		   maskg(i1-1,i2+1,i3+1)<=0 || maskg(i1  ,i2+1,i3+1)<=0 || maskg(i1+1,i2+1,i3+1)<=0 ) )
	      {
		if( !Ogen::canDiscretize(mg,iv) )
		{
		  printf("checkOverlappingGrid:ERROR: grid=%i (i1,i2,i3)=(%i,%i,%i) has mask>0 "
			 "but point can NOT be discretized\n",grid,i1,i2,i3);
		  // ok++;
		}
	      }
	      
	    }
	  }
	}
      }
      
    } // for for grid 


  } // end for multigrid levels
  

//   int fromProcessor=0;
//   broadCast(ok,fromProcessor);

  ParallelUtility::getSum(ok);
  
  return ok;
}
