#include "Overture.h"
#include "PlotIt.h"
#include "GraphicsParameters.h"
#include "ShowFileReader.h"
#include "HDF_DataBase.h"


int PlotIt::
userDefinedOutput(const realGridCollectionFunction & uv, 
                  GraphicsParameters & par,
                  const aString & callingFunctionName )
// =================================================================================================
// /Desctiption:
// 
// This routine is called from the contour plotter, for example, when the user chooses
// the "user defined output" option. A user can write a new version of this file in order
// to output values to a file in any given format. 
//
//  You can either change this file, and rebuild the Overture library, or you can make a copy of this
// file, change it, compile it and then link it ahead of the Overture library when building plotStuff
// or some other application.
//
//  /uv (input) : solution values in a grid function
//  /par (input) : graphics parameters (may include a pointer to a show file reader.
//  /callingFunctionName (input) : name of the calling function
// =================================================================================================
{

  GridCollection & gc = *uv.getGridCollection();

  GenericGraphicsInterface & gi = *Overture::getGraphicsInterface(); // graphics interface

  // Here is a pointer to a ShowFileReader (which may be non-NULL if this function was called through plotStuff)
  ShowFileReader *pShowFileReader=(ShowFileReader*)par.showFileReader;
  int solutionNumber = par.showFileSolutionNumber;
  

  printf("userDefinedOutput: callingFunctionName=%s\n",(const char*)callingFunctionName); 

  // here is a menu of possible options
  aString menu[]=  
    {
      "file name",
      "save results",
      "exit",
      ""
    };
  aString answer,answer2;
  gi.appendToTheDefaultPrompt(">user defined");

  aString fileName="user.dat"; // default file name

  aString pdeName="unknown";
  aString reactionName="unknown";
  int rc=-1, uc=-1, vc=-1, wc=-1, tc=-1, pc=-1, numberOfSpecies=0;
  real time=0., dt=0.;

  if( pShowFileReader!=NULL )
  {
    // Look for parameters in the show file
    // (These values are saved, for example, in OverBlown when the show file is written)

    ShowFileReader & showFileReader = *pShowFileReader;
    bool found;

    found=showFileReader.getGeneralParameter("pde",pdeName);
    printf(" pdeName =%s (from ShowFile)\n",(const char*)pdeName);
      
    found=showFileReader.getGeneralParameter("densityComponent",rc);
    found=showFileReader.getGeneralParameter("uComponent",uc);
    found=showFileReader.getGeneralParameter("vComponent",vc);
    found=showFileReader.getGeneralParameter("wComponent",wc);
    found=showFileReader.getGeneralParameter("temperatureComponent",tc);
    found=showFileReader.getGeneralParameter("pressureComponent",pc);

    found=showFileReader.getGeneralParameter("numberOfSpecies",numberOfSpecies);

    found=showFileReader.getGeneralParameter("reactionType",reactionName);

    // get parameters from the current frame
    HDF_DataBase *dbp= showFileReader.getFrame();
    assert( dbp!=NULL );
    HDF_DataBase & db = *dbp;
    db.get(time,"time");
    db.get(dt,"dt");
    printf(" time=%e, dt=%e (from ShowFile)\n",time,dt);
      
  }
    
  for( ;; ) // loop to check for an answer
  {
    gi.getMenuItem(menu,answer,"enter an option");
    
    if( answer=="exit" || answer=="done" )
    {
      break;
    }
    else if( answer=="file name" )
    {
      gi.inputString(fileName,sPrintF(answer,"Enter the name of the file (default=%s)\n",(const char*)fileName));
      if( fileName=="" || fileName==" " ) fileName="user.dat";
    }
    else if( answer=="save results" )
    {

      const int numberOfComponents = uv.getComponentDimension(0); // number of variables: rho,u,v,...
	

      if( gc.numberOfDimensions()==1 )
      {
	// *** Save results from contour cut where the solution is saved on a line ***

        // if callingFunctionName=="contour1d" then the next two components show be available:
	const int xc = numberOfComponents-2;    // component where the x-coordinate is stored
	const int yc = xc+1;                    // component where the y-coordinate is stored

	const int ca=0, cb=numberOfComponents-3;  // [ca,cb] : components where solution variables are stored

        printf("numberOfComponents=%i, ca=%i, cb=%i\n",numberOfComponents,ca,cb);
	for( int n=0; n<numberOfComponents; n++ )
	{
	  printf("component %i, name=[%s] \n",n,(const char*)uv.getName(n));
	}
	
	printf("Saving results to the file=[%s]\n",(const char*)fileName );
	
	FILE *file = fopen((const char*)fileName,"w" );  // open the file
  

	fprintf(file,"%e %i %i %i\n",time,numberOfComponents,1,1);
	int grid;
	for( grid=0; grid<gc.numberOfComponentGrids(); grid++ )
	{
	  realMappedGridFunction & u = uv[grid];
	  const RealDistributedArray & x = gc[grid].center();
	  const IntegerDistributedArray & mask = gc[grid].mask();
	  Index I1,I2,I3;
	  getIndex(gc[grid].gridIndexRange(),I1,I2,I3);

	  fprintf(file,"%i %i\n",I1.getLength(),1);

// 	    if( numberOfComponentsToPlot==5 )
// 	      fprintf(file,"spec-volume velocity    temperature pressure    product\n");
// 	    else if( numberOfComponentsToPlot==6 )
// 	      fprintf(file,"spec-volume velocity    temperature pressure    product     radical\n");
// 	    else
// 	      fprintf(file,"solid frac  density(s)  velocity(s) pressure(s) density(g)  velocity(g) pressure(g)\n");

	  // output the names of the components:
          int n;
          fprintf(file,"x           y           ");
//        fprintf(file,"solid frac  density(s)  velocity(s) pressure(s) density(g)  velocity(g) pressure(g)\n");
	  for( n=ca; n<=cb; n++ )
	  {
	    fprintf(file,"component(%i)",n);
//          fprintf(file,"solid frac  density(s)  velocity(s) pressure(s) density(g)  velocity(g) pressure(g)\n");
	  }
	  fprintf(file,"\n");
	
	  intArray cMask(I1);
	  cMask=mask(I1,I2,I3)!=0 && mask(I1+1,I2,I3)!=0; //  && x(I1)>=xBound(Start,0) && x(I1)<=xBound(End,0);
	  // on refinement grids do not plot cells with all corners hidden by refinement
	  if( gc.numberOfRefinementLevels()>1 )
	    cMask=cMask && !( (mask(I1  ,I2  ,I3) & MappedGrid::IShiddenByRefinement) ||
			      (mask(I1+1,I2  ,I3) & MappedGrid::IShiddenByRefinement) );

	  int i1, i2=0,i3=0;
	  for( i1=I1.getBase(); i1<=I1.getBound(); i1++ )
	  {
	    fprintf(file," %e ",x(i1,i2,i3,0));  // x holds the normalized distance
	    fprintf(file,"%e %e ",u(i1,i2,i3,xc),u(i1,i2,i3,yc));  // (x,y) coordinates
          
	    for( n=ca; n<=cb; n++ )
	    {
	      fprintf(file,"%e ",u(i1,i2,i3,n));  // output solution values
	    }

	    fprintf(file,"%i ",(cMask(i1)? 0 : 1));  // mask value

	    fprintf(file,"\n");
	  }
	}
	
	fclose(file);
      }
      else
      {
	// **** This is a 2D or 3D solution ****

	// nothing implemented in this case

        printf("No results saved -- this option not implemented yet for a 2D or 3D grid function\n");
      }
	
    }
    else 
    {
      cout << "Unknown option =" << answer << endl;
      gi.stopReadingCommandFile();
    }
  }
  gi.unAppendTheDefaultPrompt();

  return 0;
}

