//===============================================================================
//  Compare solutions for convergence testing.
//
//   o read solutions from show files that correspond to different resolutions
//   o estimate errors and convergence rates
//
//==============================================================================
#include "Overture.h"
#include "Ogshow.h"  
#include "ShowFileReader.h"
#include "interpPoints.h"
#include "display.h"
#include "FortranIO.h"
#include "PlotStuff.h"
#include "InterpolatePoints.h"
#include "gridFunctionNorms.h"
#include "ParallelUtility.h"

#define FOR_3D(i1,i2,i3,I1,I2,I3) \
int I1Base =I1.getBase(),   I2Base =I2.getBase(),  I3Base =I3.getBase();  \
int I1Bound=I1.getBound(),  I2Bound=I2.getBound(), I3Bound=I3.getBound(); \
for(i3=I3Base; i3<=I3Bound; i3++) \
for(i2=I2Base; i2<=I2Bound; i2++) \
for(i1=I1Base; i1<=I1Bound; i1++)

real
computeRate( const int & n, const RealArray & h, const RealArray & e, real & sigma, real & c )
//
// Estimate the convergence rate and actual error for a sequence
// of solutions with
//      h(i) : i=0,...,n
//      e(i) = u(h(i))-u(h(n)) i=0,1,...,n-1
// 
//  Assume the solution satisfies:
//          u(h) - u = C h^{sigma}
//
//     then 
//          e(i) == u(h(i)) - u(h(n)) = C( h^s - h(n)^s )   (s==sigma)
//
// Given at least 3 solutions (N>=2) we can get estimates for sigma and C
//   To get sigma we need to find a solution to the nonlinear equation
//      e(0)/e(1) = ( d(0)^s -1 ) /( d(1)^s -1 ) 
//   where
//        d(i) = h(i)/h(n) > 1 and 
// Thus we need to solve
//    f(s) = delta (d(1)^s -1)  - (d(0)^s -1) = 0    (delta=e(0)/e(1)
//  
{
  assert( n>=2 );
  
  if( n>2 )
  {
    printF(" ** computeRate ** WARNING: I will only consider the 3 finest solutions when computing the rates\n");
  }
  

//   real delta=e(0)/e(1);
//   real d0=h(0)/h(n);
//   real d1=h(1)/h(n);
  
  const int m0=n-2, m1=n-1;
  real delta=e(m0)/e(m1);
  real d0=h(m0)/h(n);
  real d1=h(m1)/h(n);
  
  real s= log(delta)/log(d0/d1);   // initial guess
  real ds;
  const real eps=1.e-3;
  
  if( delta>=3. )
  {
    // this doesn't seem to work if delta is too small!
    real f0,df,t0,t1;
    for( int it=0; it<21; it++ )
    {
      f0=delta*(pow(d1,s)-1.) - (pow(d0,s)-1);  
      df= delta*log(d1)*pow(d1,s)-log(d0)*pow(d0,s);
//    t1=pow(d1,s);
//    t0=pow(d0,s);
//    f0=delta*(t1-1.)/(t0-1.)-1.;
//    df=delta*( log(d1)*t1*(t0-1.)-(t1-1.)*log(d0)*t0 )/( (t0-1.)*(t0-1.) );
      ds = -f0/df;
      if( fabs(ds)>.1 )
	ds=.1*ds/fabs(ds);
      s+=ds;
      if( it>10 )
	printf("it=%i, delta=%e, f0=%e, df=%e, ds=%e, s=%e \n",it,delta,f0,df,ds,s);
      if( fabs(ds)<eps )
	break;
    }
    if( false && fabs(ds)>eps )
    {
      cout << "computeRate: ERROR no convergence\n";
      printf(" e(%i)=%e, e(m1)=%e \n",m0,e(m0),m1,e(m1));
    }
  }

  if( fabs(ds)>eps )
  { // ** here is a backup method ***
    // do a search for the best guess at s
    real sa=.1, sb=2.;
    real ns=101;
    real sBest=sa, fMin=delta*log(d1)*pow(d1,sa)-log(d0)*pow(d0,sa);;
    for( int i=0; i<ns; i++ )
    {
      real s = sa+(sb-sa)*i/ns;
      real f= delta*log(d1)*pow(d1,s)-log(d0)*pow(d0,s);
      if( f<fMin )
      {
	sBest=s;
        fMin=f;
      }
    }
    //  printf("WARNING: The estimated convergence rate is wrong \n");
  }
  
  sigma=s;
  c=e(m0)/(pow(h(m0),s)-pow(h(n),s));
  return s;
}




int 
main(int argc, char *argv[])
{
  Overture::start(argc,argv);  // initialize Overture

  if( false )
  {
    // --- test computeRate
    int n=2;
    RealArray h(n+1),u(n+1),e(n+1);
    real c0=.234, s0=2.2;
    for( int i=0; i<n+1; i++ )
    {
      h(i)=1./pow(2.,double(i+1));
      u(i)=1. + c0*pow(h(i),s0);  // our solution
    }
    for( int i=0; i<n; i++ )
      e(i)=u(i)-u(n);
    
    real sigma,c;
    computeRate(n,h,e,sigma,c);
    printf(" sigma=%e (true=%e), c=%e (true=%e) \n",sigma,s0,c,c0);
    return 0;
  }

  bool plotOption=true;
  bool closeShowAfterUse=true;
  bool useOldWay=false;
  
  aString commandFileName="";
  if( argc > 1 )
  { // look at arguments for "noplot" or some other name
    aString line;
    for( int i=1; i<argc; i++ )
    {
      line=argv[i];
      if( line=="-noplot" || line=="noplot" )
        plotOption=false;
      else if( line=="-useOld" )
        useOldWay=true;
      else if( commandFileName=="" )
      {
        commandFileName=line;    
        printf("comp: reading commands from file [%s]\n",(const char*)commandFileName);
      }
      
    }
  }

  const int maxNumberOfFiles=15;

  int numberOfFiles=0;
  aString fileName[maxNumberOfFiles];

  ShowFileReader showFileReader[maxNumberOfFiles];
  int numberOfSolutions[maxNumberOfFiles];
  int solutionNumber[maxNumberOfFiles];
  int maxSolution[maxNumberOfFiles];
  for( int i=0; i<maxNumberOfFiles; i++ )
  {
    solutionNumber[i]=-1;
    maxSolution[i]=INT_MAX;   // the largest number of solutions common to all files
  }
  
  CompositeGrid cg[maxNumberOfFiles];
  realCompositeGridFunction u[maxNumberOfFiles], ud[maxNumberOfFiles];
  RealArray h(maxNumberOfFiles);
  const int maxNumberOfComponents=20;
  RealArray time(maxNumberOfFiles);

  RealArray maxDiff(maxNumberOfFiles,maxNumberOfComponents);
  RealArray l2Diff(maxNumberOfFiles,maxNumberOfComponents);
  RealArray l1Diff(maxNumberOfFiles,maxNumberOfComponents);

  // RealArray estimatedError(maxNumberOfFiles,maxNumberOfComponents);
 
  int extra=0;  // set to -1 to not check boundary

  PlotStuff ps(plotOption,"comp");                      // create a PlotStuff object
  PlotStuffParameters psp;           // create an object that is used to pass parameters
    
  // By default start saving the command file called "comp.cmd"
  aString logFile="comp.cmd";
  ps.saveCommandFile(logFile);
  cout << "User commands are being saved in the file `" << (const char *)logFile << "'\n";

  aString outputFileName="comp.log";
  FILE *outFile = fopen((const char*)outputFileName,"w" );
  printF("Output being saved in file %s\n",(const char*)outputFileName);
  

  ps.appendToTheDefaultPrompt("comp>");

  // read from a command file if given
  if( commandFileName!="" )
  {
    cout << "read command file =" << commandFileName << endl;
    ps.readCommandFile(commandFileName);
  }

  bool errorsComputed=false;
  bool timesMatch=true;

  aString answer,answer2;
  aString menu[] = {"specify files (coarse to fine)",
                    "choose a solution",
                    "choose a solution for each file",
                    "compute errors",
                    "plot solutions",
                    "plot differences",
                    // "do not check boundaries (toggle)",
                    // "output ascii files",
                    // "output binary files",
		    "exit",
                    "" };
  char buff[80];
  for(;;)
  {
    ps.getMenuItem(menu,answer);
    if( answer=="specify files (coarse to fine)" )
    {
      for( int i=0; i<maxNumberOfFiles; i++ )
      {
	ps.inputString(fileName[numberOfFiles],"Enter the file name (`exit' to finish)");
	if( fileName[numberOfFiles]=="" || fileName[numberOfFiles]=="exit" )
          break;
	showFileReader[numberOfFiles].open(fileName[numberOfFiles]);
	numberOfSolutions[numberOfFiles]=showFileReader[numberOfFiles].getNumberOfFrames();
	maxSolution[numberOfFiles]=min(maxSolution[numberOfFiles],numberOfSolutions[numberOfFiles]);
 
	if( closeShowAfterUse )
	  showFileReader[i].close();


	numberOfFiles++;
      }
    }
    else if( answer=="choose a solution" ||
             answer=="choose a solution for each file" )
    {
      // In this case the user is asked to choose a solution to read in
      // Choosing a number that is too large will cause the last solution to be read 

      aString line;
      if( answer=="choose a solution" )
      {
	ps.inputString(line,sPrintF(buff,"Enter the solution number to read, in [1,%i] \n",maxSolution[0]));
	sScanF(line,"%i",&solutionNumber[0]);
        for( int i=1; i<maxNumberOfFiles; i++ ) solutionNumber[i]=solutionNumber[0];
      }
      else
      {
        for( int i=0; i<numberOfFiles; i++ ) 
          printF(" File %i has solutions [1,%i]\n",i,maxSolution[i]);
	ps.inputString(line,"Enter separate soltuion number of each file\n");
	assert( maxNumberOfFiles<15 );
	sScanF(line,"%i %i %i %i %i %i %i %i %i %i %i %i %i %i %i %i",
               &solutionNumber[0],&solutionNumber[1],&solutionNumber[2],&solutionNumber[3],&solutionNumber[4],
               &solutionNumber[5],&solutionNumber[6],&solutionNumber[7],&solutionNumber[8],&solutionNumber[9],
               &solutionNumber[10],&solutionNumber[11],&solutionNumber[12],&solutionNumber[13],&solutionNumber[14]);
      }
      

      for( int i=0; i<numberOfFiles; i++ )
      {
        if( closeShowAfterUse )
         showFileReader[i].open(fileName[i]);

        showFileReader[i].getASolution(solutionNumber[i],cg[i],u[i]);        // read in a grid and solution
        HDF_DataBase & db = *(showFileReader[i].getFrame());
        db.get(time(i),"t");  
        printf(" file[%i] solutionNumber=%i time=%e \n",i,solutionNumber[i],time(i));

        fPrintF(outFile,"Choosing solution %i, t=%9.3e,  from showFile=%s\n",solutionNumber[i],time(i),(const char*)fileName[i]);
	

        if( closeShowAfterUse )
	  showFileReader[i].close();

        cg[i].update(MappedGrid::THEmask );
        // cg[i].update(MappedGrid::THEmask | MappedGrid::THEcenter | MappedGrid::THEvertex |
	//	     MappedGrid::THEinverseVertexDerivative);
	
      }
      timesMatch=true;
      for( int i=1; i<numberOfFiles; i++ )
      {
        if( fabs(time(i)-time(i-1)) > REAL_EPSILON*max(fabs(time(i)),fabs(time(i-1)))*100. )
	{
          timesMatch=false;

	  printF("***************ERROR: The times of the solutions do not match! ****************\n"
		 "   times=");
	  for( int i=0; i<numberOfFiles; i++ )
	  {
	    printF("%9.3e, ",time(i));
	  }
	  printF("\n *********************************************************************************\n");

          break;
	}
      }
    }
    else if( answer=="output ascii files" )
    {
      aString outputFileName;
      ps.inputString(outputFileName,"Enter the output file name ");
      FILE *file;
      for( int i=0; i<numberOfFiles; i++ )
      {
        for( int solution=1; solution<=maxSolution[i]; solution++ )
	{
	  showFileReader[i].getASolution(solution,cg[i],u[i]);        // read in a grid and solution
	  HDF_DataBase & db = *(showFileReader[i].getFrame());
	  db.get(time(i),"t");  
	  

	  MappedGrid & mg = cg[i][0];
	  realMappedGridFunction & v = u[i][0];
	
	  const IntegerArray & gridIndexRange = cg[i][0].gridIndexRange();
          if( solution==1 )
	  {
  	    int n=gridIndexRange(End,axis1)-gridIndexRange(Start,axis1);
	    aString name;
	    name=sPrintF(buff,"%s%i.dat",(const char*)outputFileName,n);
	    printf("save file %s \n",(const char*)name);
	    file = fopen((const char*)name,"w");
	  }
	  DisplayParameters dp;
	  dp.set(file);
	  dp.set(DisplayParameters::labelNoIndicies);
	  Index I1,I2,I3;
	  getIndex(cg[i][0].gridIndexRange(),I1,I2,I3);
	  for( int c=v.getComponentBase(0); c<=v.getComponentBound(0); c++ )
	  {
	    cout << "component Name = [" << u[i].getName(c) << "]\n";
	  
	    fprintf(file,"%s\n",(const char*)u[i].getName(c));
            fprintf(file,"%e (time)\n",time(i));
	    fprintf(file,"%i %i %i\n",I1.getBound()-I1.getBase()+1,
		    I2.getBound()-I2.getBase()+1,
		    I3.getBound()-I3.getBase()+1);
	    display(u[i][0](I1,I2,I3,c),NULL,dp);
	  }
	}
	fclose(file);
      }
    }
/* --------
   else if( answer=="output binary files" )
   {
   aString outputFileName;
   ps.inputString(outputFileName,"Enter the output file name ");
   FILE *file;
   FortranIO fio;
   for( int i=0; i<numberOfFiles; i++ )
   {
   for( int solution=1; solution<=maxSolution; solution++ )
   {
   showFileReader[i].getASolution(solution,cg[i],u[i]);        // read in a grid and solution
   HDF_DataBase & db = *(showFileReader[i].getFrame());
   db.get(time(solution),"t");  
	  

   MappedGrid & mg = cg[i][0];
   realMappedGridFunction & v = u[i][0];
	
   const IntegerArray & gridIndexRange = cg[i][0].gridIndexRange();
   if( solution==1 )
   {
   int n=gridIndexRange(End,axis1)-gridIndexRange(Start,axis1);
   aString name;
   name=sPrintF(buff,"%s%i.dat",(const char*)outputFileName,n);
   printf("save file %s \n",(const char*)name);
   file = fopen((const char*)name,"w");
   fio.open((const char*)name,"unformatted","unknown");
	    
   }
   DisplayParameters dp;
   dp.set(file);
   dp.set(DisplayParameters::labelNoIndicies);
   Index I1,I2,I3;
   getIndex(cg[i][0].gridIndexRange(),I1,I2,I3);
   for( int c=v.getComponentBase(0); c<=v.getComponentBound(0); c++ )
   {
   cout << "component Name = [" << u[i].getName(c) << "], t=" << time(solution) << "\n";
	  
   int i2=(I2.getBase()+I2.getBound())/2;
   display(u[i][0](I1,i2,I3,c));

   fio.print((const char*)u[i].getName(c));
   fio.print(time(solution));
   fio.print(u[i][0](I1,I2,I3,c));
   }
   }
   fio.close();
   }
   }
   ------------- */
//     else if( answer=="do not check boundaries (toggle)" )
//     {
//       extra= extra==0 ? -1 : 0;
//     }
    else if( answer=="compute errors" )
    {
      if( solutionNumber[0]<0 )
      {
	printf(" You should `choose a solution' before computing errors\n");
	continue;
      }

      errorsComputed=true;
      

      Index I1,I2,I3, J1,J2,J3;
      const int n =numberOfFiles-1;  // finest grid

      // **** determine relative mesh spacings here ***
      // Assume the base grids are basically the same but with different numbers of grid points.
      // There may also be different numbers of refinement levels.
      for( int i=0; i<numberOfFiles; i++ )
      {
        h(i)=cg[i][0].gridSpacing(axis1);    // grid spacing on grid 0
	if( cg[i].numberOfRefinementLevels()>1 )
	{
	  int nl=cg[i].numberOfRefinementLevels();
	  int rf=cg[i].refinementLevel[nl-1].refinementFactor(0,0);
	  // NOTE: rf is the refinement factor to the base grid.
	  // printf(" numberOfLevels=%i refinementFactor=%i\n",nl,rf);
	  h(i)/=rf;
	}
      }
      for( int i=0; i<numberOfFiles; i++ )
        printf(" File %i, solution=%i, t=%9.3e, grid 0: dr = %10.3e , ratio to fine grid = %8.4f\n",i,solutionNumber[i],
                time(i),h(i),h(i)/h(n));
      
      realCompositeGridFunction & vn = u[n];  // Here is the fine grid solution 
      for( int i=0; i<n; i++ )
      {
	const realCompositeGridFunction & v = u[i];
        ud[i].updateToMatchGridFunction(v);
        ud[i]=0.;
        
        printf("\n >> Interpolate the fine grid solution onto the grid from file %i...\n",i);

        // This next call can be used to call the old or new method
	if( useOldWay )
	{
	  printF("\n +++++++++++++++++++++ USE OLD INTERP ++++++++++++++++++++++\n\n");
	  bool useNewWay=false;
	  interpolateAllPoints( vn,ud[i],useNewWay );  // interpolate ud[i] from fine grid solution vn
	}
	else
	{
	  // *new way*
	  printF("\n +++++++++++++++++++++ USE NEW INTERP ++++++++++++++++++++++\n\n");
	  InterpolatePoints interpPoints; 
	  interpPoints.setInfoLevel( 1 );
          int numGhost=0;  // no need to interpolate ghost points
          Range C(ud[i].getComponentBase(0),ud[i].getComponentBound(0));
	  interpPoints.interpolateAllPoints( vn,ud[i],C,C,numGhost );  // interpolate ud[i] from fine grid solution vn
	}
	
        ud[i]-=v;

	// zero out points in the difference that are hidden by refinement so that these are not counted 
	if( false ) // this is nolonger needed -- the norm will not count these
	{
          realCompositeGridFunction & u = ud[i];
          CompositeGrid & cg = (CompositeGrid&)(*u.gridCollection);
          const int c0=u.getComponentBase(0), c1=u.getComponentBound(0);
          Index I1,I2,I3;
	  for( int grid=0; grid<cg.numberOfComponentGrids(); grid++ )
	  {
	    if( cg.refinementLevelNumber(grid)<cg.numberOfRefinementLevels()-1 )
	    {
	      
              MappedGrid & mg = cg[grid];
              #ifdef USE_PPP
               realSerialArray uLocal; getLocalArrayWithGhostBoundaries(u[grid],uLocal);
               intSerialArray maskLocal; getLocalArrayWithGhostBoundaries(mg.mask(),maskLocal);
              #else
               const realSerialArray & uLocal = u[grid];
               const intSerialArray & maskLocal = mg.mask();
              #endif 
              getIndex(mg.gridIndexRange(),I1,I2,I3);
	      bool ok = ParallelUtility::getLocalArrayBounds(u[grid],uLocal,I1,I2,I3);
	      if( !ok ) continue;
              int count=0;
              int i1,i2,i3;
              FOR_3D(i1,i2,i3,I1,I2,I3)
	      {
		if( maskLocal(i1,i2,i3) &  MappedGrid::IShiddenByRefinement )
		{
                  count++;
                  for( int c=c0; c<=c1; c++ )
                    uLocal(i1,i2,i3,c)=0.;
		}
	      }

              printF("comp:i=%i: Zeroed out %i hidden points for grid=%i (level=%i, numLevels=%i)\n",i,count,grid, 
                       cg.refinementLevelNumber(grid),cg.numberOfRefinementLevels());

	    }
	  }
	  
	}
	

	const int useAreaWeightedNorm=1;
        for( int c=v.getComponentBase(0); c<=v.getComponentBound(0); c++ )
	{
	  l2Diff(i,c) = lpNorm(2,ud[i],c,0,0,useAreaWeightedNorm);
	  l1Diff(i,c) = lpNorm(1,ud[i],c,0,0,useAreaWeightedNorm);
	  maxDiff(i,c)=maxNorm(ud[i],c,0,0);
	}
	for( int io=0; io<=1; io++ )
	{
	  FILE *file = io==0 ? stdout : outFile;
	  fPrintF(file,"h(%i)=%e: \n",i,h(i));
	  for( int c=v.getComponentBase(0); c<=v.getComponentBound(0); c++ )
	    fPrintF(file," maxDiff(%i)=%e, l2Diff(%i)=%e , l1Diff(%i)=%e \n",c,maxDiff(i,c),c,l2Diff(i,c),c,l1Diff(i,c));
	}
	fflush(outFile);
      }
      // estimate convergence rate
      if( n>=2 )
      {
	for( int io=0; io<=1; io++ )
	{
	  FILE *file = io==0 ? stdout : outFile;
	  fPrintF(file,"\n Solutions at times=");
	  for( int i=0; i<numberOfFiles; i++ )
	  {
	    fPrintF(file,"%9.3e, ",time(i));
	  }
	  fPrintF(file,"\n");
	}
	
	for( int norm=0; norm<=2; norm++ )
	{
	  for( int io=0; io<=1; io++ )
	  {
	    FILE *file = io==0 ? stdout : outFile;
	    if( norm==0 )
	      fPrintF(file,"++++++++++++++ max norm results +++++++++++++\n");
	    else if( norm==1 )
	      fPrintF(file,"++++++++++++++ l2 norm results +++++++++++++\n");
	    else 
	      fPrintF(file,"++++++++++++++ l1 norm results +++++++++++++\n");
	    fPrintF(file,"    ee = estimated error from C*h^{rate}     \n");
	  }
	
	  const RealArray & diff = norm==0 ? maxDiff : norm==1 ? l2Diff : l1Diff;
	  real sigma,cc;
	  Range R(0,n);
	  for( int c=u[0].getComponentBase(0); c<=u[0].getComponentBound(0); c++ )
	  {
	    computeRate(n,h,diff(R,c),sigma,cc);

	    for( int io=0; io<=1; io++ )
	    {
	      FILE *file = io==0 ? stdout : outFile;

	      fPrintF(file," component=%i, %11s, rate=%5.2f, ",c,(const char*)u[0].getName(c),sigma);
	      for( int i=0; i<=n; i++ )
	      { // estimated errors:
		fPrintF(file,"ee(%i) = %8.2e, ",i,cc*pow(h(i),sigma));
	      }
	      fPrintF(file,"\n");
	    }
	  }
	}
	fflush(outFile);
	
      }
      if( !timesMatch )
      {
	for( int io=0; io<=1; io++ )
	{
	  FILE *file = io==0 ? stdout : outFile;
	  fPrintF(file,"\n***************WARNING: The times of the solutions do not match! ****************\n"
		 "   times=");
	  for( int i=0; i<numberOfFiles; i++ )
	  {
	    fPrintF(file,"%9.3e, ",time(i));
	  }
	  fPrintF(file,"\n *********************************************************************************\n");
	}
      }
      
    }
    else if( answer=="plot solutions" )
    {
      for( int i=0; i<numberOfFiles; i++ )
      {
	psp.set(GI_TOP_LABEL,sPrintF(buff,"u[%i] t=%9.3e",i,time(i)));
        ps.erase();
	PlotIt::contour(ps,u[i],psp);
      }
    }
    else if( answer=="plot differences" )
    {
      if( !errorsComputed )
      {
	printf("You should compute the errors before you can plot the differences\n");
	continue;
      }

      for( int i=0; i<numberOfFiles-1; i++ )
      {
	psp.set(GI_TOP_LABEL,sPrintF(buff,"u[%i]-u[%i] t=%9.3e, %9.3e",i,numberOfFiles-1,time(i),time(numberOfFiles-1)));
        ps.erase();
	PlotIt::contour(ps,ud[i],psp);
      }
    }
    else if( answer=="exit" )
    {
      break;
    }
  }

  fclose(outFile);

  Overture::finish();          
  return 0;
}
