#ifndef INTEGRATE_H
#define INTEGRATE_H "Integrate.h"

// =======================================================================================
//    Use this class to integrate functions on overlapping grids, both volume
//  and surface integrals.
// =======================================================================================

#include "Overture.h"
#include "BodyDefinition.h"

class Oges;
class SurfaceStitcher;

class Integrate
{
 public:

  Integrate();
  Integrate( CompositeGrid & cg );
  ~Integrate();
  
  int numberOfFacesOnASurface(const int surfaceNumber) const;
  int getFace(const int surfaceNumber,const int face, 
              int & side, int & axis, int & grid) const;

  const BodyDefinition & getBodyDefinition() const;
  
  // return a pointer to the surface stitcher (if it exists)
  SurfaceStitcher* getSurfaceStitcher() const;

  int defineSurface( const int & surfaceNumber, const int & numberOfFaces_, IntegerArray & boundary ); 

  // turn on interactive stitching (for debugging)
  void setInteractiveStitching( bool trueOrFalse );

  real surfaceIntegral(const RealCompositeGridFunction & u, const int & surfaceNumber = -1 );

  int surfaceIntegral(const RealCompositeGridFunction & u, 
                      const Range & C, 
                      RealArray & integral,
                      const int & surfaceNumber = -1 );

  int updateToMatchGrid( CompositeGrid & cg );

  int updateForAMR(CompositeGrid &cg);  // call this function when AMR grids have changed.

  real volumeIntegral( const RealCompositeGridFunction & u );
  
  RealCompositeGridFunction & integrationWeights();
  RealCompositeGridFunction & leftNullVector();

  // use AMR grids when computing integrals on grid functions that have AMR
  void useAdaptiveMeshRefinementGrids(bool trueOrFalse = true );

  // use hybrid grids to compute integrals
  void useHybridGrids( bool trueOrFalse = true );

  void setRadialAxis( int axis );

  static int debug;

 protected:


  int initialize();
  int computeWeights();
  int computeSurfaceWeights(int surfaceNumber=-1 );
  
  int computeLeftNullVector();
  int surfaceIndex( int surfaceNumber );

  // kkc 030225 added surface stitching using ugen for surface integrals
  int computeStitchedSurfaceWeightsOld();

  // new version: *wdh* 070227 
  int computeStitchedSurfaceWeights(int surfaceNumber=-1);

  int buildAdaptiveMeshRefinementSurfaceWeights(CompositeGrid & cgu, 
					        const int & surfaceNumber = -1 );
  int destroyAdaptiveMeshRefinementIntegrationArrays();


  bool useSurfaceStitching; 
  bool interactiveStitcher;  // call the stitcher in an interactive mode (for debugging)
  SurfaceStitcher *pSurfaceStitcher;  // for stitching surfaces

  BodyDefinition bd;
  CompositeGrid cg;

  int orderOfAccuracy;
  //  int numberOfSurfaces,maximumNumberOfFaces;
  //  IntegerArray surfaceIdentifier, numberOfFaces;
  //  IntegerArray boundaryFaces, boundaryHasOverlap;
  IntegerArray boundaryHasOverlap;
  IntegerArray faceWeightsDefined, surfaceWeightsDefined;
  
  realCompositeGridFunction weights, nullVector;
  
  Oges *solver;
  bool weightsComputed, leftNullVectorComputed;
  bool allFaceWeightsDefined, weightsUpdatedToMatchGrid;

  // Weights for AMR grids 
  bool useAMR;
  int *numberOfFacesPerSurface;  // total number of faces for each surface, including AMR grids
  int numberOfBoundarySurfaces;
  int numberOfBoundaryRefinementLevels;
  int *numberOfBoundaryFaces;
  realArray ***boundaryWeights;
  int **pNumberOfBoundaryGrids;
  int ***pBoundaryGrid;

  // axisymmetric stuff
  int radialAxis;
};

#endif
