#include "display.h"
#include "ParallelUtility.h"
#include "MappedGridOperators.h"
#include "SparseRep.h"

#define FOR_3D(i1,i2,i3,I1,I2,I3) \
    int I1Base =I1.getBase(),   I2Base =I2.getBase(),  I3Base =I3.getBase();  \
    int I1Bound=I1.getBound(),  I2Bound=I2.getBound(), I3Bound=I3.getBound(); \
    for(i3=I3Base; i3<=I3Bound; i3++) \
    for(i2=I2Base; i2<=I2Bound; i2++) \
    for(i1=I1Base; i1<=I1Bound; i1++)

// ===================================================================================================================
/// \brief Output the equations contained in a coefficient matrix
/// \param coeff: output the equations for this matrix.
/// \param label: write this label as the header. 
/// \param file: write results to this file. In parallel, only the local array is written to this file and thus there
///              should be a different file for each processor. 
/// \param format: output format
///
// ===================================================================================================================
int
displayCoeff(realMappedGridFunction &coeff,
	     const aString & label,
	     FILE *file /* =stdout */,
	     const aString format /* =nullString */ )
{
  MappedGrid & mg = *coeff.getMappedGrid();
  const int numberOfDimensions = mg.numberOfDimensions();

  assert( coeff.sparse!=NULL );
  SparseRepForMGF & sparse = *coeff.sparse;
  int numberOfComponentsForCoefficients = sparse.numberOfComponents;  // size of the system of equations
  int numberOfGhostLines = sparse.numberOfGhostLines;
  int stencilSize = sparse.stencilSize;
  int stencilDim=stencilSize*numberOfComponentsForCoefficients; // number of coefficients per equation

  if( label.length()>0 )
    fprintf(file,"%s\n",(const char*)label);

#ifdef USE_PPP
  realSerialArray coeffLocal; getLocalArrayWithGhostBoundaries(coeff,coeffLocal);
  intSerialArray maskLocal; getLocalArrayWithGhostBoundaries(mg.mask(),maskLocal);
#else
  realSerialArray & coeffLocal = coeff;
  intSerialArray & maskLocal = mg.mask();
#endif

  //    mce2(m1,m2,m3,c,e) = ma2(m1,m2,m3) + stencilSize*(c+ncc*e)
  //    mce3(m1,m2,m3,c,e) = ma3(m1,m2,m3) + stencilSize*(c+ncc*e)
  #define ce(c,e) stencilSize*((c)+numberOfComponentsForCoefficients*(e))

  Index I1,I2,I3;
  getIndex(mg.gridIndexRange(),I1,I2,I3,numberOfGhostLines);

  bool ok = ParallelUtility::getLocalArrayBounds(mg.mask(),maskLocal,I1,I2,I3);
  if( ok )
  {
    int i1,i2,i3;
    FOR_3D(i1,i2,i3,I1,I2,I3)
    {
      for( int e=0; e<numberOfComponentsForCoefficients; e++ )
      {
	for( int c=0; c<numberOfComponentsForCoefficients; c++ )
	{
	  if( numberOfDimensions==2 )
	  {
            if( numberOfComponentsForCoefficients==1 )
	    {
              fprintf(file," (%i,%i) coeff=",i1,i2);
	    }
	    else if( c==0 )
	    {
	      fprintf(file," (%i,%i) e=%i c=%i coeff=",i1,i2,e,c);
	    }
	    else
	      fprintf(file,"                 c=%i coeff=",c);
	  }
	  else
	  {
            if( numberOfComponentsForCoefficients==1 )
	    {
              fprintf(file," (%i,%i,%i) coeff=",i1,i2,i3,e,c);
	    }
	    else if( c==0 )
	    {
	      fprintf(file," (%i,%i,%i) e=%i c=%i coeff=",i1,i2,i3,e,c);
	    }
	    else
	      fprintf(file,"                     c=%i coeff=",c);
	  }

	  for( int m=ce(c,e); m<ce(c+1,e); m++ )
	  {
	    fprintf(file,"%9.2e ",coeffLocal(m,i1,i2,i3));
	  }
          fprintf(file,"\n");
	}
      }
    }
  }

  return 0;
}
