#include "Ogen.h"
#include "PlotStuff.h"
#include "MatrixTransform.h"
#include "Ogshow.h"

// ========================================================================================================
// Moving Grid Example: 
//   o read in a grid from a data-base file, rotate a component grid and recompute the overlapping grid.
//   o interpolate a grid function, update the interpolate for the new grid
//   o save solutions in a show file
// 
// Usage:
//   move1 [-grid=<name>] [-shift] [-rotate] [-interpolate]
// 
// Examples:
//   move1 -grid=cic
//  srun -N1 -n1 -ppdebug move1 -grid=cic
//  totalview srun -a -N1 -n1 -ppdebug move1 -grid=cic
// ========================================================================================================
int 
main(int argc, char *argv[])
{
  Overture::start(argc,argv);  // initialize Overture

  Mapping::debug=0; 

  int numGhost=2;  // for second-order accurate (1 is good enough for implicit)
  MappedGrid::setMinimumNumberOfDistributedGhostLines(numGhost);

  printF(" moving grid demo: \n"
         " move1 [-grid=<name>] [-shift] [-rotate] [-interpolate]\n");

  enum
  {
    rotate,
    shift
  } moveOption=shift;

  int numberOfSteps=20;
  real deltaAngle=5.*Pi/180.;
  real xShift=-.01;

  // aString nameOfOGFile = "cice2.order2.hdf";
  aString nameOfOGFile = "cic.hdf";

  int plotOption=true;
  bool interpolate=false;
  if( argc > 1 )
  { // look at arguments for "noplot"
    aString line;
    int len=0;
    for( int i=1; i<argc; i++ )
    {
      line=argv[i];
      if( line=="-noplot" || line=="noplot" )
        plotOption=false;
      else if( len=line.matches("-grid=") )
      {
	nameOfOGFile=line(len,line.length()-1);
      }
      else if( line=="-shift" )
      {
	moveOption=shift;
      }
      else if( line=="-rotate" )
      {
	moveOption=rotate;
      }
      else if( line=="-interpolate" )
      {
	interpolate=true;
      }
      
      
    }
  }


  PlotStuff ps(plotOption,"Moving Grid Example");         // for plotting
  PlotStuffParameters psp;

  // Create two CompositeGrid objects, cg[0] and cg[1]
  CompositeGrid cg[2];                             
  getFromADataBase(cg[0],nameOfOGFile);             // read cg[0] from a data-base file
//cg[0].update();
  cg[1]=cg[0];                                      // copy cg[0] into cg[1]

  if( true )
  {
    psp.set(GI_TOP_LABEL,"initial grid");  // set title
    PlotIt::plot(ps,cg[0],psp);
  }
  

  // Rotate component grid 1 (do this by changing the mapping)
  int gridToMove=1;
  Mapping & mappingToMove = *(cg[0][gridToMove].mapping().mapPointer);

  // Use this MatrixTransform to change the existing Mapping, the MatrixTransform
  // can rotate/scale and shift any Mapping, keep a transform for each composite grid
  MatrixTransform transform0(mappingToMove); 
  MatrixTransform transform1(mappingToMove); 

  // Replace the mapping of the component grid that we want to move:
  cg[0][gridToMove].reference(transform0); 
  cg[0].update();
  cg[1][gridToMove].reference(transform1); 
  cg[1].update();
  
  if( true )
  {
    ps.erase();
    psp.set(GI_TOP_LABEL,"cg[0] after reference to transform");  // set title
    PlotIt::plot(ps,cg[0],psp);
  }


  // now we destroy all the data on the new grid -- it will be shared with the old grid
  // this is not necessary to do but it will save space
  cg[1].destroy(CompositeGrid::EVERYTHING);  

  // we tell the grid generator which grids have changed
  LogicalArray hasMoved(2);
  hasMoved    = LogicalFalse;
  hasMoved(gridToMove) = LogicalTrue;  // Only this grid will move.
  char buff[80];

  // Here is the overlapping grid generator
  Ogen gridGenerator(ps);

  // update the initial grid, since the above reference destroys the mask
  gridGenerator.updateOverlap(cg[0]);

  if( true )
  {
    ps.erase();
    psp.set(GI_TOP_LABEL,"cg[0] gridGenerator.updateOverlap");  // set title
    PlotIt::plot(ps,cg[0],psp);
  }
  

  // Here is an interpolant
  Interpolant interpolant(cg[0]);
  realCompositeGridFunction u(cg[0]);
  // Here is a show file
  Ogshow show("move1.show");
  show.setIsMovingGridProblem(true);


  // ---- Move the grid a bunch of times.----
  for (int i=1; i<=numberOfSteps; i++) 
  {
    int newCG = i % 2;        // new grid
    int oldCG = (i+1) % 2;    // old grid
    
    printF("--- take step %i\n",i);

    if( i>0 )
    {
      ps.erase();
      psp.set(GI_TOP_LABEL,sPrintF(buff,"Solution at step=%i",i));  // set title
      PlotIt::plot(ps,cg[oldCG],psp);      // plot the current overlapping grid
      psp.set(GI_PLOT_THE_OBJECT_AND_EXIT,true);     // set this to run in "movie" mode (after first plot)
      ps.redraw(true);
    }
    
    // Rotate the grid by rotating the mapping
    // After the first step we must double the angle since we start from the old grid
    if( moveOption==rotate )
    {
      real angle = i==1 ? deltaAngle : deltaAngle*2.; 
      if( newCG==0 )
	transform0.rotate(axis3,angle);
      else
	transform1.rotate(axis3,angle);
    }
    else
    {
      real delta = i==1 ? xShift : xShift*2.; 
      if( newCG==0 )
	transform0.shift(delta,0.,0.);
      else
	transform1.shift(delta,0.,0.);
    }
    // Update the overlapping newCG, starting with and sharing data with oldCG.
    Ogen::MovingGridOption option = Ogen::useOptimalAlgorithm;
    int useFullAlgorithmInterval=10; // 10000;
    if( i% useFullAlgorithmInterval == useFullAlgorithmInterval-1  )
    {
      printF("\n +++++++++++ use full algorithm in updateOverlap +++++++++++++++ \n");
      option=Ogen::useFullAlgorithm;
    }
    gridGenerator.updateOverlap(cg[newCG], cg[oldCG], hasMoved, option );

    if( interpolate )
    {
      interpolant.updateToMatchGrid(cg[newCG]);
      u.updateToMatchGrid(cg[newCG]);
    }
    
    // assign values to u
    Index I1,I2,I3;
    for( int grid=0; grid<cg[newCG].numberOfComponentGrids(); grid++ )
    {
      MappedGrid & mg = cg[newCG][grid];
      getIndex(mg.dimension(),I1,I2,I3);
      real freq = 2.*i/numberOfSteps;
      u[grid](I1,I2,I3)=cos(freq*mg.vertex()(I1,I2,I3,axis1))*sin(freq*mg.vertex()(I1,I2,I3,axis2));  
    }
    if( interpolate )
      u.interpolate();
    // save the result in a show file, every fourth step
    if( (i % 4) == 1 )
    {
      show.startFrame();
      show.saveComment(0,sPrintF(buff,"Solution form move1 ar step = %i",i));
      show.saveSolution(u);
    }
  } 
  printF("Results saved in move1.show, use Overture/bin/plotStuff to view this file\n");
  Overture::finish();          
  return 0;
}

